@@ -263,24 +263,24 @@ int sergensio_ri_b(struct sergensio_b *sbio, int *ri);
  * requested the mask be changed.  buf points to an unsigned integer
  * holding the modem or line state.
  */
-#define GENSIO_EVENT_SER_MODEMSTATE	(SERGENIO_EVENT_BASE + 1)
-#define GENSIO_EVENT_SER_LINESTATE	(SERGENIO_EVENT_BASE + 2)
+#define GENSIO_EVENT_SER_MODEMSTATE	(SERGENSIO_EVENT_BASE + 1)
+#define GENSIO_EVENT_SER_LINESTATE	(SERGENSIO_EVENT_BASE + 2)
 
 /*
  * On the server side, these are for reporting that the client is
  * requesting the signature.  Not for client.
  */
-#define GENSIO_EVENT_SER_SIGNATURE	(SERGENIO_EVENT_BASE + 3)
+#define GENSIO_EVENT_SER_SIGNATURE	(SERGENSIO_EVENT_BASE + 3)
 
 /*
  * The remote end is asking the user to flow control or flush.  Client
  * or server.
  */
-#define GENSIO_EVENT_SER_FLOW_STATE	(SERGENIO_EVENT_BASE + 4)
-#define GENSIO_EVENT_SER_FLUSH		(SERGENIO_EVENT_BASE + 5)
+#define GENSIO_EVENT_SER_FLOW_STATE	(SERGENSIO_EVENT_BASE + 4)
+#define GENSIO_EVENT_SER_FLUSH		(SERGENSIO_EVENT_BASE + 5)
 
 /* Got a sync from the other end.  Client or server. */
-#define GENSIO_EVENT_SER_SYNC		(SERGENIO_EVENT_BASE + 6)
+#define GENSIO_EVENT_SER_SYNC		(SERGENSIO_EVENT_BASE + 6)
 
 /*
  * Server callbacks.  These only come in in server mode, you must
@@ -288,15 +288,15 @@ int sergensio_ri_b(struct sergensio_b *sbio, int *ri);
  * though the done callback is ignored in that case.  buf points to
  * an integer holding the value.
  */
-#define GENSIO_EVENT_SER_BAUD		(SERGENIO_EVENT_BASE + 7)
-#define GENSIO_EVENT_SER_DATASIZE	(SERGENIO_EVENT_BASE + 8)
-#define GENSIO_EVENT_SER_PARITY		(SERGENIO_EVENT_BASE + 9)
-#define GENSIO_EVENT_SER_STOPBITS	(SERGENIO_EVENT_BASE + 10)
-#define GENSIO_EVENT_SER_FLOWCONTROL	(SERGENIO_EVENT_BASE + 11)
-#define GENSIO_EVENT_SER_IFLOWCONTROL	(SERGENIO_EVENT_BASE + 12)
-#define GENSIO_EVENT_SER_SBREAK		(SERGENIO_EVENT_BASE + 13)
-#define GENSIO_EVENT_SER_DTR		(SERGENIO_EVENT_BASE + 14)
-#define GENSIO_EVENT_SER_RTS		(SERGENIO_EVENT_BASE + 15)
+#define GENSIO_EVENT_SER_BAUD		(SERGENSIO_EVENT_BASE + 7)
+#define GENSIO_EVENT_SER_DATASIZE	(SERGENSIO_EVENT_BASE + 8)
+#define GENSIO_EVENT_SER_PARITY		(SERGENSIO_EVENT_BASE + 9)
+#define GENSIO_EVENT_SER_STOPBITS	(SERGENSIO_EVENT_BASE + 10)
+#define GENSIO_EVENT_SER_FLOWCONTROL	(SERGENSIO_EVENT_BASE + 11)
+#define GENSIO_EVENT_SER_IFLOWCONTROL	(SERGENSIO_EVENT_BASE + 12)
+#define GENSIO_EVENT_SER_SBREAK		(SERGENSIO_EVENT_BASE + 13)
+#define GENSIO_EVENT_SER_DTR		(SERGENSIO_EVENT_BASE + 14)
+#define GENSIO_EVENT_SER_RTS		(SERGENSIO_EVENT_BASE + 15)
 
 GENSIO_DLL_PUBLIC
 bool sergensio_is_client(struct sergensio *sio);
