@@ -80,6 +80,8 @@ struct gensio {
     gensio_func func;
     void *gensio_data;
 
+    struct gensio_frdata *frdata;
+
     const char *typename;
 
     struct gensio *child;
@@ -172,6 +174,11 @@ gensio_data_free(struct gensio *io)
 {
     assert(gensio_list_empty(&io->waiters));
 
+    gensio_clear_sync(io);
+
+    if (io->frdata && io->frdata->freed)
+	io->frdata->freed(io, io->frdata);
+
     while (io->classes) {
 	struct gensio_classobj *c = io->classes;
 
@@ -198,12 +205,6 @@ gensio_get_cb(struct gensio *io)
     return io->cb;
 }
 
-void gensio_set_cb(struct gensio *io, gensio_event cb, void *user_data)
-{
-    io->cb = cb;
-    io->user_data = user_data;
-}
-
 int
 gensio_cb(struct gensio *io, int event, int err,
 	  unsigned char *buf, gensiods *buflen, const char *const *auxdata)
@@ -273,6 +274,8 @@ struct gensio_accepter {
     gensio_acc_func func;
     void *gensio_acc_data;
 
+    struct gensio_acc_frdata *frdata;
+
     const char *typename;
 
     struct gensio_accepter *child;
@@ -328,6 +331,9 @@ gensio_acc_data_alloc(struct gensio_os_funcs *o,
 void
 gensio_acc_data_free(struct gensio_accepter *acc)
 {
+    if (acc->frdata && acc->frdata->freed)
+	acc->frdata->freed(acc, acc->frdata);
+
     while (acc->classes) {
 	struct gensio_classobj *c = acc->classes;
 
@@ -374,6 +380,7 @@ gensio_acc_cb(struct gensio_accepter *acc, int event, void *data)
     }
     if (!acc->cb)
 	return GE_NOTSUP;
+
     return acc->cb(acc, acc->user_data, event, data);
 }
 
@@ -418,6 +425,19 @@ gensio_acc_remove_pending_gensio(struct gensio_accepter *acc,
     gensio_list_rm(&acc->pending_ios, &io->link);
 }
 
+void
+gensio_acc_set_frdata(struct gensio_accepter *acc,
+		      struct gensio_acc_frdata *frdata)
+{
+    acc->frdata = frdata;
+}
+
+struct gensio_acc_frdata *
+gensio_acc_get_frdata(struct gensio_accepter *acc)
+{
+    return acc->frdata;
+}
+
 int
 gensio_scan_args(struct gensio_os_funcs *o,
 		 const char **rstr, int *argc, const char ***args)
@@ -497,7 +517,8 @@ gensio_write(struct gensio *io, gensiods *count,
     struct gensio_sg sg;
 
     if (buflen == 0) {
-	*count = 0;
+	if (count)
+	    *count = 0;
 	return 0;
     }
     sg.buf = buf;
@@ -511,7 +532,8 @@ gensio_write_sg(struct gensio *io, gensiods *count,
 		const char *const *auxdata)
 {
     if (sglen == 0) {
-	*count = 0;
+	if (count)
+	    *count = 0;
 	return 0;
     }
     return io->func(io, GENSIO_FUNC_WRITE_SG, count, sg, sglen, NULL, auxdata);
@@ -535,7 +557,7 @@ gensio_raddr_to_str(struct gensio *io, gensiods *pos,
 	curlen = 0;
 	data = buf;
     }
-    rv = gensio_control(io, GENSIO_CONTROL_DEPTH_FIRST, true,
+    rv = gensio_control(io, GENSIO_CONTROL_DEPTH_FIRST, GENSIO_CONTROL_GET,
 			GENSIO_CONTROL_RADDR, data, &curlen);
     if (!rv)
 	*pos += curlen;
@@ -545,7 +567,7 @@ gensio_raddr_to_str(struct gensio *io, gensiods *pos,
 int
 gensio_get_raddr(struct gensio *io, void *addr, gensiods *addrlen)
 {
-    return gensio_control(io, GENSIO_CONTROL_DEPTH_FIRST, true,
+    return gensio_control(io, GENSIO_CONTROL_DEPTH_FIRST, GENSIO_CONTROL_GET,
 			  GENSIO_CONTROL_RADDR_BIN, addr, addrlen);
 }
 
@@ -851,6 +873,18 @@ gensio_set_is_encrypted(struct gensio *io, bool is_encrypted)
     io->is_encrypted = is_encrypted;
 }
 
+void
+gensio_set_frdata(struct gensio *io, struct gensio_frdata *frdata)
+{
+    io->frdata = frdata;
+}
+
+struct gensio_frdata *
+gensio_get_frdata(struct gensio *io)
+{
+    return io->frdata;
+}
+
 void
 gensio_set_attr_from_child(struct gensio *io, struct gensio *child)
 {
@@ -1188,12 +1222,8 @@ add_default_gensio_accepters(void *cb_data)
     REG_GENSIO_ACC(o, "tcp", str_to_tcp_gensio_accepter);
     REG_GENSIO_ACC(o, "udp", str_to_udp_gensio_accepter);
     REG_GENSIO_ACC(o, "sctp", str_to_sctp_gensio_accepter);
-#if HAVE_STDIO
     REG_GENSIO_ACC(o, "stdio", str_to_stdio_gensio_accepter);
-#endif
-#if HAVE_UNIX
     REG_GENSIO_ACC(o, "unix", str_to_unix_gensio_accepter);
-#endif
     REG_FILT_GENSIO_ACC(o, "ssl", str_to_ssl_gensio_accepter,
 			ssl_gensio_accepter_alloc);
     REG_FILT_GENSIO_ACC(o, "mux", str_to_mux_gensio_accepter,
@@ -1401,23 +1431,15 @@ add_default_gensios(void *cb_data)
     REG_GENSIO(o, "tcp", str_to_tcp_gensio);
     REG_GENSIO(o, "udp", str_to_udp_gensio);
     REG_GENSIO(o, "sctp", str_to_sctp_gensio);
-#if HAVE_UNIX
     REG_GENSIO(o, "unix", str_to_unix_gensio);
-#endif
-#if HAVE_STDIO
     REG_GENSIO(o, "stdio", str_to_stdio_gensio);
-#endif
-#if HAVE_PTY
     REG_GENSIO(o, "pty", str_to_pty_gensio);
-#endif
     REG_FILT_GENSIO(o, "ssl", str_to_ssl_gensio, ssl_gensio_alloc);
     REG_FILT_GENSIO(o, "mux", str_to_mux_gensio, mux_gensio_alloc);
     REG_FILT_GENSIO(o, "certauth", str_to_certauth_gensio,
 		    certauth_gensio_alloc);
     REG_FILT_GENSIO(o, "telnet", str_to_telnet_gensio, telnet_gensio_alloc);
-#if HAVE_SERIALDEV
     REG_GENSIO(o, "serialdev", str_to_serialdev_gensio);
-#endif
     REG_GENSIO(o, "echo", str_to_echo_gensio);
     REG_GENSIO(o, "file", str_to_file_gensio);
     REG_GENSIO(o, "ipmisol", str_to_ipmisol_gensio);
@@ -1502,7 +1524,13 @@ str_to_gensio(const char *str,
 	return err;
     }
 
-#if HAVE_SERIALDEV
+#if _WIN32
+    if (strncmp(str, "COM", 3) == 0) {
+	err = str_to_serialdev_gensio(str, NULL, o, cb, user_data,
+				      gensio);
+	goto out;
+    }
+#else
     if (*str == '/') {
 	err = str_to_serialdev_gensio(str, NULL, o, cb, user_data,
 				      gensio);
@@ -1958,6 +1986,9 @@ struct gensio_def_entry builtin_defaults[] = {
     { "tcpd",		GENSIO_DEFAULT_ENUM,	.enums = tcpd_enums,
 						.def.intval = GENSIO_TCPD_ON },
 #endif
+    /* UDP only */
+    { "mttl",		GENSIO_DEFAULT_INT,	.min = 1, .max = 255,
+						.def.intval = 1 },
     /* SCTP only */
     { "instreams",	GENSIO_DEFAULT_INT,	.min = 1, .max = INT_MAX,
 						.def.intval = 1 },
@@ -1976,6 +2007,11 @@ struct gensio_def_entry builtin_defaults[] = {
     { "hangup_when_done", GENSIO_DEFAULT_BOOL,	.def.intval = 0 },
     { "custspeed",	GENSIO_DEFAULT_BOOL,	.def.intval = 0 },
     { "rs485",		GENSIO_DEFAULT_STR,	.def.strval = NULL },
+    { "nouucplock",	GENSIO_DEFAULT_BOOL,	.def.intval = 0 },
+    { "drain_time",	GENSIO_DEFAULT_INT,	.min = -1, .max = INT_MAX,
+						.def.intval = -1, },
+    { "char_drain_wait",GENSIO_DEFAULT_INT,	.min = -1, .max = INT_MAX,
+						.def.intval = 50, },
     /* serialdev and SOL */
     { "speed",		GENSIO_DEFAULT_STR,	.def.strval = "9600N81" },
     { "nobreak",	GENSIO_DEFAULT_BOOL,	.def.intval = 0 },
@@ -2477,7 +2513,7 @@ gensio_get_default(struct gensio_os_funcs *o,
 		err = GE_NOMEM;
 		goto out_unlock;
 	    }
-	    memcpy(str, val->strval, val->intval + 1); /* copy terminating \0 */
+	    memcpy(str, val->strval, val->intval + 1L); /* copy emding \0 */
 	    *strval = str;
 	    *intval = val->intval;
 	} else {
@@ -2700,6 +2736,7 @@ gensio_wait_no_cb(struct gensio *io, struct gensio_waiter *waiter,
     struct gensio_nocbwait wait;
     int rv = 0;
 
+    memset(&wait, 0, sizeof(wait));
     wait.waiter = waiter;
     o->lock(io->lock);
     if (io->cb_count != 0) {
@@ -3082,7 +3119,7 @@ i_gensio_acc_accept_s(struct gensio_accepter *acc, gensio_time *timeout,
     struct gensio_os_funcs *o = acc->o;
     struct gensio_waiting_accept wa;
     struct gensio_link *l;
-    int rv;
+    int rv = 0;
 
     memset(&wa, 0, sizeof(wa));
     wa.waiter = o->alloc_waiter(o);
@@ -3107,9 +3144,8 @@ i_gensio_acc_accept_s(struct gensio_accepter *acc, gensio_time *timeout,
 	gensio_list_rm(&acc->waiting_accepts, &wa.link);
     } else if (gensio_list_empty(&acc->waiting_ios)) {
 	rv = GE_LOCALCLOSED;
-    } else {
+    } else if (!rv) {
     got_one:
-	rv = 0;
 	l = gensio_list_first(&acc->waiting_ios);
 	gensio_list_rm(&acc->waiting_ios, l);
 	*new_io = gensio_container_of(l, struct gensio, link);
@@ -3266,7 +3302,7 @@ gensio_alloc_vsprintf(struct gensio_os_funcs *o, const char *fmt, va_list va)
     char c[1], *str;
 
     va_copy(va2, va);
-    len = vsnprintf(c, 0, fmt, va) + 1;
+    len = vsnprintf(c, 0, fmt, va) + 1L;
     str = o->zalloc(o, len);
     if (str)
 	vsnprintf(str, len, fmt, va2);
@@ -3288,6 +3324,36 @@ gensio_alloc_sprintf(struct gensio_os_funcs *o, const char *fmt, ...)
     return s;
 }
 
+char *
+gensio_quote_string(struct gensio_os_funcs *o, const char *str)
+{
+    const char *ic;
+    char *ostr, *oc;
+    gensiods count = 3; /* Space for two quotes and a \0. */
+
+    /* We need two characters for all \ and ". */
+    for (ic = str; *ic; ic++) {
+	count++;
+	if (*ic == '\\' || *ic == '"')
+	    count++;
+    }
+
+    ostr = o->zalloc(o, count);
+    if (!ostr)
+	return NULL;
+
+    oc = ostr;
+    *oc++ = '"';
+    for (ic = str; *ic; ic++) {
+	if (*ic == '\\' || *ic == '"')
+	    *oc++ = '\\';
+	*oc++ = *ic;
+    }
+    *oc++ = '"';
+
+    return ostr;
+}
+
 static const char *gensio_errs[] = {
     /*   0 */    "No error",
     /*   1 */    "Out of memory",
@@ -3325,7 +3391,8 @@ static const char *gensio_errs[] = {
     /*  33 */    "Operation was interrupted by a signal",
     /*  34 */    "Operation on shutdown fd",
     /*  35 */    "Local end closed connection",
-    /*  36 */    "Permission denied"
+    /*  36 */    "Permission denied",
+    /*  37 */    "Application error"
 };
 const unsigned int errno_len = sizeof(gensio_errs) / sizeof(char *);
 
@@ -3336,3 +3403,92 @@ gensio_err_to_str(int err)
 	return "Unknown error";
     return gensio_errs[err];
 }
+
+int
+gensio_addr_create(struct gensio_os_funcs *o,
+		   int nettype, const void *iaddr, gensiods len,
+		   unsigned int port, struct gensio_addr **newaddr)
+{
+    return o->addr_create(o, nettype, iaddr, len, port, newaddr);
+}
+
+bool
+gensio_addr_equal(const struct gensio_addr *a1,
+		  const struct gensio_addr *a2,
+		  bool compare_ports, bool compare_all)
+{
+    return a1->o->addr_equal(a1, a2, compare_ports, compare_all);
+}
+
+int
+gensio_addr_to_str(const struct gensio_addr *addr,
+		   char *buf, gensiods *pos, gensiods buflen)
+{
+    return addr->o->addr_to_str(addr, buf, pos, buflen);
+}
+
+int
+gensio_addr_to_str_all(const struct gensio_addr *addr,
+		       char *buf, gensiods *pos, gensiods buflen)
+{
+    return addr->o->addr_to_str_all(addr, buf, pos, buflen);
+}
+
+struct gensio_addr *
+gensio_addr_dup(const struct gensio_addr *iaddr)
+{
+    return iaddr->o->addr_dup(iaddr);
+}
+
+struct gensio_addr *
+gensio_addr_cat(const struct gensio_addr *addr1,
+		const struct gensio_addr *addr2)
+{
+    return addr1->o->addr_cat(addr1, addr2);
+}
+
+bool
+gensio_addr_addr_present(const struct gensio_addr *gai,
+			 const void *addr, gensiods addrlen,
+			 bool compare_ports)
+{
+    return gai->o->addr_addr_present(gai, addr, addrlen, compare_ports);
+}
+
+void
+gensio_addr_free(struct gensio_addr *addr)
+{
+    addr->o->addr_free(addr);
+}
+
+bool
+gensio_addr_next(struct gensio_addr *addr)
+{
+    return addr->o->addr_next(addr);
+}
+
+void
+gensio_addr_rewind(struct gensio_addr *addr)
+{
+    addr->o->addr_rewind(addr);
+}
+
+int
+gensio_addr_get_nettype(const struct gensio_addr *addr)
+{
+    return addr->o->addr_get_nettype(addr);
+}
+
+bool
+gensio_addr_family_supports(const struct gensio_addr *addr,
+			    int family, int flags)
+{
+    return addr->o->addr_family_supports(addr, family, flags);
+}
+
+void
+gensio_addr_getaddr(const struct gensio_addr *addr,
+		    void *oaddr, gensiods *rlen)
+{
+    addr->o->addr_getaddr(addr, oaddr, rlen);
+}
