@@ -17,7 +17,7 @@
 
 #ifdef DEBUG_DATA
 #define ENABLE_PRBUF 1
-#include <utils/utils.h>
+#include "utils.h"
 #endif
 
 /*
@@ -176,6 +176,7 @@ struct basen_data {
     bool xmit_enabled;
     bool in_xmit_ready;
     bool redo_xmit_ready;
+    bool ll_can_write;
 
     /*
      * We got an error from the lower layer, it's probably not working
@@ -220,8 +221,10 @@ struct gensio_filter {
     void *user_data;
 };
 
-static void handle_ioerr(struct basen_data *ndata, int err);
+static void i_handle_ioerr(struct basen_data *ndata, int err, int line);
 static void basen_filter_try_close(struct basen_data *ndata, bool was_timeout);
+static void basen_check_open_close_ops(struct basen_data *ndata);
+#define handle_ioerr(n, e) i_handle_ioerr(n, e, __LINE__)
 
 static void
 i_basen_lock(struct basen_data *ndata)
@@ -312,6 +315,13 @@ basen_start_timer(struct basen_data *ndata, gensio_time *timeout)
 	basen_ref(ndata);
 }
 
+static void
+basen_stop_timer(struct basen_data *ndata)
+{
+    if (ndata->o->stop_timer(ndata->timer) == 0)
+	basen_deref(ndata);
+}
+
 #ifdef DEBUG_STATE
 static void
 i_basen_add_trace(struct basen_data *ndata,
@@ -371,7 +381,7 @@ static void
 filter_io_err(struct basen_data *ndata, int err)
 {
     if (ndata->filter)
-	return gensio_filter_io_err(ndata->filter, err);
+	gensio_filter_io_err(ndata->filter, err);
 }
 
 static bool
@@ -379,7 +389,7 @@ filter_ll_can_write(struct basen_data *ndata)
 {
     if (ndata->filter)
 	return gensio_filter_ll_can_write(ndata->filter);
-    return true;
+    return ndata->ll_can_write;
 }
 
 static bool
@@ -464,11 +474,21 @@ static int
 ll_write(struct basen_data *ndata, gensiods *rcount,
 	 const struct gensio_sg *sg, gensiods sglen, const char *const *auxdata)
 {
+    int rv;
+
 #ifdef DEBUG_DATA
     printf("LL write:");
-    prbuf(buf, buflen);
+    do {
+	unsigned int i;
+	for (i = 0; i < sglen; i++)
+	    prbuf(sg[i].buf, sg[i].buflen);
+    } while (false);
 #endif
-    return gensio_ll_write(ndata->ll, rcount, sg, sglen, auxdata);
+    rv = gensio_ll_write(ndata->ll, rcount, sg, sglen, auxdata);
+#ifdef DEBUG_DATA
+    printf("LL write returned %d accepted %ld\n", rv, rcount ? *rcount : 0);
+#endif
+    return rv;
 }
 
 /*
@@ -492,6 +512,7 @@ basen_ll_close_done(void *cb_data, void *close_data)
     i_basen_add_trace(ndata, 1100, __LINE__);
     switch(ndata->state) {
     case BASEN_IN_LL_CLOSE:
+    case BASEN_IO_ERR_CLOSE:
 	/* Don't move to BASEN_CLOSED until later to avoid races. */
 	i_basen_add_trace(ndata, 102, __LINE__);
 	ndata->deferred_close = true;
@@ -500,6 +521,18 @@ basen_ll_close_done(void *cb_data, void *close_data)
 
     case BASEN_IN_LL_IO_ERR_CLOSE:
 	basen_set_state(ndata, BASEN_IO_ERR_CLOSE);
+	/*
+	 * This is kind of a hack.  This can come from an error two
+	 * places, either filter open or open.  If it's in filter
+	 * open, we need to deliver the open failure to the user,
+	 * otherwise we will already be set to deliver a read/write
+	 * failture.  So if open_done is set, we are in filter open,
+	 * and know to deliver the open failure.
+	 */
+	if (ndata->open_done) {
+	    ndata->deferred_open = true;
+	    basen_sched_deferred_op(ndata);
+	}
 	break;
 
     default:
@@ -509,10 +542,14 @@ basen_ll_close_done(void *cb_data, void *close_data)
 }
 
 static int
-ll_close(struct basen_data *ndata)
+i_ll_close(struct basen_data *ndata, int line)
 {
-    return gensio_ll_close(ndata->ll, basen_ll_close_done, ndata);
+    int rv;
+
+    rv = gensio_ll_close(ndata->ll, basen_ll_close_done, ndata);
+    return rv;
 }
+#define ll_close(n) i_ll_close(n, __LINE__);
 
 static void
 ll_set_read_callback_enable(struct basen_data *ndata, bool enable)
@@ -537,14 +574,7 @@ basen_set_ll_enables(struct basen_data *ndata)
 	return;
     }
 
-    if (ndata->xmit_enabled && ndata->ll_err) {
-	ndata->deferred_write = true;
-	basen_sched_deferred_op(ndata);
-    } else if (filter_ll_write_pending(ndata) ||
-	(filter_ll_can_write(ndata) && ndata->xmit_enabled))
-	ll_set_write_callback_enable(ndata, true);
-    else
-	ll_set_write_callback_enable(ndata, false);
+    ll_set_write_callback_enable(ndata, !ndata->ll_can_write);
 
     enabled = false;
     if (ndata->in_read)
@@ -569,6 +599,10 @@ basen_set_ll_enables(struct basen_data *ndata)
 	enabled = enabled || filter_ll_read_needed(ndata);
 	break;
 
+    case BASEN_IN_LL_CLOSE:
+	enabled = false;
+	break;
+
     default:
 	enabled = true;
 	break;
@@ -583,8 +617,37 @@ basen_write_data_handler(void *cb_data, gensiods *rcount,
 			 const char *const *auxdata)
 {
     struct basen_data *ndata = cb_data;
+    int rv;
+    gensiods i, total = 0, count = 0;
 
-    return ll_write(ndata, rcount, sg, sglen, auxdata);
+    for (i = 0; i < sglen; i++)
+	total += sg[i].buflen;
+    rv = ll_write(ndata, &count, sg, sglen, auxdata);
+    if (!rv && count < total)
+	ndata->ll_can_write = false;
+    if (rcount)
+	*rcount = count;
+    return rv;
+}
+
+static int
+basen_filter_ul_push(struct basen_data *ndata, bool check_open_close)
+{
+    if (!ndata->ll_err && ndata->ll_can_write &&
+		filter_ll_write_pending(ndata)) {
+	int err;
+
+	err = filter_ul_write(ndata, basen_write_data_handler, NULL, NULL, 0,
+			      NULL);
+	if (err) {
+	    handle_ioerr(ndata, err);
+	    return err;
+	}
+	if (check_open_close)
+	    basen_check_open_close_ops(ndata);
+    }
+
+    return 0;
 }
 
 static bool
@@ -625,19 +688,29 @@ basen_write(struct basen_data *ndata, gensiods *rcount,
      * to here.
      */
     if (ndata->in_write_count == 0 && ndata->ll_want_close) {
-	int rv = ll_close(ndata);
-	if (rv) {
-	    if (ndata->state == BASEN_CLOSE_WAIT_DRAIN) {
-		basen_set_state(ndata, BASEN_IN_FILTER_CLOSE);
-		basen_filter_try_close(ndata, false);
-	    } else if (ndata->state == BASEN_IN_LL_IO_ERR_CLOSE) {
-		basen_set_state(ndata, BASEN_IO_ERR_CLOSE);
-	    } else if (ndata->state == BASEN_IN_LL_CLOSE) {
+	int rv;
+
+	switch (ndata->state) {
+	case BASEN_CLOSE_WAIT_DRAIN:
+	    basen_set_state(ndata, BASEN_IN_LL_CLOSE);
+	    rv = ll_close(ndata);
+	    if (rv) {
 		ndata->deferred_close = true;
 		basen_sched_deferred_op(ndata);
-	    } else {
-		assert(0);
 	    }
+	    break;
+
+	case BASEN_IN_LL_IO_ERR_CLOSE:
+	    basen_set_state(ndata, BASEN_IO_ERR_CLOSE);
+	    break;
+
+	case BASEN_IN_LL_CLOSE:
+	    ndata->deferred_close = true;
+	    basen_sched_deferred_op(ndata);
+	    break;
+
+	default:
+	    assert(0);
 	}
     }
  out_unlock:
@@ -711,10 +784,11 @@ basen_read_data_handler(void *cb_data,
 }
 
 static void
-handle_ioerr(struct basen_data *ndata, int err)
+i_handle_ioerr(struct basen_data *ndata, int err, int line)
 {
     int rv;
 
+    i_basen_add_trace(ndata, 905, line);
     assert(err);
 
     if (ndata->ll_err)
@@ -726,6 +800,9 @@ handle_ioerr(struct basen_data *ndata, int err)
     ndata->ll_err = err;
     ndata->open_err = err;
 
+    /* Strange looking, but don't enable ll write if we get an error. */
+    ndata->ll_can_write = true;
+
     switch(ndata->state) {
     case BASEN_CLOSED:
     case BASEN_IN_LL_OPEN:
@@ -735,8 +812,6 @@ handle_ioerr(struct basen_data *ndata, int err)
 
     case BASEN_IN_FILTER_OPEN:
 	filter_io_err(ndata, err);
-	ndata->deferred_open = true;
-	basen_sched_deferred_op(ndata);
 	basen_set_state(ndata, BASEN_IN_LL_IO_ERR_CLOSE);
 	/*
 	 * No need to check for in_write_count here, it can't be
@@ -850,6 +925,9 @@ basen_finish_close(struct basen_data *ndata)
 static void
 basen_finish_open(struct basen_data *ndata, int err)
 {
+    gensio_done_err open_done;
+    void *open_data;
+
     i_basen_add_trace(ndata, 100, __LINE__);
     if (!err) {
 	assert(ndata->state == BASEN_IN_FILTER_OPEN || ndata->state == BASEN_OPEN);
@@ -858,8 +936,11 @@ basen_finish_open(struct basen_data *ndata, int err)
 	    basen_start_timer(ndata, &ndata->pending_timer);
     }
 
+    open_done = ndata->open_done;
+    ndata->open_done = NULL;
+    open_data = ndata->open_data;
     basen_unlock(ndata);
-    ndata->open_done(ndata->io, err, ndata->open_data);
+    open_done(ndata->io, err, open_data);
     basen_lock(ndata);
 }
 
@@ -875,6 +956,7 @@ basen_in_read_callbackable_state(struct basen_data *ndata)
 	    ndata->state == BASEN_CLOSE_WAIT_DRAIN ||
 	    ndata->state == BASEN_IN_LL_IO_ERR_CLOSE ||
 	    ndata->state == BASEN_IN_LL_CLOSE ||
+	    ndata->state == BASEN_IN_FILTER_CLOSE ||
 	    ndata->state == BASEN_IO_ERR_CLOSE);
 }
 
@@ -894,6 +976,8 @@ basen_deferred_op(struct gensio_runner *runner, void *cbdata)
     int err;
 
     basen_lock(ndata);
+    ndata->deferred_op_pending = false;
+
     if (ndata->deferred_open) {
 	ndata->deferred_open = false;
 	i_basen_add_trace(ndata, 100, __LINE__);
@@ -931,24 +1015,21 @@ basen_deferred_op(struct gensio_runner *runner, void *cbdata)
  skip_read:
     /* The write side is primarily for delivery when an error occurs. */
     while (ndata->deferred_write) {
+	ndata->deferred_write = false;
 	if (ndata->in_xmit_ready)
 	    goto skip_write;
-	ndata->deferred_write = false;
-	while (basen_in_write_callbackable_state(ndata) && ndata->ll_err
+	ndata->in_xmit_ready = true;
+	while (basen_in_write_callbackable_state(ndata) &&
+	       (filter_ll_can_write(ndata) || ndata->ll_err)
 	       && ndata->xmit_enabled) {
 	    basen_unlock(ndata);
 	    gensio_cb(ndata->io, GENSIO_EVENT_WRITE_READY, 0, NULL, 0, NULL);
 	    basen_lock(ndata);
 	}
+	ndata->in_xmit_ready = false;
     }
 
  skip_write:
-    /*
-     * Close callback may call open, if open defers make sure we
-     * run the runner.
-     */
-    ndata->deferred_op_pending = false;
-
     if (ndata->deferred_close) {
 	if (!(ndata->in_xmit_ready || ndata->in_read)) {
 	    ndata->deferred_close = false;
@@ -957,8 +1038,10 @@ basen_deferred_op(struct gensio_runner *runner, void *cbdata)
 	}
     }
 
-    if (ndata->state != BASEN_CLOSED)
+    if (ndata->state != BASEN_CLOSED) {
+	basen_filter_ul_push(ndata, true);
 	basen_set_ll_enables(ndata);
+    }
     basen_deref_and_unlock(ndata); /* Ref from basen_sched_deferred_op */
 }
 
@@ -979,8 +1062,16 @@ basen_filter_try_connect(struct basen_data *ndata, bool was_timeout)
     gensio_time timeout = {0, 0};
 
     err = filter_try_connect(ndata, &timeout, was_timeout);
-    if (!err || err == GE_INPROGRESS || err == GE_RETRY)
-	basen_set_ll_enables(ndata);
+    if (!err || err == GE_INPROGRESS || err == GE_RETRY) {
+	int err2 = basen_filter_ul_push(ndata, false);
+	if (err2) {
+	    basen_set_ll_enables(ndata);
+	    return err2;
+	} else if (err == GE_INPROGRESS) {
+	    err = filter_try_connect(ndata, &timeout, false);
+	    basen_set_ll_enables(ndata);
+	}
+    }
     if (err == GE_INPROGRESS)
 	return GE_INPROGRESS;
     if (err == GE_RETRY) {
@@ -1047,6 +1138,9 @@ basen_open(struct basen_data *ndata, gensio_done_err open_done, void *open_data)
 {
     int err = GE_INUSE;
 
+    if (!open_done)
+	return GE_INVAL;
+
     basen_lock(ndata);
     if (ndata->state == BASEN_CLOSED) {
 	err = filter_setup(ndata);
@@ -1062,6 +1156,7 @@ basen_open(struct basen_data *ndata, gensio_done_err open_done, void *open_data)
 	ndata->deferred_close = false;
 	ndata->read_enabled = false;
 	ndata->xmit_enabled = false;
+	ndata->ll_can_write = true;
 	ndata->timer_start_pending = false;
 	ndata->close_requested = false;
 	ndata->ll_want_close = false;
@@ -1101,6 +1196,9 @@ basen_open_nochild(struct basen_data *ndata,
 {
     int err = GE_INUSE;
 
+    if (!open_done)
+	return GE_INVAL;
+
     basen_lock(ndata);
     if (ndata->state == BASEN_CLOSED) {
 	err = filter_setup(ndata);
@@ -1135,6 +1233,7 @@ basen_open_nochild(struct basen_data *ndata,
 	    basen_set_state(ndata, BASEN_CLOSED);
 	else
 	    basen_ref(ndata); /* Ref for open */
+
 	basen_set_ll_enables(ndata);
     }
  out_err:
@@ -1149,10 +1248,13 @@ basen_filter_try_close(struct basen_data *ndata, bool was_timeout)
     int err;
     gensio_time timeout = {0, 0};
 
-
     err = filter_try_disconnect(ndata, &timeout, was_timeout);
-    if (err == GE_INPROGRESS || err == GE_RETRY)
+    if (err == GE_INPROGRESS || err == GE_RETRY) {
+	basen_filter_ul_push(ndata, false);
+	if (err == GE_INPROGRESS)
+	    err = filter_try_disconnect(ndata, &timeout, was_timeout);
 	basen_set_ll_enables(ndata);
+    }
     if (err == GE_INPROGRESS)
 	return;
     if (err == GE_RETRY) {
@@ -1191,8 +1293,6 @@ basen_i_close(struct basen_data *ndata,
     if (ndata->state == BASEN_IN_LL_OPEN ||
 		ndata->state == BASEN_IN_FILTER_OPEN) {
 	basen_set_state(ndata, BASEN_IN_LL_CLOSE);
-	ndata->deferred_open = true;
-	basen_sched_deferred_op(ndata);
 	/*
 	 * No need to check for in_write_count here, it can't be
 	 * pending because we can't have started a write yet.
@@ -1326,6 +1426,7 @@ basen_timeout(struct gensio_timer *timer, void *cb_data)
     default:
 	break;
     }
+    basen_filter_ul_push(ndata, true);
     basen_set_ll_enables(ndata);
     basen_deref_and_unlock(ndata);
 }
@@ -1373,12 +1474,10 @@ basen_set_write_callback_enable(struct basen_data *ndata, bool enabled)
     if (!basen_in_write_callbackable_state(ndata))
 	goto out_unlock;
     ndata->xmit_enabled = enabled;
-    if (enabled && ndata->ll_err && ndata->state != BASEN_OPEN) {
-	/* We got an error from the lower layer, just deliver it. */
+    if (enabled && (filter_ll_can_write(ndata) || ndata->ll_err)) {
+	/* We can write, schedule the callback as a deferred op. */
 	ndata->deferred_write = true;
 	basen_sched_deferred_op(ndata);
-    } else {
-	basen_set_ll_enables(ndata);
     }
  out_unlock:
     basen_unlock(ndata);
@@ -1397,13 +1496,13 @@ gensio_base_func(struct gensio *io, int func, gensiods *count,
 	return basen_write(ndata, count, cbuf, buflen, auxdata);
 
     case GENSIO_FUNC_OPEN:
-	return basen_open(ndata, cbuf, buf);
+	return basen_open(ndata, (void *) cbuf, buf);
 
     case GENSIO_FUNC_OPEN_NOCHILD:
-	return basen_open_nochild(ndata, cbuf, buf);
+	return basen_open_nochild(ndata, (void *) cbuf, buf);
 
     case GENSIO_FUNC_CLOSE:
-	return basen_close(ndata, cbuf, buf);
+	return basen_close(ndata, (void *) cbuf, buf);
 
     case GENSIO_FUNC_FREE:
 	basen_free(ndata);
@@ -1488,7 +1587,6 @@ basen_ll_read(void *cb_data, int readerr,
     prbuf(buf, buflen);
 #endif
     basen_lock_and_ref(ndata);
-    ll_set_read_callback_enable(ndata, false);
     if (readerr) {
 	handle_ioerr(ndata, readerr);
 	goto out_finish;
@@ -1499,14 +1597,14 @@ basen_ll_read(void *cb_data, int readerr,
 	goto out_finish;
     }
 
-    if (ndata->in_read)
+    if (ndata->in_read) {
 	/* Currently in a deferred read, just let that handle it. */
+	ll_set_read_callback_enable(ndata, false);
 	goto out_unlock;
+    }
 
     while (buflen > 0 &&
-	   (ndata->read_enabled || filter_ll_read_needed(ndata) ||
-	    ndata->state == BASEN_IN_LL_CLOSE ||
-	    ndata->state == BASEN_CLOSE_WAIT_DRAIN)) {
+	   (ndata->read_enabled || filter_ll_read_needed(ndata))) {
 	ndata->in_read = true;
 	do {
 	    gensiods wrlen = 0;
@@ -1543,7 +1641,17 @@ basen_ll_read(void *cb_data, int readerr,
 	} while (ndata->read_enabled && buflen > 0);
 	ndata->in_read = false;
 
+	basen_filter_ul_push(ndata, true);
+
 	basen_check_open_close_ops(ndata);
+
+	while (basen_in_write_callbackable_state(ndata) &&
+	       (filter_ll_can_write(ndata) || ndata->ll_err)
+	       && ndata->xmit_enabled) {
+	    basen_unlock(ndata);
+	    gensio_cb(ndata->io, GENSIO_EVENT_WRITE_READY, 0, NULL, 0, NULL);
+	    basen_lock(ndata);
+	}
     }
 
  out_finish:
@@ -1551,6 +1659,9 @@ basen_ll_read(void *cb_data, int readerr,
  out_unlock:
     basen_deref_and_unlock(ndata);
 
+#ifdef DEBUG_DATA
+    printf("LL read returns %ld\n", buf - ibuf);
+#endif
     return buf - ibuf;
 }
 
@@ -1576,9 +1687,9 @@ basen_ll_write_ready(void *cb_data)
 	ndata->redo_xmit_ready = true;
 	goto out;
     }
+    ndata->ll_can_write = true;
     ndata->in_xmit_ready = true;
  retry:
-    ll_set_write_callback_enable(ndata, false);
     if (filter_ll_write_pending(ndata)) {
 	err = filter_ul_write(ndata, basen_write_data_handler, NULL, NULL, 0,
 			      NULL);
@@ -1588,10 +1699,12 @@ basen_ll_write_ready(void *cb_data)
 	}
     }
 
+    basen_filter_ul_push(ndata, true);
+
     basen_check_open_close_ops(ndata);
 
-    if (basen_in_read_callbackable_state(ndata) &&
-	   (!filter_ll_write_pending(ndata) || ndata->ll_err)
+    while (basen_in_write_callbackable_state(ndata) &&
+	   (filter_ll_can_write(ndata) || ndata->ll_err)
 	   && ndata->xmit_enabled) {
 	basen_unlock(ndata);
 	gensio_cb(ndata->io, GENSIO_EVENT_WRITE_READY, 0, NULL, 0, NULL);
@@ -1654,6 +1767,16 @@ basen_start_timer_op(void *cb_data, gensio_time *timeout)
     }
 }
 
+static void
+basen_stop_timer_op(void *cb_data)
+{
+    struct basen_data *ndata = cb_data;
+
+    if (ndata->state == BASEN_OPEN || ndata->state == BASEN_CLOSE_WAIT_DRAIN) {
+	basen_stop_timer(ndata);
+    }
+}
+
 static int
 gensio_base_filter_cb(void *cb_data, int op, void *data)
 {
@@ -1666,6 +1789,10 @@ gensio_base_filter_cb(void *cb_data, int op, void *data)
 	basen_start_timer_op(cb_data, data);
 	return 0;
 
+    case GENSIO_FILTER_CB_STOP_TIMER:
+	basen_stop_timer_op(cb_data);
+	return 0;
+
     default:
 	return GE_NOTSUP;
     }
@@ -1731,6 +1858,10 @@ gensio_i_alloc(struct gensio_os_funcs *o,
 	    goto out_nomem;
 	}
 
+	/* This is not ideal, it should really return GE_NOMEM. */
+	if (!open_done)
+	    goto out_nomem;
+
 	ndata->open_done = open_done;
 	ndata->open_data = open_data;
     }
@@ -1784,12 +1915,15 @@ base_gensio_server_start(struct gensio *io)
     basen_set_state(ndata, BASEN_IN_FILTER_OPEN);
     err = basen_filter_try_connect(ndata, false);
     if (!err) {
-	/* We are fully open, schedule it. */
-	basen_set_state(ndata, BASEN_OPEN);
-	ndata->deferred_open = true;
-	basen_sched_deferred_op(ndata);
+	err = basen_filter_ul_push(ndata, true);
+	if (!err) {
+	    /* We are fully open, schedule it. */
+	    basen_set_state(ndata, BASEN_OPEN);
+	    ndata->deferred_open = true;
+	    basen_sched_deferred_op(ndata);
+	}
     } else if (err == GE_INPROGRESS) {
-	err = 0;
+	err = basen_filter_ul_push(ndata, true);
     } else {
 	basen_set_state(ndata, BASEN_CLOSED);
 	err = GE_NOMEM;
@@ -1797,6 +1931,7 @@ base_gensio_server_start(struct gensio *io)
     }
     basen_ref(ndata); /* For the open. */
     basen_set_ll_enables(ndata);
+
  out_unlock:
     basen_unlock(ndata);
 
@@ -1830,10 +1965,13 @@ bool
 gensio_filter_ll_can_write(struct gensio_filter *filter)
 {
     bool val = true;
+    int err;
 
     /* If not implemented, this will just be ignored. */
-    filter->func(filter, GENSIO_FILTER_FUNC_LL_CAN_WRITE,
-		 NULL, &val, NULL, NULL, NULL, 0, NULL);
+    err = filter->func(filter, GENSIO_FILTER_FUNC_LL_CAN_WRITE,
+		       NULL, &val, NULL, NULL, NULL, 0, NULL);
+    if (err)
+	return filter->ndata->ll_can_write;
     return val;
 }
 
