@@ -131,7 +131,7 @@ f_open(struct gensio_os_funcs *o,
        const char *fn, int flags, int mode, FILE **rf)
 {
     char *fmode;
-    struct FILE *f;
+    FILE *f;
 
     if (flags & F_O_RDONLY) {
 	fmode = "r";
@@ -141,7 +141,7 @@ f_open(struct gensio_os_funcs *o,
 	else
 	    fmode = "r+";
     } else {
-	return NULL;
+	return GE_INVAL;
     }
     f = fopen(fn, fmode);
     if (!f)
@@ -315,7 +315,7 @@ filen_deferred_op(struct gensio_runner *runner, void *cb_data)
 
     while (ndata->state == FILEN_OPEN &&
 	   (f_ready(ndata->inf) || ndata->read_err) && ndata->read_enabled) {
-	gensiods count;
+	gensiods count = 0;
 
 	if (ndata->data_pending_len == 0 && !ndata->read_err) {
 	    int rv = f_read(ndata->o, ndata->inf, ndata->read_data,
@@ -543,10 +543,10 @@ gensio_file_func(struct gensio *io, int func, gensiods *count,
 	return filen_write(io, count, cbuf, buflen);
 
     case GENSIO_FUNC_OPEN:
-	return filen_open(io, cbuf, buf);
+	return filen_open(io, (void *) cbuf, buf);
 
     case GENSIO_FUNC_CLOSE:
-	return filen_close(io, cbuf, buf);
+	return filen_close(io, (void *) cbuf, buf);
 
     case GENSIO_FUNC_FREE:
 	filen_free(io);
@@ -640,7 +640,10 @@ file_ndata_setup(struct gensio_os_funcs *o, struct file_ndata_data *data,
 static int
 process_file_args(const char * const args[], struct file_ndata_data *data)
 {
-    unsigned int umode = 6, gmode = 6, omode = 6, i, mode;
+#if !USE_FILE_STDIO
+    unsigned int mode;
+#endif
+    unsigned int umode = 6, gmode = 6, omode = 6, i;
 
     memset(data, 0, sizeof(*data));
     data->read_close = true;
@@ -678,7 +681,7 @@ process_file_args(const char * const args[], struct file_ndata_data *data)
 }
 
 int
-file_gensio_alloc(const char * const argv[], const char * const args[],
+file_gensio_alloc(const char * const args[],
 		  struct gensio_os_funcs *o,
 		  gensio_event cb, void *user_data,
 		  struct gensio **new_gensio)
@@ -717,14 +720,5 @@ str_to_file_gensio(const char *str, const char * const args[],
 		   gensio_event cb, void *user_data,
 		   struct gensio **new_gensio)
 {
-    int err;
-    const char **argv;
-
-    err = gensio_str_to_argv(o, str, NULL, &argv, NULL);
-    if (!err) {
-	err = file_gensio_alloc(argv, args, o, cb, user_data, new_gensio);
-	gensio_argv_free(o, argv);
-    }
-
-    return err;
+    return file_gensio_alloc(args, o, cb, user_data, new_gensio);
 }
