@@ -12,6 +12,12 @@
 #include "gensio_filter_certauth.h"
 #include "gensio_filter_ssl.h"
 
+#ifdef _WIN32
+#define DIRSEP '\\'
+#else
+#define DIRSEP '/'
+#endif
+
 struct gensio_certauth_filter_data {
     struct gensio_os_funcs *o;
     bool is_client;
@@ -21,9 +27,12 @@ struct gensio_certauth_filter_data {
     char *username;
     char *password;
     char *service;
+    char *val_2fa;
+    unsigned int len_2fa;
     bool allow_authfail;
     bool use_child_auth;
     bool enable_password;
+    bool do_2fa; /* Ask for two-factor authentication, version 2+ */
 
     /*
      * The following is only used for testing. so certauth can be run
@@ -67,7 +76,7 @@ static void EVP_MD_CTX_free(EVP_MD_CTX *c)
 
 #define GENSIO_CERTAUTH_DATA_SIZE	2048
 #define GENSIO_CERTAUTH_CHALLENGE_SIZE	32
-#define GENSIO_CERTAUTH_VERSION		1
+#define GENSIO_CERTAUTH_VERSION		3
 
 /*
  * Passwords are always sent in this size buffer to keep an attacker
@@ -109,6 +118,9 @@ enum certauth_state {
      * If apps says to continue verification, send the SERVERHELLO
      * (containing the version, random challenge, and optional
      * options) and goes into CHALLENGE_RESPONSE.
+     *
+     * Message contains a VERSION element, an optional USERNAME
+     * element, and an optional SERVICE element.
      */
     CERTAUTH_CLIENTHELLO = 1,
 
@@ -120,6 +132,10 @@ enum certauth_state {
      *
      * Client may also receive a SERVERDONE in this state if the
      * authorization is rejected or authorized on username alone.
+     *
+     * Message contains a VERSION element, a CHALLENGE_DATA element,
+     * and an optional AUX_DATA element.  AUX_DATA is version 2 or
+     * later.
      */
     CERTAUTH_SERVERHELLO = 2,
 
@@ -138,15 +154,23 @@ enum certauth_state {
      * verification, send a SERVERDONE giving the result and go into
      * passthrough mode.  If the certificate does not verify, send
      * a PASSWORD_REQUEST (no data) and go into PASSWORD mode.
+     *
+     * Message contains a CERTIFICATE element, a CHALLENGE_RSP
+     * element, and an optional AUX_DATA element.  AUX_DATA is version
+     * 2 or later.  AUX_DATA has to be here, and not in CLIENTHELLO,
+     * because the client doesn't know the remote version when sending
+     * a CLIENTHELLO.
      */
     CERTAUTH_CHALLENGE_RESPONSE = 3,
 
     /*
-     * Client waits for PASSWORD_REQUEST.  One byte, after the request
-     * tells whether to actually send the password (1) or send a dummy (2).
+     * Client waits for PASSWORD_REQUEST.
      *
      * Client may also receive a SERVERDONE in this state if the
      * authorization is rejected or authorized by the certificate.
+     *
+     * Message contains a PASSWORD_TYPE element to tell how to handle
+     * the request.
      */
     CERTAUTH_PASSWORD_REQUEST = 4,
 
@@ -154,12 +178,17 @@ enum certauth_state {
      * Server waits for a password.  When received. the application is
      * notified of the password.  Send a SERVERDONE with error result
      * from the app.
+     *
+     * Message contains either a PASSWORD_DATA element or a DUMMY_DATA
+     * element and then an optional 2FA_DATA element.
      */
     CERTAUTH_PASSWORD = 5,
 
     /*
      * Client waits for SERVERDONE and goes into passthrough mode if
      * successful, contains the result.
+     *
+     * Message contains a RESULT element.
      */
     CERTAUTH_SERVERDONE = 6,
 
@@ -249,11 +278,28 @@ enum certauth_elements {
     CERTAUTH_DUMMY_DATA		= 109,
 
     /*
-     * 110 <n>
-     * What password data we are asking for.
+     * 110 2 <val>
+     * The bottom two bits is what password data we are asking for, either
+     * send the password (1) or send a dummy (2).  Bit 8 requests that a
+     * 2FA_DATA element be sent, also.
      */
     CERTAUTH_PASSWORD_TYPE	= 110,
 
+    /*
+     * 111 <n> <2fa data length n>
+     *
+     * The service is used to transfer 2-factor auth data.  Added in version 2.
+     */
+    CERTAUTH_2FA_DATA		= 111,
+
+    /*
+     * 112 <n> <data>
+     *
+     * Holds a block of generic data.  It has no meaning to certauth,
+     * it is used to transfer useful information to the other end.
+     */
+    CERTAUTH_AUX_DATA		= 112,
+
     /*
      * 200
      *
@@ -262,14 +308,22 @@ enum certauth_elements {
     CERTAUTH_END		= 200
 };
 #define CERTAUTH_MIN_ELEMENT CERTAUTH_VERSION
-#define CERTAUTH_MAX_ELEMENT CERTAUTH_PASSWORD_TYPE
+#define CERTAUTH_MAX_ELEMENT CERTAUTH_AUX_DATA
 
 #define CERTAUTH_RESULT_SUCCESS	1
 #define CERTAUTH_RESULT_FAILURE	2
 #define CERTAUTH_RESULT_ERR	3
 
+/* Request a password. */
 #define CERTAUTH_PASSWORD_TYPE_REQ	1
+/* Don't send a password, just send dummy data. */
 #define CERTAUTH_PASSWORD_TYPE_DUMMY	2
+#define CERTAUTH_PASSWORD_TYPE_MASK	0xff
+/*
+ * Request 2 factor authentication data on top of password or dummy data.
+ * Added in version 2.
+ */
+#define CERTAUTH_PASSWORD_TYPE_BIT_2FA	(1 << 8)
 
 struct certauth_filter {
     struct gensio_filter *filter;
@@ -287,6 +341,9 @@ struct certauth_filter {
     /* Version number from the remote end. */
     unsigned int version;
 
+    /* My version number, may be lowered due to openssl capabilities. */
+    unsigned int my_version;
+
     /* Result from the server or local verification. */
     unsigned int result;
 
@@ -302,6 +359,9 @@ struct certauth_filter {
     /* Enable password authentication. */
     bool enable_password;
 
+    /* Enable 2-factor authentication. */
+    bool do_2fa;
+
     char *username;
     size_t username_len;
 
@@ -309,6 +369,16 @@ struct certauth_filter {
     char *password;
     size_t password_len;
 
+    bool req_2fa_val;
+    unsigned char *val_2fa;
+    gensiods len_2fa;
+
+    /* Aux data locally and from the remote end. */
+    unsigned char *val_aux;
+    gensiods len_aux;
+    unsigned char *val_rem_aux;
+    gensiods len_rem_aux;
+
     char *service;
     size_t service_len;
 
@@ -325,6 +395,8 @@ struct certauth_filter {
     BUF_MEM cert_buf_mem;
     BIO *cert_bio;
     const EVP_MD *rsa_md5;
+    const EVP_MD *sha3_512;
+    const EVP_MD *digest;
 
     unsigned char *read_buf;
     gensiods read_buf_len;
@@ -635,7 +707,6 @@ certauth_get_cert(struct certauth_filter *sfilter)
 	gca_logs_err(sfilter, "Failure reading cert from network");
 	return GE_NOCERT;
     }
-    sfilter->write_buf_len += sfilter->cert_buf_mem.length;
 
     sfilter->sk_ca = sk_X509_new_null();
     if (!sfilter->sk_ca) {
@@ -672,7 +743,7 @@ certauth_add_challenge_rsp(struct certauth_filter *sfilter)
 	gca_log_err(sfilter, "Unable to allocate signature context");
 	return GE_NOMEM;
     }
-    if (!EVP_SignInit(sign_ctx, sfilter->rsa_md5)) {
+    if (!EVP_SignInit(sign_ctx, sfilter->digest)) {
 	gca_logs_err(sfilter, "Signature init failed");
 	goto out_nomem;
     }
@@ -714,7 +785,7 @@ certauth_check_challenge(struct certauth_filter *sfilter)
 	gca_log_err(sfilter, "Unable to allocate verify context");
 	return GE_NOMEM;
     }
-    if (!EVP_VerifyInit(sign_ctx, sfilter->rsa_md5)) {
+    if (!EVP_VerifyInit(sign_ctx, sfilter->digest)) {
 	gca_logs_err(sfilter, "Verify init failed");
 	goto out_nomem;
     }
@@ -786,13 +857,24 @@ certauth_send_server_done(struct certauth_filter *sfilter)
     certauth_write_byte(sfilter, CERTAUTH_END);
 }
 
+static void
+set_digest(struct certauth_filter *sfilter)
+{
+    if (sfilter->version >= 3 && sfilter->my_version >= 3)
+	sfilter->digest = sfilter->sha3_512;
+    else
+	sfilter->digest = sfilter->rsa_md5;
+}
+
 static int
 certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 {
     struct certauth_filter *sfilter = filter_to_certauth(filter);
     struct gensio *io;
+    gensiods len;
     bool password_requested = false;
     int err, rv;
+    unsigned int req;
 
     certauth_lock(sfilter);
     if (sfilter->pending_err)
@@ -806,7 +888,7 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	certauth_write_byte(sfilter, CERTAUTH_CLIENTHELLO);
 	certauth_write_byte(sfilter, CERTAUTH_VERSION);
 	certauth_write_u16(sfilter, 2);
-	certauth_write_u16(sfilter, GENSIO_CERTAUTH_VERSION);
+	certauth_write_u16(sfilter, sfilter->my_version);
 	if (sfilter->username && sfilter->username_len) {
 	    certauth_write_byte(sfilter, CERTAUTH_USERNAME);
 	    certauth_write_u16(sfilter, sfilter->username_len);
@@ -831,6 +913,16 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	    sfilter->pending_err = GE_DATAMISSING;
 	    break;
 	}
+
+	/* Verify support for things requested. */
+	if (sfilter->do_2fa && sfilter->version < 2) {
+	    gca_log_err(sfilter,
+		"2-factor auth requested but other end doesn't have it");
+	    sfilter->pending_err = GE_INVAL;
+	    break;
+	}
+	set_digest(sfilter);
+
 	io = gensio_filter_get_gensio(filter);
 	if (sfilter->use_child_auth) {
 	    if (gensio_is_authenticated(io)) {
@@ -873,7 +965,7 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	certauth_write_byte(sfilter, CERTAUTH_SERVERHELLO);
 	certauth_write_byte(sfilter, CERTAUTH_VERSION);
 	certauth_write_u16(sfilter, 2);
-	certauth_write_u16(sfilter, GENSIO_CERTAUTH_VERSION);
+	certauth_write_u16(sfilter, sfilter->my_version);
 	certauth_write_byte(sfilter, CERTAUTH_CHALLENGE_DATA);
 	certauth_write_u16(sfilter, sfilter->challenge_data_size);
 	if (!RAND_bytes(sfilter->challenge_data,
@@ -884,6 +976,11 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	}
 	certauth_write(sfilter, sfilter->challenge_data,
 		       sfilter->challenge_data_size);
+	if (sfilter->version >= 2 && sfilter->len_aux && sfilter->val_aux) {
+	    certauth_write_byte(sfilter, CERTAUTH_AUX_DATA);
+	    certauth_write_u16(sfilter, sfilter->len_aux);
+	    certauth_write(sfilter, sfilter->val_aux, sfilter->len_aux);
+	}
 
 	certauth_write_byte(sfilter, CERTAUTH_END);
 
@@ -901,6 +998,7 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	    sfilter->pending_err = GE_DATAMISSING;
 	    break;
 	}
+	set_digest(sfilter);
 	sfilter->write_buf_len = 0;
 	certauth_write_byte(sfilter, CERTAUTH_CHALLENGE_RESPONSE);
 
@@ -921,6 +1019,11 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	    /* Mask the fact we are not sending a response. */
 	    certauth_add_dummy(sfilter, 256);
 	}
+	if (sfilter->version >= 2 && sfilter->len_aux && sfilter->val_aux) {
+	    certauth_write_byte(sfilter, CERTAUTH_AUX_DATA);
+	    certauth_write_u16(sfilter, sfilter->len_aux);
+	    certauth_write(sfilter, sfilter->val_aux, sfilter->len_aux);
+	}
 
 	certauth_write_byte(sfilter, CERTAUTH_END);
 
@@ -935,6 +1038,7 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	     */
 	    goto try_password;
 	}
+
 	if (!!sfilter->cert != !!sfilter->response_result) {
 	    gca_log_err(sfilter, "Remote end did not send cert and response");
 	    sfilter->pending_err = GE_PROTOERR;
@@ -984,22 +1088,25 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	 */
 
     try_password:
+	if (!sfilter->result && sfilter->enable_password)
+	    req = CERTAUTH_PASSWORD_TYPE_REQ;
+	else
+	    req = CERTAUTH_PASSWORD_TYPE_DUMMY;
+	/* Request 2 factor authentication data. */
+	if (sfilter->do_2fa)
+	    req |= CERTAUTH_PASSWORD_TYPE_BIT_2FA;
 	sfilter->write_buf_len = 0;
 	certauth_write_byte(sfilter, CERTAUTH_PASSWORD_REQUEST);
 	certauth_write_byte(sfilter, CERTAUTH_PASSWORD_TYPE);
 	certauth_write_u16(sfilter, 2);
-	certauth_write_u16(sfilter,
-			   sfilter->result || !sfilter->enable_password ?
-			   CERTAUTH_PASSWORD_TYPE_DUMMY :
-			   CERTAUTH_PASSWORD_TYPE_REQ);
+	certauth_write_u16(sfilter, req);
 	certauth_write_byte(sfilter, CERTAUTH_END);
 	sfilter->state = CERTAUTH_PASSWORD;
 	break;
 
     case CERTAUTH_PASSWORD_REQUEST:
 	if (!sfilter->password_req_val) {
-	    gca_log_err(sfilter,
-			   "Remote client didn't send request value");
+	    gca_log_err(sfilter, "Remote client didn't send request value");
 	    sfilter->pending_err = GE_DATAMISSING;
 	    goto finish_result;
 	}
@@ -1012,6 +1119,10 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	}
 
 	if (!sfilter->enable_password) {
+	    /*
+	     * If we don't have passwords enabled but the other end
+	     * requests, send all zeros.
+	     */
 	    certauth_write_byte(sfilter, CERTAUTH_PASSWORD);
 	    certauth_write_byte(sfilter, CERTAUTH_PASSWORD_DATA);
 	    certauth_write_u16(sfilter, sfilter->password_len);
@@ -1051,15 +1162,82 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	    memset(sfilter->password, 0, sfilter->password_len);
 
     password_done:
+	if (sfilter->val_2fa) {
+	    if (sfilter->version < 2) {
+		gca_log_err(sfilter,
+		    "2-factor auth given, but server doesn't support it");
+		sfilter->pending_err = GE_INVAL;
+		goto finish_result;
+	    }
+	    goto send_2fa;
+	}
+	if (!sfilter->req_2fa_val)
+	    goto password_end;
+
+	password_requested = false;
+	if (!sfilter->val_2fa) {
+	    /* Empty 2fa data, ask the user. */
+	    certauth_unlock(sfilter);
+	    err = gensio_filter_do_event(sfilter->filter,
+					 GENSIO_EVENT_REQUEST_2FA, 0,
+					 (unsigned char *) &sfilter->val_2fa,
+					 &sfilter->len_2fa,
+					 NULL);
+	    certauth_lock(sfilter);
+	    if (err) {
+		if (err != GE_NOTSUP) {
+		    gca_log_err(sfilter, "Error fetching 2-factor auth: %s",
+				gensio_err_to_str(err));
+		    sfilter->pending_err = err;
+		    goto finish_result;
+		}
+	    }
+	    password_requested = true;
+	    if (sfilter->len_2fa > 65535 || sfilter->len_2fa < 1) {
+		gca_log_err(sfilter, "2-factor auth data bad size: %lld",
+			    (unsigned long long) sfilter->len_2fa);
+		sfilter->pending_err = GE_TOOBIG;
+		goto clear_2fa;
+	    }
+	}
+
+    send_2fa:
+	certauth_write_byte(sfilter, CERTAUTH_2FA_DATA);
+	certauth_write_u16(sfilter, sfilter->len_2fa);
+	if (sfilter->len_2fa)
+	    certauth_write(sfilter, sfilter->val_2fa, sfilter->len_2fa);
+    clear_2fa:
+	if (password_requested) {
+	    memset(sfilter->val_2fa, 0, sfilter->len_2fa);
+	    sfilter->o->free(sfilter->o, sfilter->val_2fa);
+	    sfilter->val_2fa = NULL;
+	    sfilter->len_2fa = 0;
+	}
+	if (sfilter->pending_err)
+	    goto finish_result;
+    password_end:
 	certauth_write_byte(sfilter, CERTAUTH_END);
 
 	sfilter->state = CERTAUTH_SERVERDONE;
 	break;
 
     case CERTAUTH_PASSWORD:
-	if (sfilter->result)
-	    /* Already verified, the rest was for show. */
+	if (sfilter->do_2fa && !sfilter->val_2fa) {
+	    /* Remote end didn't send a password and we requested one. */
+	    gca_log_err(sfilter, "Remote client didn't send 2fa data");
+	    sfilter->pending_err = GE_DATAMISSING;
 	    goto finish_result;
+	}
+
+	if (sfilter->result) {
+	    /* Already verified by certificate, the password was for show. */
+	    if (sfilter->val_2fa
+			&& sfilter->result == CERTAUTH_RESULT_SUCCESS) {
+		sfilter->result = 0;
+		goto check_2fa;
+	    }
+	    goto finish_result;
+	}
 
 	if (sfilter->enable_password && !sfilter->password) {
 	    /* Remote end didn't send a password and we requested one. */
@@ -1071,12 +1249,20 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	if (!sfilter->password || !*sfilter->password)
 	    goto finish_result;
 
+	len = sfilter->password_len;
 	certauth_unlock(sfilter);
 	err = gensio_filter_do_event(sfilter->filter,
 				     GENSIO_EVENT_PASSWORD_VERIFY, 0,
 				     (unsigned char *) sfilter->password,
-				     NULL, NULL);
+				     &len, NULL);
 	certauth_lock(sfilter);
+	if (sfilter->state != CERTAUTH_PASSWORD)
+	    /*
+	     * Either something went wrong, or the user called
+	     * GENSIO_CONTROL_FINISH_INIT.  We just exit in either
+	     * case.
+	     */
+	    goto out_finish;
 	memset(sfilter->password, 0, sfilter->password_len);
 	if (!err) {
 	    sfilter->result = CERTAUTH_RESULT_SUCCESS;
@@ -1087,6 +1273,36 @@ certauth_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 			gensio_err_to_str(err));
 	    sfilter->pending_err = err;
 	}
+	goto finish_result;
+
+    check_2fa:
+	len = sfilter->len_2fa;
+	certauth_unlock(sfilter);
+	err = gensio_filter_do_event(sfilter->filter,
+				     GENSIO_EVENT_2FA_VERIFY, 0,
+				     (unsigned char *) sfilter->val_2fa,
+				     &len, NULL);
+	certauth_lock(sfilter);
+	if (sfilter->state != CERTAUTH_PASSWORD)
+	    /*
+	     * Either something went wrong, or the user called
+	     * GENSIO_CONTROL_FINISH_INIT.  We just exit in either
+	     * case.
+	     */
+	    goto out_finish;
+	memset(sfilter->val_2fa, 0, sfilter->len_2fa);
+	sfilter->o->free(sfilter->o, sfilter->val_2fa);
+	sfilter->val_2fa = NULL;
+	sfilter->len_2fa = 0;
+	if (!err) {
+	    sfilter->result = CERTAUTH_RESULT_SUCCESS;
+	} else if (err == GE_CERTINVALID) {
+	    gca_log_err(sfilter, "Application rejected 2-factor auth");
+	} else if (err != GE_NOTSUP) {
+	    gca_log_err(sfilter, "Error from application at 2-factor auth: %s",
+			gensio_err_to_str(err));
+	    sfilter->pending_err = err;
+	}
 
     finish_result:
 	certauth_send_server_done(sfilter);
@@ -1229,7 +1445,7 @@ certauth_handle_new_element(struct certauth_filter *sfilter)
 	    sfilter->pending_err = GE_PROTOERR;
 	    break;
 	}
-	sfilter->username = o->zalloc(0, sfilter->curr_elem_len + 1);
+	sfilter->username = o->zalloc(o, sfilter->curr_elem_len + 1);
 	sfilter->username_len = sfilter->curr_elem_len;
 	if (!sfilter->username) {
 	    gca_log_err(sfilter, "Unable to allocate memory for username");
@@ -1246,7 +1462,7 @@ certauth_handle_new_element(struct certauth_filter *sfilter)
 	    sfilter->pending_err = GE_PROTOERR;
 	    break;
 	}
-	sfilter->service = o->zalloc(0, sfilter->curr_elem_len + 1);
+	sfilter->service = o->zalloc(o, sfilter->curr_elem_len + 1);
 	sfilter->service_len = sfilter->curr_elem_len;
 	if (!sfilter->service) {
 	    gca_log_err(sfilter, "Unable to allocate memory for service");
@@ -1268,6 +1484,9 @@ certauth_handle_new_element(struct certauth_filter *sfilter)
 	    sfilter->pending_err = GE_PROTOERR;
 	} else {
 	    sfilter->password_req_val = certauth_buf_to_u16(sfilter->read_buf);
+	    sfilter->req_2fa_val = (sfilter->password_req_val &
+				    CERTAUTH_PASSWORD_TYPE_BIT_2FA);
+	    sfilter->password_req_val &= CERTAUTH_PASSWORD_TYPE_MASK;
 	    if (!sfilter->password_req_val) {
 		gca_log_err(sfilter, "password req was zero");
 		sfilter->pending_err = GE_PROTOERR;
@@ -1283,7 +1502,7 @@ certauth_handle_new_element(struct certauth_filter *sfilter)
 	}
 	sfilter->password_len = limited_strlen((char *) sfilter->read_buf,
 					       sfilter->curr_elem_len);
-	sfilter->password = o->zalloc(0, sfilter->password_len + 1);
+	sfilter->password = o->zalloc(o, sfilter->password_len + 1);
 	if (!sfilter->password) {
 	    gca_log_err(sfilter, "Unable to allocate memory for password");
 	    sfilter->pending_err = GE_NOMEM;
@@ -1293,6 +1512,39 @@ certauth_handle_new_element(struct certauth_filter *sfilter)
 	}
 	break;
 
+    case CERTAUTH_2FA_DATA:
+	if (sfilter->len_2fa) {
+	    gca_log_err(sfilter, "2-factor auth received when already set");
+	    sfilter->pending_err = GE_PROTOERR;
+	    break;
+	}
+	sfilter->len_2fa = sfilter->curr_elem_len;
+	sfilter->val_2fa = o->zalloc(o, sfilter->len_2fa + 1);
+	if (!sfilter->val_2fa) {
+	    gca_log_err(sfilter, "Unable to allocate memory for 2-factor auth");
+	    sfilter->pending_err = GE_NOMEM;
+	} else {
+	    memcpy(sfilter->val_2fa, sfilter->read_buf, sfilter->len_2fa);
+	}
+	break;
+
+    case CERTAUTH_AUX_DATA:
+	if (sfilter->len_rem_aux) {
+	    gca_log_err(sfilter, "Remote aux data received when already set");
+	    sfilter->pending_err = GE_PROTOERR;
+	    break;
+	}
+	sfilter->len_rem_aux = sfilter->curr_elem_len;
+	sfilter->val_rem_aux = o->zalloc(o, sfilter->len_rem_aux + 1);
+	if (!sfilter->val_rem_aux) {
+	    gca_log_err(sfilter, "Unable to allocate memory for remote aux");
+	    sfilter->pending_err = GE_NOMEM;
+	} else {
+	    memcpy(sfilter->val_rem_aux, sfilter->read_buf,
+		   sfilter->len_rem_aux);
+	}
+	break;
+
     case CERTAUTH_OPTIONS:
 	break;
 
@@ -1302,7 +1554,7 @@ certauth_handle_new_element(struct certauth_filter *sfilter)
 	    sfilter->pending_err = GE_PROTOERR;
 	    break;
 	}
-	sfilter->challenge_data = o->zalloc(0, sfilter->curr_elem_len);
+	sfilter->challenge_data = o->zalloc(o, sfilter->curr_elem_len);
 	sfilter->challenge_data_size = sfilter->curr_elem_len;
 	if (!sfilter->challenge_data) {
 	    gca_log_err(sfilter, "Unable to allocate memory for challenge");
@@ -1416,7 +1668,7 @@ certauth_ll_write(struct gensio_filter *filter,
 	    sfilter->state = CERTAUTH_SERVERDONE;
 	} else if (sfilter->curr_msg_type != sfilter->state) {
 	    gca_log_err(sfilter, "Expected message type %d, got %d",
-			sfilter->curr_msg_type, sfilter->state);
+			sfilter->state, sfilter->curr_msg_type);
 	    sfilter->pending_err = GE_PROTOERR;
 	    goto out_unlock;
 	}
@@ -1511,10 +1763,11 @@ static void
 certauth_cleanup(struct gensio_filter *filter)
 {
     struct certauth_filter *sfilter = filter_to_certauth(filter);
+    struct gensio_os_funcs *o = sfilter->o;
 
     if (sfilter->is_client) {
 	if (sfilter->challenge_data)
-	    sfilter->o->free(sfilter->o, sfilter->challenge_data);
+	    o->free(o, sfilter->challenge_data);
 	sfilter->challenge_data = NULL;
 	memset(sfilter->password, 0, sfilter->password_len);
     } else {
@@ -1527,19 +1780,32 @@ certauth_cleanup(struct gensio_filter *filter)
 	if (sfilter->password)
 	    memset(sfilter->password, 0, sfilter->password_len);
 	if (!sfilter->is_client && sfilter->password) {
-	    sfilter->o->free(sfilter->o, sfilter->password);
+	    o->free(o, sfilter->password);
 	    sfilter->password = NULL;
 	    sfilter->password_len = 0;
 	}
 	if (sfilter->username)
-	    sfilter->o->free(sfilter->o, sfilter->username);
+	    o->free(o, sfilter->username);
 	sfilter->username = NULL;
 	sfilter->username_len = 0;
 	if (sfilter->service)
-	    sfilter->o->free(sfilter->o, sfilter->service);
+	    o->free(o, sfilter->service);
 	sfilter->service = NULL;
 	sfilter->service_len = 0;
     }
+    if (sfilter->val_2fa)
+	o->free(o, sfilter->val_2fa);
+    sfilter->val_2fa = NULL;
+    sfilter->len_2fa = 0;
+
+    if (sfilter->val_aux)
+	o->free(o, sfilter->val_aux);
+    sfilter->val_aux = NULL;
+    sfilter->len_aux = 0;
+    if (sfilter->val_rem_aux)
+	o->free(o, sfilter->val_rem_aux);
+    sfilter->val_rem_aux = NULL;
+    sfilter->len_rem_aux = 0;
 
     sfilter->pending_err = 0;
     sfilter->password_req_val = 0;
@@ -1555,6 +1821,8 @@ certauth_cleanup(struct gensio_filter *filter)
 static void
 sfilter_free(struct certauth_filter *sfilter)
 {
+    struct gensio_os_funcs *o = sfilter->o;
+
     if (sfilter->cert)
 	X509_free(sfilter->cert);
     if (sfilter->sk_ca)
@@ -1562,30 +1830,30 @@ sfilter_free(struct certauth_filter *sfilter)
     if (sfilter->cert_bio)
 	BIO_free(sfilter->cert_bio);
     if (sfilter->lock)
-	sfilter->o->free_lock(sfilter->lock);
+	o->free_lock(sfilter->lock);
     if (sfilter->read_buf) {
 	memset(sfilter->read_buf, 0, sfilter->max_read_size);
-	sfilter->o->free(sfilter->o, sfilter->read_buf);
+	o->free(o, sfilter->read_buf);
     }
     if (sfilter->write_buf)
-	sfilter->o->free(sfilter->o, sfilter->write_buf);
+	o->free(o, sfilter->write_buf);
     if (sfilter->pkey)
 	EVP_PKEY_free(sfilter->pkey);
     if (sfilter->password) {
 	memset(sfilter->password, 0, sfilter->password_len);
-	sfilter->o->free(sfilter->o, sfilter->password);
+	o->free(o, sfilter->password);
     }
     if (sfilter->username)
-	sfilter->o->free(sfilter->o, sfilter->username);
+	o->free(o, sfilter->username);
     if (sfilter->service)
-	sfilter->o->free(sfilter->o, sfilter->service);
+	o->free(o, sfilter->service);
     if (sfilter->challenge_data)
-	sfilter->o->free(sfilter->o, sfilter->challenge_data);
+	o->free(o, sfilter->challenge_data);
     if (sfilter->filter)
 	gensio_filter_free_data(sfilter->filter);
     if (sfilter->verify_store)
 	X509_STORE_free(sfilter->verify_store);
-    sfilter->o->free(sfilter->o, sfilter);
+    o->free(o, sfilter);
 }
 
 static void
@@ -1603,6 +1871,7 @@ certauth_filter_control(struct gensio_filter *filter, bool get, int op,
     struct certauth_filter *sfilter = filter_to_certauth(filter);
     X509_STORE *store;
     char *CApath = NULL, *CAfile = NULL;
+    int rv = 0;
 
     switch (op) {
     case GENSIO_CONTROL_GET_PEER_CERT_NAME:
@@ -1618,8 +1887,6 @@ certauth_filter_control(struct gensio_filter *filter, bool get, int op,
 	return gensio_cert_to_buf(sfilter->cert, data, datalen);
 
     case GENSIO_CONTROL_USERNAME: {
-	int rv = 0;
-
 	certauth_lock(sfilter);
 	if (get) {
 	    if (!sfilter->username) {
@@ -1639,13 +1906,151 @@ certauth_filter_control(struct gensio_filter *filter, bool get, int op,
 	    }
 	    if (sfilter->username)
 		sfilter->o->free(sfilter->o, sfilter->username);
-	    sfilter->username = data;
+	    sfilter->username = newusername;
 	}
 	out_username:
 	certauth_unlock(sfilter);
 	return rv;
     }
 
+    case GENSIO_CONTROL_PASSWORD: {
+	certauth_lock(sfilter);
+	if (get) {
+	    if (!sfilter->password) {
+		rv = GE_DATAMISSING;
+		goto out_username;
+	    }
+	    *datalen = snprintf(data, *datalen, "%s", sfilter->password);
+	} else {
+	    char *newpw = NULL;
+
+	    if (data) {
+		newpw = gensio_strdup(sfilter->o, data);
+		if (!newpw) {
+		    rv = GE_NOMEM;
+		    goto out_pw;
+		}
+	    }
+	    if (sfilter->password)
+		sfilter->o->free(sfilter->o, sfilter->password);
+	    sfilter->password = newpw;
+	}
+	out_pw:
+	certauth_unlock(sfilter);
+	return rv;
+    }
+
+    case GENSIO_CONTROL_2FA: {
+	gensiods len;
+
+	certauth_lock(sfilter);
+	if (get) {
+	    if (!sfilter->val_2fa) {
+		rv = GE_DATAMISSING;
+		goto out_username;
+	    }
+	    len = sfilter->len_2fa;
+	    if (len > *datalen)
+		len = *datalen;
+	    memcpy(data, sfilter->val_2fa, len);
+	    *datalen = sfilter->len_2fa;
+	} else {
+	    unsigned char *new2fa = NULL;
+
+	    if (*datalen == 0)
+		data = NULL;
+	    if (data) {
+		new2fa = sfilter->o->zalloc(sfilter->o, *datalen);
+		if (!new2fa) {
+		    rv = GE_NOMEM;
+		    goto out_2fa;
+		}
+		memcpy(new2fa, data, *datalen);
+	    }
+	    if (sfilter->val_2fa)
+		sfilter->o->free(sfilter->o, sfilter->val_2fa);
+	    sfilter->val_2fa = new2fa;
+	    sfilter->len_2fa = *datalen;
+	}
+	out_2fa:
+	certauth_unlock(sfilter);
+	return rv;
+    }
+
+    case GENSIO_CONTROL_AUX_DATA: {
+	gensiods len;
+
+	certauth_lock(sfilter);
+	if (get) {
+	    if (!sfilter->val_aux) {
+		rv = GE_DATAMISSING;
+		goto out_username;
+	    }
+	    len = sfilter->len_aux;
+	    if (len > *datalen)
+		len = *datalen;
+	    memcpy(data, sfilter->val_aux, len);
+	    *datalen = sfilter->len_aux;
+	} else {
+	    unsigned char *newaux = NULL;
+
+	    if (*datalen == 0)
+		data = NULL;
+	    if (data) {
+		newaux = sfilter->o->zalloc(sfilter->o, *datalen);
+		if (!newaux) {
+		    rv = GE_NOMEM;
+		    goto out_aux;
+		}
+		memcpy(newaux, data, *datalen);
+	    }
+	    if (sfilter->val_aux)
+		sfilter->o->free(sfilter->o, sfilter->val_aux);
+	    sfilter->val_aux = newaux;
+	    sfilter->len_aux = *datalen;
+	}
+	out_aux:
+	certauth_unlock(sfilter);
+	return rv;
+    }
+
+    case GENSIO_CONTROL_REM_AUX_DATA: {
+	gensiods len;
+
+	certauth_lock(sfilter);
+	if (get) {
+	    if (!sfilter->val_rem_aux) {
+		rv = GE_DATAMISSING;
+		goto out_username;
+	    }
+	    len = sfilter->len_rem_aux;
+	    if (len > *datalen)
+		len = *datalen;
+	    memcpy(data, sfilter->val_rem_aux, len);
+	    *datalen = sfilter->len_rem_aux;
+	} else {
+	    unsigned char *newrem_aux = NULL;
+
+	    if (*datalen == 0)
+		data = NULL;
+	    if (data) {
+		newrem_aux = sfilter->o->zalloc(sfilter->o, *datalen);
+		if (!newrem_aux) {
+		    rv = GE_NOMEM;
+		    goto out_rem_aux;
+		}
+		memcpy(newrem_aux, data, *datalen);
+	    }
+	    if (sfilter->val_rem_aux)
+		sfilter->o->free(sfilter->o, sfilter->val_rem_aux);
+	    sfilter->val_rem_aux = newrem_aux;
+	    sfilter->len_rem_aux = *datalen;
+	}
+	out_rem_aux:
+	certauth_unlock(sfilter);
+	return rv;
+    }
+
     case GENSIO_CONTROL_SERVICE:
 	if (get) {
 	    gensiods to_copy;
@@ -1677,7 +2082,7 @@ certauth_filter_control(struct gensio_filter *filter, bool get, int op,
 	store = X509_STORE_new();
 	if (!store)
 	    return GE_NOMEM;
-	if (data[strlen(data) - 1] == '/')
+	if (data[strlen(data) - 1] == DIRSEP)
 	    CApath = data;
 	else
 	    CAfile = data;
@@ -1707,7 +2112,7 @@ certauth_filter_control(struct gensio_filter *filter, bool get, int op,
 
 static
 int gensio_certauth_filter_func(struct gensio_filter *filter, int op,
-				const void *func, void *data,
+				void *func, void *data,
 				gensiods *count,
 				void *buf, const void *cbuf,
 				gensiods buflen,
@@ -1769,9 +2174,10 @@ gensio_certauth_filter_raw_alloc(struct gensio_os_funcs *o,
 				 X509 *cert, STACK_OF(X509) *sk_ca,
 				 EVP_PKEY *pkey,
 				 const char *username, const char *password,
+				 const char *val_2fa, gensiods len_2fa,
 				 const char *service,
 				 bool allow_authfail, bool use_child_auth,
-				 bool enable_password,
+				 bool enable_password, bool do_2fa,
 				 struct gensio_filter **rfilter)
 {
     struct certauth_filter *sfilter;
@@ -1785,12 +2191,17 @@ gensio_certauth_filter_raw_alloc(struct gensio_os_funcs *o,
     sfilter->is_client = is_client;
     sfilter->allow_authfail = allow_authfail;
     sfilter->use_child_auth = use_child_auth;
-    sfilter->enable_password = enable_password,
+    sfilter->enable_password = enable_password;
+    sfilter->do_2fa = do_2fa;
+    sfilter->my_version = GENSIO_CERTAUTH_VERSION;
     sfilter->rsa_md5 = EVP_get_digestbyname("ssl3-md5");
     if (!sfilter->rsa_md5) {
 	rv = GE_IOERR;
 	goto out_err;
     }
+    sfilter->sha3_512 = EVP_get_digestbyname("sha3-512");
+    if (!sfilter->sha3_512)
+	sfilter->my_version = 2;
 
     if (is_client) {
 	/* Extra byte at the end so it's always nil terminated. */
@@ -1811,6 +2222,16 @@ gensio_certauth_filter_raw_alloc(struct gensio_os_funcs *o,
 
 	    strncpy(sfilter->password, password, GENSIO_CERTAUTH_PASSWORD_LEN);
 	}
+
+	if (val_2fa) {
+	    sfilter->val_2fa = o->zalloc(o, len_2fa);
+	    if (!sfilter->val_2fa) {
+		rv = GE_NOMEM;
+		goto out_err;
+	    }
+	    memcpy(sfilter->val_2fa, val_2fa, len_2fa);
+	    sfilter->len_2fa = len_2fa;
+	}
     }
 
     sfilter->lock = o->alloc_lock(o);
@@ -1896,6 +2317,10 @@ gensio_certauth_filter_config_free(struct gensio_certauth_filter_data *data)
 	memset(data->password, 0, strlen(data->password));
 	o->free(o, data->password);
     }
+    if (data->val_2fa) {
+	memset(data->val_2fa, 0, data->len_2fa);
+	o->free(o, data->val_2fa);
+    }
     if (data->username)
 	o->free(o, data->username);
     if (data->service)
@@ -1953,7 +2378,7 @@ gensio_certauth_filter_config(struct gensio_os_funcs *o,
 	    data->is_client = false;
 	else {
 	    gensio_log(o, GENSIO_LOG_ERR,
-		       "Unknown default certauth mode (%s), ignoring", str);
+		       "Unknown default certauth mode (%s), ignoring", fstr);
 	}
 	o->free(o, fstr);
     }
@@ -1990,6 +2415,15 @@ gensio_certauth_filter_config(struct gensio_os_funcs *o,
 		goto out_err;
 	    continue;
 	}
+	if (gensio_check_keyvalue(args[i], "2fa", &str) > 0) {
+	    data->len_2fa = strlen(str);
+	    if (data->len_2fa == 0)
+		goto out_err;
+	    data->val_2fa = gensio_strdup(o, str);
+	    if (!data->val_2fa)
+		goto out_err;
+	    continue;
+	}
 	if (gensio_check_keyvalue(args[i], "service", &str) > 0) {
 	    data->service = gensio_strdup(o, str);
 	    if (!data->service)
@@ -2008,6 +2442,9 @@ gensio_certauth_filter_config(struct gensio_os_funcs *o,
 	if (gensio_check_keybool(args[i], "enable-password",
 				 &data->enable_password) > 0)
 	    continue;
+	if (gensio_check_keybool(args[i], "enable-2fa",
+				 &data->do_2fa) > 0)
+	    continue;
 	if (gensio_check_keybool(args[i], "allow-unencrypted",
 				 &data->allow_unencrypted) > 0)
 	    continue;
@@ -2085,12 +2522,12 @@ gensio_certauth_filter_config(struct gensio_os_funcs *o,
     }
 
     if (data->is_client) {
-	if (data->CAfilepath) {
+	if (data->CAfilepath || data->do_2fa) {
 	    rv = GE_INVAL;
 	    goto out_err;
 	}
     } else {
-	if (data->keyfile || data->username) {
+	if (data->keyfile || data->username || data->val_2fa) {
 	    rv = GE_INVAL;
 	    goto out_err;
 	}
@@ -2215,7 +2652,7 @@ gensio_certauth_filter_alloc(struct gensio_certauth_filter_data *data,
     if (data->CAfilepath && data->CAfilepath[0]) {
 	char *CAfile = NULL, *CApath = NULL;
 
-	if (data->CAfilepath[strlen(data->CAfilepath) - 1] == '/')
+	if (data->CAfilepath[strlen(data->CAfilepath) - 1] == DIRSEP)
 	    CApath = data->CAfilepath;
 	else
 	    CAfile = data->CAfilepath;
@@ -2237,10 +2674,12 @@ gensio_certauth_filter_alloc(struct gensio_certauth_filter_data *data,
     rv = gensio_certauth_filter_raw_alloc(o, data->is_client, store,
 					  cert, sk_ca, pkey,
 					  data->username, data->password,
+					  data->val_2fa, data->len_2fa,
 					  data->service,
 					  data->allow_authfail,
 					  data->use_child_auth,
 					  data->enable_password,
+					  data->do_2fa,
 					  &filter);
     if (rv)
 	goto err;
