@@ -140,10 +140,16 @@ msgdelim_try_connect(struct gensio_filter *filter, gensio_time *timeout)
     return 0;
 }
 
+#include <stdio.h>
 static int
 msgdelim_try_disconnect(struct gensio_filter *filter, gensio_time *timeout)
 {
-    return 0;
+    struct msgdelim_filter *mfilter = filter_to_msgdelim(filter);
+
+    if (mfilter->write_data_len == 0 || !mfilter->out_msg_complete)
+	return 0;
+    else
+	return GE_INPROGRESS;
 }
 
 static void
@@ -390,11 +396,11 @@ msgdelim_control(struct gensio_filter *filter, bool get, int op, char *data,
 }
 
 static int gensio_msgdelim_filter_func(struct gensio_filter *filter, int op,
-				     const void *func, void *data,
-				     gensiods *count,
-				     void *buf, const void *cbuf,
-				     gensiods buflen,
-				     const char *const *auxdata)
+				       void *func, void *data,
+				       gensiods *count,
+				       void *buf, const void *cbuf,
+				       gensiods buflen,
+				       const char *const *auxdata)
 {
     switch (op) {
     case GENSIO_FILTER_FUNC_UL_READ_PENDING:
