@@ -382,6 +382,7 @@ first_xmitpkt_to_send(struct relpkt_filter *rfilter)
 	    return &(rfilter->xmitpkts[pos]);
     }
     assert(0);
+    return NULL;
 }
 
 static void
@@ -794,7 +795,8 @@ relpkt_ul_write(struct relpkt_filter *rfilter,
 		    rfilter->nr_waiting_xmitpkt--;
 		    rfilter->send_since_timeout = true;
 		} else {
-		    *endbool = false;
+		    if (endbool)
+			*endbool = false;
 		    if (finish_close) {
 			rfilter->err = GE_REMCLOSE;
 			err = GE_REMCLOSE;
@@ -1175,7 +1177,7 @@ relpkt_filter_timeout(struct relpkt_filter *rfilter)
 }
 
 static int gensio_relpkt_filter_func(struct gensio_filter *filter, int op,
-				     const void *func, void *data,
+				     void *func, void *data,
 				     gensiods *count,
 				     void *buf, const void *cbuf,
 				     gensiods buflen,
