@@ -20,6 +20,12 @@
 #include <openssl/bio.h>
 #include <openssl/err.h>
 
+#ifdef _WIN32
+#define DIRSEP '\\'
+#else
+#define DIRSEP '/'
+#endif
+
 /* In gensio_filter_ssl.c, semi-private. */
 int gensio_cert_get_name(X509 *cert, char *data, gensiods *datalen);
 int gensio_cert_to_buf(X509 *cert, char *buf, gensiods *datalen);
@@ -86,9 +92,10 @@ struct ssl_filter {
     gensiods write_data_len;
 
     /* This is data from BIO_read() waiting to be sent to the lower layer. */
-    unsigned char xmit_buf[1024];
+    unsigned char *xmit_buf;
     gensiods xmit_buf_pos;
     gensiods xmit_buf_len;
+    gensiods max_xmit_buf;
 
     /*
      * SSL has asked for something.
@@ -220,6 +227,18 @@ ssl_ll_read_needed(struct gensio_filter *filter)
     return rv;
 }
 
+static bool
+ssl_ll_can_write(struct gensio_filter *filter, bool *val)
+{
+    struct ssl_filter *sfilter = filter_to_ssl(filter);
+
+    ssl_lock(sfilter);
+    *val = sfilter->write_data_len == 0 && sfilter->xmit_buf_len == 0;
+    ssl_unlock(sfilter);
+
+    return 0;
+}
+
 static int
 ssl_verify_cb(int preverify_ok, X509_STORE_CTX *x509_ctx)
 {
@@ -318,7 +337,6 @@ ssl_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	    break;
 
 	case SSL_ERROR_SSL:
-	err_rpt:
 	    gssl_logs_err(sfilter, "Failed SSL startup");
 	    rv = GE_PROTOERR;
 	    break;
@@ -328,7 +346,8 @@ ssl_try_connect(struct gensio_filter *filter, gensio_time *timeout)
 	    break;
 
 	default:
-	    gssl_log_err(sfilter, "Failed SSL startup");
+	err_rpt:
+	    gssl_log_err(sfilter, "Failed SSL startup: 0x%8.8x", err);
 	    rv = GE_COMMERR;
 	}
     }
@@ -505,13 +524,12 @@ ssl_ul_write(struct gensio_filter *filter,
 
     if (!err && sfilter->xmit_buf_len == 0) {
 	int rdlen = BIO_read(sfilter->io_bio, sfilter->xmit_buf,
-			     sizeof(sfilter->xmit_buf));
+			     sfilter->max_xmit_buf);
 
 	if (rdlen <= 0) {
 	    if (!BIO_should_retry(sfilter->io_bio)) {
 		gssl_log_err(sfilter, "Failed BIO read");
 		err = GE_COMMERR;
-		rdlen = 0;
 	    }
 	} else {
 	    sfilter->xmit_buf_len = rdlen;
@@ -728,6 +746,8 @@ sfilter_free(struct ssl_filter *sfilter)
 	memset(sfilter->read_data, 0, sfilter->max_read_size);
 	sfilter->o->free(sfilter->o, sfilter->read_data);
     }
+    if (sfilter->xmit_buf)
+	sfilter->o->free(sfilter->o, sfilter->xmit_buf);
     if (sfilter->write_data)
 	sfilter->o->free(sfilter->o, sfilter->write_data);
     if (sfilter->filter)
@@ -870,7 +890,7 @@ ssl_filter_control(struct gensio_filter *filter, bool get, int op, char *data,
 	store = X509_STORE_new();
 	if (!store)
 	    return GE_NOMEM;
-	if (data[strlen(data) - 1] == '/')
+	if (data[strlen(data) - 1] == DIRSEP)
 	    CApath = data;
 	else
 	    CAfile = data;
@@ -942,7 +962,7 @@ ssl_filter_control(struct gensio_filter *filter, bool get, int op, char *data,
 }
 
 static int gensio_ssl_filter_func(struct gensio_filter *filter, int op,
-				  const void *func, void *data,
+				  void *func, void *data,
 				  gensiods *count,
 				  void *buf, const void *cbuf,
 				  gensiods buflen,
@@ -962,6 +982,9 @@ static int gensio_ssl_filter_func(struct gensio_filter *filter, int op,
     case GENSIO_FILTER_FUNC_LL_READ_NEEDED:
 	return ssl_ll_read_needed(filter);
 
+    case GENSIO_FILTER_FUNC_LL_CAN_WRITE:
+	return ssl_ll_can_write(filter, data);
+
     case GENSIO_FILTER_FUNC_CHECK_OPEN_DONE:
 	return ssl_check_open_done(filter, data);
 
@@ -1095,14 +1118,21 @@ gensio_ssl_filter_raw_alloc(struct gensio_os_funcs *o,
     if (!sfilter->lock)
 	goto out_nomem;
 
-    sfilter->read_data = o->zalloc(o, max_read_size);
+    sfilter->read_data = o->zalloc(o, sfilter->max_read_size);
     if (!sfilter->read_data)
 	goto out_nomem;
 
-    sfilter->write_data = o->zalloc(o, max_write_size);
+    sfilter->write_data = o->zalloc(o, sfilter->max_write_size);
     if (!sfilter->write_data)
 	goto out_nomem;
 
+    sfilter->max_xmit_buf = sfilter->max_write_size + 128;
+    if (sfilter->max_xmit_buf < 1024)
+	sfilter->max_xmit_buf = 1024; /* Enough room for the protocol. */
+    sfilter->xmit_buf = o->zalloc(o, sfilter->max_xmit_buf);
+    if (!sfilter->xmit_buf)
+	goto out_nomem;
+
     sfilter->filter = gensio_filter_alloc_data(o, gensio_ssl_filter_func,
 					       sfilter);
     if (!sfilter->filter)
@@ -1309,7 +1339,7 @@ gensio_ssl_filter_alloc(struct gensio_ssl_filter_data *data,
     if (data->CAfilepath && data->CAfilepath[0]) {
 	char *CAfile = NULL, *CApath = NULL;
 
-	if (data->CAfilepath[strlen(data->CAfilepath) - 1] == '/')
+	if (data->CAfilepath[strlen(data->CAfilepath) - 1] == DIRSEP)
 	    CApath = data->CAfilepath;
 	else
 	    CAfile = data->CAfilepath;
