@@ -106,7 +106,7 @@ struct fd_ll {
     gensio_ll_cb cb;
     void *cb_data;
 
-    int fd;
+    struct gensio_iod *iod;
 
     enum fd_state state;
 
@@ -134,6 +134,7 @@ struct fd_ll {
     const char *const *auxdata;
 
     bool in_read;
+    bool in_write;
 
     /*
      * Used to run read callbacks from the selector to avoid running
@@ -155,7 +156,7 @@ struct fd_ll {
 
 #define ll_to_fd(v) ((struct fd_ll *) gensio_ll_get_user_data(v))
 
-static void fd_handle_write_ready(struct fd_ll *fdll, int fd);
+static void fd_handle_write_ready(struct fd_ll *fdll, struct gensio_iod *iod);
 
 static void fd_finish_free(struct fd_ll *fdll)
 {
@@ -336,10 +337,10 @@ fd_write(struct gensio_ll *ll, gensiods *rcount,
     struct fd_ll *fdll = ll_to_fd(ll);
 
     if (fdll->ops->write)
-	return fdll->ops->write(fdll->handler_data, fdll->fd,
+	return fdll->ops->write(fdll->handler_data, fdll->iod,
 				rcount, sg, sglen, auxdata);
 
-    return gensio_os_write(fdll->o, fdll->fd, sg, sglen, rcount);
+    return fdll->o->write(fdll->iod, sg, sglen, rcount);
 }
 
 static void
@@ -349,9 +350,11 @@ fd_deliver_read_data(struct fd_ll *fdll, int err)
 	gensiods count;
 
     retry:
+	fd_unlock(fdll);
 	count = gensio_fd_ll_callback(fdll->ll, GENSIO_LL_CB_READ, err,
 				      fdll->read_data + fdll->read_data_pos,
 				      fdll->read_data_len, fdll->auxdata);
+	fd_lock(fdll);
 	if (err || count >= fdll->read_data_len) {
 	    fdll->read_data_pos = 0;
 	    fdll->read_data_len = 0;
@@ -383,10 +386,10 @@ fd_finish_open(struct fd_ll *fdll, int err)
 
     if (fdll->state == FD_OPEN) {
 	if (fdll->read_enabled)
-	    fdll->o->set_read_handler(fdll->o, fdll->fd, true);
+	    fdll->o->set_read_handler(fdll->iod, true);
 	if (fdll->write_enabled)
-	    fdll->o->set_write_handler(fdll->o, fdll->fd, true);
-	fdll->o->set_except_handler(fdll->o, fdll->fd,
+	    fdll->o->set_write_handler(fdll->iod, true);
+	fdll->o->set_except_handler(fdll->iod,
 				    fdll->read_enabled || fdll->write_enabled);
     }
 }
@@ -418,19 +421,16 @@ fd_deferred_op(struct gensio_runner *runner, void *cbdata)
 
     if (fdll->deferred_except && fdll->write_enabled) {
 	fdll->deferred_except = false;
-	if (fdll->fd >= 0)
-	    fd_handle_write_ready(fdll, fdll->fd);
+	if (fdll->iod)
+	    fd_handle_write_ready(fdll, fdll->iod);
     }
 
     while (fdll->deferred_read) {
 	fdll->deferred_read = false;
 
 	fdll->in_read = true;
-	while (fdll->read_enabled && fdll->read_data_len) {
-	    fd_unlock(fdll);
+	while (fdll->read_enabled && fdll->read_data_len)
 	    fd_deliver_read_data(fdll, 0);
-	    fd_lock(fdll);
-	}
 	fdll->in_read = false;
     }
 
@@ -441,10 +441,10 @@ fd_deferred_op(struct gensio_runner *runner, void *cbdata)
 
     fdll->deferred_op_pending = false;
     if (fdll->state == FD_OPEN) {
-	fdll->o->set_read_handler(fdll->o, fdll->fd, fdll->read_enabled);
-	fdll->o->set_except_handler(fdll->o, fdll->fd,
+	fdll->o->set_read_handler(fdll->iod, fdll->read_enabled);
+	fdll->o->set_except_handler(fdll->iod,
 				    fdll->read_enabled || fdll->write_enabled);
-	fdll->o->set_write_handler(fdll->o, fdll->fd, fdll->write_enabled);
+	fdll->o->set_write_handler(fdll->iod, fdll->write_enabled);
     }
     fd_deref_and_unlock(fdll);
 }
@@ -464,21 +464,21 @@ static void
 fd_start_close(struct fd_ll *fdll)
 {
     if (fdll->ops->check_close)
-	fdll->ops->check_close(fdll->handler_data,
+	fdll->ops->check_close(fdll->handler_data, fdll->iod,
 			       GENSIO_LL_CLOSE_STATE_START, NULL);
-    if (fdll->fd == -1) {
+    if (!fdll->iod) {
 	fdll->deferred_close = true;
 	fd_sched_deferred_op(fdll);
     } else if (fdll->state != FD_OPEN_ERR_WAIT &&
 	       fdll->state != FD_IN_OPEN_RETRY) {
-	fdll->o->clear_fd_handlers(fdll->o, fdll->fd);
+	fdll->o->clear_fd_handlers(fdll->iod);
     }
     fd_set_state(fdll, FD_IN_CLOSE);
 }
 
 static void
 fd_handle_incoming(struct fd_ll *fdll,
-		   int (*doread)(int fd, void *buf, gensiods count,
+		   int (*doread)(struct gensio_iod *iod, void *buf, gensiods count,
 				 gensiods *rcount, const char ***auxdata,
 				 void *cb_data),
 		   const char **auxdata, void *cb_data)
@@ -487,17 +487,16 @@ fd_handle_incoming(struct fd_ll *fdll,
     gensiods count;
 
     fd_lock_and_ref(fdll);
-    fdll->o->set_read_handler(fdll->o, fdll->fd, false);
-    fdll->o->set_except_handler(fdll->o, fdll->fd, fdll->write_enabled);
     if (fdll->in_read || fdll->state == FD_ERR_WAIT ||
 		fdll->state == FD_OPEN_ERR_WAIT)
-	goto out;
+	goto out_disable;
     fdll->in_read = true;
-    fd_unlock(fdll);
 
     if (!fdll->read_data_len) {
-	err = doread(fdll->fd, fdll->read_data, fdll->read_data_size, &count,
+	fd_unlock(fdll);
+	err = doread(fdll->iod, fdll->read_data, fdll->read_data_size, &count,
 		     &auxdata, cb_data);
+	fd_lock(fdll);
 	if (!err) {
 	    fdll->read_data_len = count;
 	    fdll->auxdata = auxdata;
@@ -506,7 +505,6 @@ fd_handle_incoming(struct fd_ll *fdll,
 
     fd_deliver_read_data(fdll, err);
 
-    fd_lock(fdll);
     if (err) {
 	switch(fdll->state) {
 	case FD_IN_OPEN:
@@ -517,8 +515,8 @@ fd_handle_incoming(struct fd_ll *fdll,
 	    break;
 
 	case FD_OPEN:
-	    fdll->o->set_write_handler(fdll->o, fdll->fd, false);
-	    fdll->o->set_except_handler(fdll->o, fdll->fd, false);
+	    fdll->o->set_write_handler(fdll->iod, false);
+	    fdll->o->set_except_handler(fdll->iod, false);
 	    fd_set_state(fdll, FD_ERR_WAIT);
 	    break;
 
@@ -533,15 +531,18 @@ fd_handle_incoming(struct fd_ll *fdll,
      * if the user is doing their job right, it shouldn't matter.
      */
     if (fdll->state == FD_OPEN && fdll->read_enabled) {
-	fdll->o->set_read_handler(fdll->o, fdll->fd, true);
-	fdll->o->set_except_handler(fdll->o, fdll->fd, true);
+	fdll->o->set_read_handler(fdll->iod, true);
+	fdll->o->set_except_handler(fdll->iod, true);
+    } else {
+    out_disable:
+	fdll->o->set_read_handler(fdll->iod, false);
+	fdll->o->set_except_handler(fdll->iod, fdll->write_enabled);
     }
- out:
     fd_deref_and_unlock(fdll);
 }
 
 void gensio_fd_ll_handle_incoming(struct gensio_ll *ll,
-				  int (*doread)(int fd, void *buf,
+				  int (*doread)(struct gensio_iod *iod, void *buf,
 						gensiods count,
 						gensiods *rcount,
 						const char ***auxdata,
@@ -555,21 +556,19 @@ void gensio_fd_ll_handle_incoming(struct gensio_ll *ll,
 }
 
 static int
-gensio_ll_fd_read(int fd, void *buf, gensiods count, gensiods *rcount,
+gensio_ll_fd_read(struct gensio_iod *iod, void *buf, gensiods count, gensiods *rcount,
 		  const char ***auxdata, void *cb_data)
 {
-    struct fd_ll *fdll = cb_data;
-
-    return gensio_os_read(fdll->o, fd, buf, count, rcount);
+    return iod->f->read(iod, buf, count, rcount);
 }
 
 static void
-fd_read_ready(int fd, void *cbdata)
+fd_read_ready(struct gensio_iod *iod, void *cbdata)
 {
     struct fd_ll *fdll = cbdata;
 
     if (fdll->ops->read_ready) {
-	fdll->ops->read_ready(fdll->handler_data, fdll->fd);
+	fdll->ops->read_ready(fdll->handler_data, fdll->iod);
 	return;
     }
 
@@ -579,56 +578,71 @@ fd_read_ready(int fd, void *cbdata)
 static int fd_setup_handlers(struct fd_ll *fdll);
 
 static void
-fd_handle_write_ready(struct fd_ll *fdll, int fd)
+fd_handle_write_ready(struct fd_ll *fdll, struct gensio_iod *iod)
 {
-    fdll->o->set_write_handler(fdll->o, fd, false);
-    fdll->o->set_except_handler(fdll->o, fd, fdll->read_enabled);
     if (fdll->state == FD_IN_OPEN) {
 	int err;
 
-	err = fdll->ops->check_open(fdll->handler_data, fdll->fd);
-	if (err && fdll->ops->retry_open) {
+	fdll->o->set_write_handler(iod, false);
+	fdll->o->set_except_handler(iod, fdll->read_enabled);
+	err = fdll->ops->check_open(fdll->handler_data, fdll->iod);
+	/*
+	 * The GE_NOMEM check is strange here, but it really has more
+	 * to do with testing.  check_open() is not going to return
+	 * GE_NOMEM unless it's an error trigger failure, and we really
+	 * want to fail in that case or we will get a "error triggered
+	 * but no failure" in the test.
+	 */
+	if (err && err != GE_NOMEM && fdll->ops->retry_open) {
 	    fd_set_state(fdll, FD_IN_OPEN_RETRY);
-	    fdll->o->clear_fd_handlers(fdll->o, fdll->fd);
+	    fdll->o->clear_fd_handlers(fdll->iod);
 	} else {
 	    if (err) {
 		fdll->open_err = err;
 		fd_set_state(fdll, FD_OPEN_ERR_WAIT);
-		fdll->o->clear_fd_handlers(fdll->o, fdll->fd);
+		fdll->o->clear_fd_handlers(fdll->iod);
 	    } else {
 		fd_finish_open(fdll, 0);
 	    }
 	}
-    } else if (fdll->state == FD_OPEN && fdll->write_enabled) {
+    } else if (fdll->state == FD_OPEN && fdll->write_enabled &&
+	       !fdll->in_write) {
+	fdll->in_write = true;
 	fd_unlock(fdll);
 
-	if (fdll->ops->write_ready) {
-	    fdll->ops->write_ready(fdll->handler_data, fdll->fd);
-	    fd_lock(fdll);
-	} else {
+	if (fdll->ops->write_ready)
+	    fdll->ops->write_ready(fdll->handler_data, fdll->iod);
+	else
 	    gensio_fd_ll_callback(fdll->ll, GENSIO_LL_CB_WRITE_READY, 0,
 				  NULL, 0, NULL);
-	    fd_lock(fdll);
-	    if (fdll->state == FD_OPEN && fdll->write_enabled) {
-		fdll->o->set_write_handler(fdll->o, fdll->fd, true);
-		fdll->o->set_except_handler(fdll->o, fdll->fd, true);
-	    }
+	fd_lock(fdll);
+	fdll->in_write = false;
+	if ((fdll->state == FD_OPEN || fdll->state == FD_IN_CLOSE) &&
+		fdll->write_enabled) {
+	    fdll->o->set_write_handler(fdll->iod, true);
+	    fdll->o->set_except_handler(fdll->iod, true);
+	} else {
+	    fdll->o->set_write_handler(iod, false);
+	    fdll->o->set_except_handler(iod, fdll->read_enabled);
 	}
+    } else {
+	fdll->o->set_write_handler(iod, false);
+	fdll->o->set_except_handler(iod, fdll->read_enabled);
     }
 }
 
 static void
-fd_write_ready(int fd, void *cbdata)
+fd_write_ready(struct gensio_iod *iod, void *cbdata)
 {
     struct fd_ll *fdll = cbdata;
 
     fd_lock_and_ref(fdll);
-    fd_handle_write_ready(fdll, fd);
+    fd_handle_write_ready(fdll, iod);
     fd_deref_and_unlock(fdll);
 }
 
 static void
-fd_except_ready(int fd, void *cbdata)
+fd_except_ready(struct gensio_iod *iod, void *cbdata)
 {
     struct fd_ll *fdll = cbdata;
     int rv = 0;
@@ -640,11 +654,11 @@ fd_except_ready(int fd, void *cbdata)
      */
     if (fdll->state == FD_IN_OPEN || fdll->state == FD_IN_OPEN_RETRY) {
 	fd_ref(fdll);
-	fd_handle_write_ready(fdll, fd);
+	fd_handle_write_ready(fdll, iod);
 	fd_deref_and_unlock(fdll);
     } else if (fdll->ops->except_ready) {
 	fd_unlock(fdll);
-	rv = fdll->ops->except_ready(fdll->handler_data, fdll->fd);
+	rv = fdll->ops->except_ready(fdll->handler_data, fdll->iod);
 	if (rv) {
 	    fd_lock(fdll);
 	    goto handle_except_internal;
@@ -653,10 +667,10 @@ fd_except_ready(int fd, void *cbdata)
     handle_except_internal:
 	if (fdll->read_enabled) {
 	    fd_unlock(fdll);
-	    fd_read_ready(fd, fdll);
+	    fd_read_ready(iod, fdll);
 	} else {
 	    if (fdll->write_enabled)
-		fd_handle_write_ready(fdll, fd);
+		fd_handle_write_ready(fdll, iod);
 	    else
 		fdll->deferred_except = true;
 	    fd_unlock(fdll);
@@ -667,8 +681,8 @@ fd_except_ready(int fd, void *cbdata)
 static void
 fd_finish_cleared(struct fd_ll *fdll)
 {
-    if (fdll->fd != -1)
-	gensio_os_close(fdll->o, &fdll->fd);
+    if (fdll->iod)
+	fdll->o->close(&fdll->iod);
     if (fdll->state == FD_OPEN_ERR_WAIT)
 	fdll->deferred_open = true;
     fdll->deferred_close = true;
@@ -680,8 +694,8 @@ gensio_fd_ll_close_now(struct gensio_ll *ll)
 {
     struct fd_ll *fdll = ll_to_fd(ll);
 
-    if (fdll->fd != -1)
-	gensio_os_close(fdll->o, &fdll->fd);
+    if (fdll->iod)
+	fdll->o->close(&fdll->iod);
 }
 
 static void
@@ -690,9 +704,12 @@ fd_check_close(struct fd_ll *fdll)
     gensio_time timeout;
     int err = 0;
 
-    if (fdll->ops->check_close)
-	err = fdll->ops->check_close(fdll->handler_data,
+    if (fdll->ops->check_close) {
+	err = fdll->ops->check_close(fdll->handler_data, fdll->iod,
 				     GENSIO_LL_CLOSE_STATE_DONE, &timeout);
+	if (err != GE_INPROGRESS)
+	    fdll->iod = NULL;
+    }
 
     if (err == GE_INPROGRESS) {
 	fd_ref(fdll);
@@ -713,23 +730,26 @@ fd_close_timeout(struct gensio_timer *t, void *cb_data)
 }
 
 static void
-fd_cleared(int fd, void *cb_data)
+fd_cleared(struct gensio_iod *iod, void *cb_data)
 {
     struct fd_ll *fdll = cb_data;
     int err;
 
     fd_lock_and_ref(fdll);
     if (fdll->state == FD_IN_OPEN_RETRY) {
-	gensio_os_close(fdll->o, &fdll->fd);
-	err = fdll->ops->retry_open(fdll->handler_data, &fdll->fd);
-	if (err == GE_INPROGRESS)
+	fdll->o->close(&fdll->iod);
+	err = fdll->ops->retry_open(fdll->handler_data, &fdll->iod);
+	if (err == GE_INPROGRESS) {
 	    err = fd_setup_handlers(fdll);
+	    if (!err)
+		fd_set_state(fdll, FD_IN_OPEN);
+	}
 	if (err) {
 	    fd_deref(fdll);
 	    fd_finish_open(fdll, err);
 	} else {
-	    fdll->o->set_write_handler(fdll->o, fdll->fd, true);
-	    fdll->o->set_except_handler(fdll->o, fdll->fd, true);
+	    fdll->o->set_write_handler(fdll->iod, true);
+	    fdll->o->set_except_handler(fdll->iod, true);
 	}
     } else {
 	fd_check_close(fdll);
@@ -757,12 +777,12 @@ fd_open(struct gensio_ll *ll, gensio_ll_open_done done, void *open_data)
     fdll->read_data_len = 0;
     fdll->read_data_pos = 0;
 
-    err = fdll->ops->sub_open(fdll->handler_data, &fdll->fd);
+    err = fdll->ops->sub_open(fdll->handler_data, &fdll->iod);
     if (err == GE_INPROGRESS || err == 0) {
 	int err2 = fd_setup_handlers(fdll);
 	if (err2) {
 	    err = err2;
-	    gensio_os_close(fdll->o, &fdll->fd);
+	    fdll->o->close(&fdll->iod);
 	    goto out;
 	}
 
@@ -770,8 +790,8 @@ fd_open(struct gensio_ll *ll, gensio_ll_open_done done, void *open_data)
 	fdll->open_data = open_data;
 	if (err == GE_INPROGRESS) {
 	    fd_set_state(fdll, FD_IN_OPEN);
-	    fdll->o->set_write_handler(fdll->o, fdll->fd, true);
-	    fdll->o->set_except_handler(fdll->o, fdll->fd, true);
+	    fdll->o->set_write_handler(fdll->iod, true);
+	    fdll->o->set_except_handler(fdll->iod, true);
 	} else {
 	    fd_set_state(fdll, FD_OPEN);
 	}
@@ -786,7 +806,7 @@ fd_open(struct gensio_ll *ll, gensio_ll_open_done done, void *open_data)
 static int
 fd_setup_handlers(struct fd_ll *fdll)
 {
-    if (fdll->o->set_fd_handlers(fdll->o, fdll->fd, fdll, fd_read_ready,
+    if (fdll->o->set_fd_handlers(fdll->iod, fdll, fd_read_ready,
 				 fd_write_ready, fd_except_ready,
 				 fd_cleared))
 	return GE_NOMEM;
@@ -850,9 +870,8 @@ fd_set_read_callback_enable(struct gensio_ll *ll, bool enabled)
 	fdll->deferred_read = true;
 	fd_sched_deferred_op(fdll);
     } else {
-	fdll->o->set_read_handler(fdll->o, fdll->fd, enabled);
-	fdll->o->set_except_handler(fdll->o, fdll->fd,
-				    enabled || fdll->write_enabled);
+	fdll->o->set_read_handler(fdll->iod, enabled);
+	fdll->o->set_except_handler(fdll->iod, enabled || fdll->write_enabled);
     }
  out_unlock:
     fd_unlock(fdll);
@@ -867,9 +886,8 @@ fd_set_write_callback_enable(struct gensio_ll *ll, bool enabled)
     fdll->write_enabled = enabled;
     if (fdll->state == FD_OPEN || fdll->state == FD_IN_OPEN ||
 		fdll->state == FD_IN_OPEN_RETRY) {
-	fdll->o->set_write_handler(fdll->o, fdll->fd, enabled);
-	fdll->o->set_except_handler(fdll->o, fdll->fd,
-				    enabled || fdll->read_enabled);
+	fdll->o->set_write_handler(fdll->iod, enabled);
+	fdll->o->set_except_handler(fdll->iod, enabled || fdll->read_enabled);
     } else if (fdll->deferred_except) {
 	fd_sched_deferred_op(fdll);
     }
@@ -910,7 +928,7 @@ static int fd_control(struct gensio_ll *ll, bool get, unsigned int option,
     if (!fdll->ops->control)
 	return GE_NOTSUP;
 
-    return fdll->ops->control(fdll->handler_data, fdll->fd, get, option, data,
+    return fdll->ops->control(fdll->handler_data, fdll->iod, get, option, data,
 			      datalen);
 }
 
@@ -920,8 +938,8 @@ static void fd_disable(struct gensio_ll *ll)
 
     fd_set_state(fdll, FD_CLOSED);
     fd_deref(fdll);
-    fdll->o->clear_fd_handlers_norpt(fdll->o, fdll->fd);
-    gensio_os_close(fdll->o, &fdll->fd);
+    fdll->o->clear_fd_handlers_norpt(fdll->iod);
+    fdll->o->close(&fdll->iod);
 }
 
 static int
@@ -931,17 +949,17 @@ gensio_ll_fd_func(struct gensio_ll *ll, int op, gensiods *count,
 {
     switch (op) {
     case GENSIO_LL_FUNC_SET_CALLBACK:
-	fd_set_callbacks(ll, cbuf, buf);
+	fd_set_callbacks(ll, (void *) cbuf, buf);
 	return 0;
 
     case GENSIO_LL_FUNC_WRITE_SG:
 	return fd_write(ll, count, cbuf, buflen, auxdata);
 
     case GENSIO_LL_FUNC_OPEN:
-	return fd_open(ll, cbuf, buf);
+	return fd_open(ll, (void *) cbuf, buf);
 
     case GENSIO_LL_FUNC_CLOSE:
-	return fd_close(ll, cbuf, buf);
+	return fd_close(ll, (void *) cbuf, buf);
 
     case GENSIO_LL_FUNC_SET_READ_CALLBACK:
 	fd_set_read_callback_enable(ll, buflen);
@@ -977,7 +995,7 @@ gensio_fd_ll_get_handler_data(struct gensio_ll *ll)
 
 struct gensio_ll *
 fd_gensio_ll_alloc(struct gensio_os_funcs *o,
-		   int fd,
+		   struct gensio_iod *iod,
 		   const struct gensio_fd_ll_ops *ops,
 		   void *handler_data,
 		   gensiods max_read_size,
@@ -991,10 +1009,10 @@ fd_gensio_ll_alloc(struct gensio_os_funcs *o,
 
     fdll->o = o;
     fdll->handler_data = handler_data;
-    fdll->fd = fd;
+    fdll->iod = iod;
     fdll->refcount = 1;
     fdll->write_only = write_only;
-    if (fd == -1) {
+    if (!iod) {
 	fd_set_state(fdll, FD_CLOSED);
     } else {
 	fd_set_state(fdll, FD_OPEN);
@@ -1024,7 +1042,7 @@ fd_gensio_ll_alloc(struct gensio_os_funcs *o,
     if (!fdll->ll)
 	goto out_nomem;
 
-    if (fd != -1) {
+    if (iod) {
 	int err = fd_setup_handlers(fdll);
 	if (err)
 	    goto out_nomem;
