@@ -51,7 +51,9 @@ gio_mem_free(void *data)
 
 struct os_hnd_fd_id_s
 {
+    struct gensio_os_funcs *o;
     int             fd;
+    struct gensio_iod *iod;
     void            *cb_data;
     os_data_ready_t data_ready;
     os_data_ready_t write_ready;
@@ -61,36 +63,37 @@ struct os_hnd_fd_id_s
 };
 
 static void
-fd_read_handler(int fd, void *data)
+fd_read_handler(struct gensio_iod *iod, void *data)
 {
     os_hnd_fd_id_t *fd_data = (os_hnd_fd_id_t *) data;
 
-    fd_data->data_ready(fd, fd_data->cb_data, fd_data);
+    fd_data->data_ready(fd_data->fd, fd_data->cb_data, fd_data);
 }
 
 static void
-fd_write_handler(int fd, void *data)
+fd_write_handler(struct gensio_iod *iod, void *data)
 {
     os_hnd_fd_id_t *fd_data = (os_hnd_fd_id_t *) data;
 
-    fd_data->write_ready(fd, fd_data->cb_data, fd_data);
+    fd_data->write_ready(fd_data->fd, fd_data->cb_data, fd_data);
 }
 
 static void
-fd_except_handler(int fd, void *data)
+fd_except_handler(struct gensio_iod *iod, void *data)
 {
     os_hnd_fd_id_t *fd_data = (os_hnd_fd_id_t *) data;
 
-    fd_data->except_ready(fd, fd_data->cb_data, fd_data);
+    fd_data->except_ready(fd_data->fd, fd_data->cb_data, fd_data);
 }
 
 static void
-free_fd_data(int fd, void *data)
+free_fd_data(struct gensio_iod *iod, void *data)
 {
     os_hnd_fd_id_t *fd_data = data;
 
+    fd_data->o->release_iod(fd_data->iod);
     if (fd_data->freed)
-        fd_data->freed(fd, fd_data->cb_data);
+        fd_data->freed(fd_data->fd, fd_data->cb_data);
     free(data);
 }
 
@@ -110,22 +113,30 @@ gio_add_fd_to_wait_for(os_handler_t       *handler,
 
     fd_data = malloc(sizeof(*fd_data));
     if (!fd_data)
-	return GE_NOMEM;
+	return ENOMEM;
+    fd_data->o = o;
 
+    rv = o->add_iod(o, GENSIO_IOD_SOCKET, fd, &fd_data->iod);
+    if (rv) {
+	free(fd_data);
+	return rv;
+    }
     fd_data->fd = fd;
     fd_data->cb_data = cb_data;
     fd_data->data_ready = data_ready;
     fd_data->handler = handler;
     fd_data->freed = freed;
-    rv = o->set_fd_handlers(o, fd, fd_data, fd_read_handler, fd_write_handler,
+    rv = o->set_fd_handlers(fd_data->iod, fd_data,
+			    fd_read_handler, fd_write_handler,
 			    fd_except_handler, free_fd_data);
     if (rv) {
+	o->release_iod(fd_data->iod);
 	free(fd_data);
 	return rv;
     }
-    o->set_write_handler(o, fd, false);
-    o->set_except_handler(o, fd, false);
-    o->set_read_handler(o, fd, true);
+    o->set_write_handler(fd_data->iod, false);
+    o->set_except_handler(fd_data->iod, false);
+    o->set_read_handler(fd_data->iod, true);
 
     *id = fd_data;
     return 0;
@@ -138,8 +149,8 @@ gio_remove_fd_to_wait_for(os_handler_t   *handler,
     struct igensio_info *info = handler->internal_data;
     struct gensio_os_funcs *o = info->o;
 
-    o->set_read_handler(o, id->fd, false);
-    o->clear_fd_handlers(o, id->fd);
+    o->set_read_handler(id->iod, false);
+    o->clear_fd_handlers(id->iod);
     return 0;
 }
 
@@ -181,12 +192,12 @@ gio_alloc_timer(os_handler_t      *handler,
 
     timer = malloc(sizeof(*timer));
     if (!timer)
-	return GE_NOMEM;
+	return ENOMEM;
 
     timer->lock = o->alloc_lock(o);
     if (!timer->lock) {
 	free(timer);
-	return GE_NOMEM;
+	return ENOMEM;
     }
 
     timer->running = false;
@@ -197,7 +208,7 @@ gio_alloc_timer(os_handler_t      *handler,
     if (!timer->timer) {
 	o->free_lock(timer->lock);
 	free(timer);
-	return GE_NOMEM;
+	return ENOMEM;
     }
 
     *rtimer = timer;
@@ -231,7 +242,7 @@ gio_start_timer(os_handler_t      *handler,
 
     o->lock(timer->lock);
     if (timer->running) {
-	rv = GE_NOTREADY;
+	rv = EAGAIN;
 	goto out_unlock;
     }
 
@@ -265,7 +276,7 @@ gio_stop_timer(os_handler_t *handler,
 	timer->running = 0;
 	o->stop_timer(timer->timer);
     } else {
-	rv = GE_TIMEDOUT;
+	rv = ETIMEDOUT;
     }
     o->unlock(timer->lock);
 
@@ -287,12 +298,12 @@ gio_create_lock(os_handler_t  *handler,
 
     lock = malloc(sizeof(*lock));
     if (!lock)
-	return GE_NOMEM;
+	return ENOMEM;
 
     lock->lock = o->alloc_lock(o);
     if (!lock->lock) {
 	free(lock);
-	return GE_NOMEM;
+	return ENOMEM;
     }
 
     *rlock = lock;
@@ -341,7 +352,7 @@ gio_get_random(os_handler_t  *handler,
 {
     struct igensio_info *info = handler->internal_data;
 
-    return gensio_os_get_random(info->o, data, len);
+    return info->o->get_random(info->o, data, len);
 }
 
 static void
@@ -429,9 +440,9 @@ gio_set_fd_enables(os_handler_t *handler, os_hnd_fd_id_t *id,
     struct igensio_info *info = handler->internal_data;
     struct gensio_os_funcs *o = info->o;
 
-    o->set_read_handler(o, id->fd, read);
-    o->set_write_handler(o, id->fd, write);
-    o->set_except_handler(o, id->fd, except);
+    o->set_read_handler(id->iod, read);
+    o->set_write_handler(id->iod, write);
+    o->set_except_handler(id->iod, except);
     return 0;
 }
 
@@ -517,9 +528,24 @@ enum sol_state {
     SOL_IN_CLOSE
 };
 
+struct sol_ll;
+
+struct sol_op_done {
+    struct sol_ll *solll;
+    bool started;
+    bool use_runner;
+    sergensio_done cb;
+    int done_val;
+    int val;
+    void *cb_data;
+    int (*func)(ipmi_sol_conn_t *, int, ipmi_sol_transmit_complete_cb, void *);
+    struct sol_op_done *next;
+};
+
 struct sol_ll {
     struct gensio_ll *ll;
     struct gensio_os_funcs *o;
+    struct sergensio *sio;
 
     struct gensio_lock *lock;
 
@@ -584,6 +610,31 @@ struct sol_ll {
     unsigned int ack_retries;
     bool deassert_CTS_DCD_DSR_on_connect;
     bool shared_serial_alert_behavior;
+
+    /* Pending transmit done handling. */
+    bool xmit_dones_pending;
+    struct gensio_list xmit_dones;
+    struct gensio_lock *xmit_done_lock;
+    struct gensio_runner *xmit_done_runner;
+
+    /*
+     * We got a flush/break while one was still pending, do these when
+     * it finishes.
+     */
+    int pending_flush;
+    int pending_break;
+
+    struct sol_op_done *cts_done;
+    struct sol_op_done *dcd_dsr_done;
+    struct sol_op_done *ri_done;
+};
+
+/* Used to hold information about pending transmits. */
+struct sol_tc {
+    unsigned int size;
+    struct sol_ll *solll;
+    int err;
+    struct gensio_link link;
 };
 
 os_handler_t *gensio_os_handler;
@@ -610,10 +661,20 @@ sol_ref(struct sol_ll *solll)
 
 static void sol_finish_free(struct sol_ll *solll)
 {
+    if (solll->sol) {
+	ipmi_sol_close(solll->sol);
+	ipmi_sol_free(solll->sol);
+    }
+    if (solll->ipmi)
+	solll->ipmi->close_connection(solll->ipmi);
     if (solll->ll)
 	gensio_ll_free_data(solll->ll);
     if (solll->lock)
 	solll->o->free_lock(solll->lock);
+    if (solll->xmit_done_lock)
+	solll->o->free_lock(solll->xmit_done_lock);
+    if (solll->xmit_done_runner)
+	solll->o->free_runner(solll->xmit_done_runner);
     if (solll->read_data.buf)
 	solll->o->free(solll->o, solll->read_data.buf);
     if (solll->deferred_op_runner)
@@ -651,6 +712,8 @@ static int sol_xlat_ipmi_err(struct gensio_os_funcs *o, int err)
 	    err = GE_COMMERR;
 	else if (err == IPMI_SOL_DEACTIVATED)
 	    err = GE_HOSTDOWN;
+	else
+	    err = GE_COMMERR;
     } else if (IPMI_IS_RMCPP_ERR(err)) {
 	err = IPMI_GET_RMCPP_ERR(err);
 	if (err == IPMI_RMCPP_INVALID_PAYLOAD_TYPE)
@@ -718,6 +781,9 @@ check_for_write_ready(struct sol_ll *solll)
     }
 }
 
+static void sol_op_done(struct sol_ll *solll, int err,
+			struct sol_op_done **op_done);
+
 static void
 sol_deferred_op(struct gensio_runner *runner, void *cbdata)
 {
@@ -727,6 +793,13 @@ sol_deferred_op(struct gensio_runner *runner, void *cbdata)
     while (solll->deferred_op_pending) {
 	solll->deferred_op_pending = false;
 
+	if (solll->cts_done && solll->cts_done->use_runner)
+	    sol_op_done(solll, 0, &solll->cts_done);
+	if (solll->dcd_dsr_done && solll->dcd_dsr_done->use_runner)
+	    sol_op_done(solll, 0, &solll->dcd_dsr_done);
+	if (solll->ri_done && solll->ri_done->use_runner)
+	    sol_op_done(solll, 0, &solll->ri_done);
+
 	while (solll->deferred_read) {
 	    solll->deferred_read = false;
 	    check_for_read_delivery(solll);
@@ -761,13 +834,65 @@ sol_set_callbacks(struct gensio_ll *ll, gensio_ll_cb cb, void *cb_data)
     solll->cb_data = cb_data;
 }
 
-struct sol_tc {
-    unsigned int size;
-    struct sol_ll *solll;
-};
-
 static void connection_closed(ipmi_con_t *ipmi, void *cb_data);
 
+static void
+handle_xmit_dones(struct gensio_runner *runner, void *cbdata)
+{
+    struct sol_ll *solll = cbdata;
+    struct gensio_os_funcs *o = solll->o;
+    unsigned int deref_count = 0;
+
+    sol_lock(solll);
+    o->lock(solll->xmit_done_lock);
+    solll->xmit_dones_pending = false;
+    while (!gensio_list_empty(&solll->xmit_dones)) {
+	struct gensio_link *l = gensio_list_first(&solll->xmit_dones);
+	struct sol_tc *tc = gensio_container_of(l, struct sol_tc, link);
+
+	gensio_list_rm(&solll->xmit_dones, l);
+	o->unlock(solll->xmit_done_lock);
+
+	if (tc->err && solll->state != SOL_IN_CLOSE) {
+	    solll->read_err = tc->err;
+	    check_for_read_delivery(solll);
+	} else {
+	    solll->write_outstanding -= tc->size;
+	    if (solll->state == SOL_IN_CLOSE) {
+		if (solll->write_outstanding == 0) {
+		    tc->err = ipmi_sol_close(solll->sol);
+		    if (tc->err)
+			tc->err = solll->ipmi->close_connection_done(
+					     solll->ipmi,
+					     connection_closed,
+					     solll);
+		    if (tc->err) {
+			solll->state = SOL_CLOSED;
+			solll->ipmi = NULL;
+			if (solll->close_done)
+			    solll->close_done(solll->cb_data, solll->open_data);
+		    }
+		}
+	    } else {
+		check_for_write_ready(solll);
+	    }
+	}
+	o->free(o, tc);
+	deref_count++;
+
+	o->lock(solll->xmit_done_lock);
+    }
+    o->unlock(solll->xmit_done_lock);
+
+    if (deref_count >= 1) {
+	assert(solll->refcount >= deref_count);
+	solll->refcount -= deref_count - 1;
+	sol_deref_and_unlock(solll);
+    } else {
+	sol_unlock(solll);
+    }
+}
+
 static void
 transmit_complete(ipmi_sol_conn_t *conn,
 		  int             err,
@@ -775,36 +900,24 @@ transmit_complete(ipmi_sol_conn_t *conn,
 {
     struct sol_tc *tc = cb_data;
     struct sol_ll *solll = tc->solll;
+    struct gensio_os_funcs *o = solll->o;
 
     if (err)
-	err = sol_xlat_ipmi_err(solll->o, err);
+	err = sol_xlat_ipmi_err(o, err);
+    tc->err = err;
 
-    sol_lock(solll);
-    if (err && solll->state != SOL_IN_CLOSE) {
-	solll->read_err = err;
-	check_for_read_delivery(solll);
-    } else {
-	solll->write_outstanding -= tc->size;
-	if (solll->state == SOL_IN_CLOSE) {
-	    if (solll->write_outstanding == 0) {
-		err = ipmi_sol_close(solll->sol);
-		if (err)
-		    err = solll->ipmi->close_connection_done(solll->ipmi,
-							     connection_closed,
-							     solll);
-		if (err) {
-		    solll->state = SOL_CLOSED;
-		    solll->ipmi = NULL;
-		    if (solll->close_done)
-			solll->close_done(solll->cb_data, solll->open_data);
-		}
-	    }
-	} else {
-	    check_for_write_ready(solll);
-	}
+    /*
+     * Unfortunately, OpenIPMI isn't quite as nice as gensio, you can
+     * get callbacks from user function calls.  So we need to run the
+     * transmit complete handling in a runner.
+     */
+    o->lock(solll->xmit_done_lock);
+    gensio_list_add_tail(&solll->xmit_dones, &tc->link);
+    if (!solll->xmit_dones_pending) {
+	solll->xmit_dones_pending = true;
+	o->run(solll->xmit_done_runner);
     }
-    solll->o->free(solll->o, tc);
-    sol_deref_and_unlock(solll);
+    o->unlock(solll->xmit_done_lock);
 }
 
 static int
@@ -829,19 +942,16 @@ sol_write(struct gensio_ll *ll, gensiods *rcount,
 	total_write += sg[i].buflen;
     if (total_write > left)
 	total_write = left;
-    if (total_write == 0)
+    if (total_write == 0) {
+	pos = 0;
 	goto out_finish;
+    }
 
     buf = solll->o->zalloc(solll->o, total_write);
     if (!buf) {
 	err = GE_NOMEM;
 	goto out_unlock;
     }
-    tc = solll->o->zalloc(solll->o, sizeof(*tc));
-    if (!tc) {
-	err = GE_NOMEM;
-	goto out_unlock;
-    }
     for (i = 0; i < sglen; i++) {
 	if (sg[i].buflen >= total_write - pos) {
 	    memcpy(buf + pos, sg[i].buf, total_write - pos);
@@ -852,21 +962,45 @@ sol_write(struct gensio_ll *ll, gensiods *rcount,
 	}
     }
 
-    tc->size = total_write;
-    tc->solll = solll;
-    err = ipmi_sol_write(solll->sol, buf, total_write, transmit_complete, tc);
-    if (err) {
-	err = sol_xlat_ipmi_err(solll->o, err);
-	free(tc);
-	goto out_unlock;
-    } else {
-	solll->write_outstanding += total_write;
-	sol_ref(solll);
+    pos = 0;
+    while (pos < total_write) {
+	tc = solll->o->zalloc(solll->o, sizeof(*tc));
+	if (!tc) {
+	    if (pos == 0) {
+		/* Nothing transmitted, return an error. */
+		err = GE_NOMEM;
+		goto out_unlock;
+	    }
+	    goto out_finish;
+	}
+	if (total_write - pos > 255)
+	    tc->size = 255;
+	else
+	    tc->size = total_write - pos;
+	tc->solll = solll;
+	err = ipmi_sol_write(solll->sol, buf + pos, tc->size,
+			     transmit_complete, tc);
+	if (err) {
+	    solll->o->free(solll->o, tc);
+	    if (pos == 0 && err != EAGAIN) {
+		/*
+		 * Nothing transmitted and it's not full buffers,
+		 * return an error.
+		 */
+		err = sol_xlat_ipmi_err(solll->o, err);
+		goto out_unlock;
+	    }
+	    err = 0;
+	    goto out_finish;
+	} else {
+	    solll->write_outstanding += tc->size;
+	    sol_ref(solll);
+	    pos += tc->size;
+	}
     }
-    
  out_finish:
     if (rcount)
-	*rcount = total_write;
+	*rcount = pos;
  out_unlock:
     if (buf)
 	solll->o->free(solll->o, buf);
@@ -1010,13 +1144,26 @@ conn_changed(ipmi_con_t   *ipmi,
     if (solll->state == SOL_IN_OPEN || solll->state == SOL_IN_SOL_OPEN) {
 	if (any_port_up && solll->state == SOL_IN_OPEN) {
 	    solll->state = SOL_IN_SOL_OPEN;
-	    sol_unlock(solll);
-	    ipmi_sol_open(solll->sol);
-	    return;
-	} else if (!any_port_up && (err || solll->read_err)) {
+	    err = ipmi_sol_open(solll->sol);
+	    if (!err) {
+		sol_unlock(solll);
+		return;
+	    }
+	    any_port_up = 0;
+	    err = sol_xlat_ipmi_err(solll->o, err);
+	}
+	if (!any_port_up && (err || solll->read_err)) {
 	    solll->state = SOL_CLOSED;
 	    if (solll->read_err)
 		err = solll->read_err; /* Prefer the first error we got. */
+	    if (solll->sol) {
+		ipmi_sol_free(solll->sol);
+		solll->sol = NULL;
+	    }
+	    if (solll->ipmi) {
+		solll->ipmi->close_connection(solll->ipmi);
+		solll->ipmi = NULL;
+	    }
 	    sol_unlock(solll);
 	    solll->open_done(solll->cb_data, err, solll->open_data);
 	    return;
@@ -1030,8 +1177,10 @@ conn_changed(ipmi_con_t   *ipmi,
 	}
     } else if (err) {
 	solll->read_err = err;
+	check_for_read_delivery(solll);
     } else if (!any_port_up) {
 	solll->read_err = GE_NOTREADY;
+	check_for_read_delivery(solll);
     }
 	
  out_unlock:
@@ -1061,7 +1210,7 @@ sol_open(struct gensio_ll *ll, gensio_ll_open_done done, void *open_data)
     err = ipmi_args_setup_con(solll->args, gensio_os_handler, NULL,
 			      &solll->ipmi);
     if (err)
-	return err;
+	goto out_unlock;
 
     err = ipmi_sol_create(solll->ipmi, &solll->sol);
     if (err)
@@ -1109,7 +1258,9 @@ sol_open(struct gensio_ll *ll, gensio_ll_open_done done, void *open_data)
     solll->open_done = done;
     solll->open_data = open_data;
 
-    solll->ipmi->start_con(solll->ipmi);
+    err = solll->ipmi->start_con(solll->ipmi);
+    if (err)
+	goto out_err;
 
     sol_unlock(solll);
     return GE_INPROGRESS;
@@ -1126,6 +1277,8 @@ sol_open(struct gensio_ll *ll, gensio_ll_open_done done, void *open_data)
     }
  out_unlock:
     sol_unlock(solll);
+    if (err)
+	err = sol_xlat_ipmi_err(solll->o, err);
     return err;
 }
 
@@ -1220,18 +1373,29 @@ static void sol_disable(struct gensio_ll *ll)
     }
 }
 
-static int sol_control(struct gensio_ll *ll, bool get, unsigned int option,
-		       char *data, gensiods *datalen)
-{
-    if (option != GENSIO_CONTROL_RADDR)
-	return GE_NOTSUP;
-    if (!get)
-	return GE_NOTSUP;
-    if (strtoul(data, NULL, 0) > 0)
-	return GE_NOTFOUND;
+static int ipmisol_do_break(struct gensio_ll *ll);
 
-    *datalen = gensio_pos_snprintf(data, *datalen, NULL, "ipmisol");
-    return 0;
+static int
+sol_control(struct gensio_ll *ll, bool get, unsigned int option,
+	    char *data, gensiods *datalen)
+{
+    switch(option) {
+    case GENSIO_CONTROL_RADDR:
+	if (!get)
+	    return GE_NOTSUP;
+	if (strtoul(data, NULL, 0) > 0)
+	    return GE_NOTFOUND;
+	*datalen = gensio_pos_snprintf(data, *datalen, NULL, "ipmisol");
+	return 0;
+
+    case GENSIO_CONTROL_SEND_BREAK:
+	if (get)
+	    return GE_NOTSUP;
+	return ipmisol_do_break(ll);
+
+    default:
+	return GE_NOTSUP;
+    }
 }
 
 static int
@@ -1277,6 +1441,296 @@ gensio_ll_sol_func(struct gensio_ll *ll, int op, gensiods *count,
     }
 }
 
+static void
+ipmisol_flush_done(ipmi_sol_conn_t *conn, int error,
+		   int queue_selectors_flushed, void *cb_data)
+{
+    struct sol_ll *solll = cb_data;
+    int rv;
+
+    sol_lock(solll);
+    if (solll->state == SOL_OPEN && solll->pending_break) {
+	/* A flush came in while one was pending, do it. */
+	rv = ipmi_sol_flush(solll->sol, solll->pending_break,
+			    ipmisol_flush_done, solll);
+	if (!rv) {
+	    solll->pending_flush = 0;
+	    sol_ref(solll);
+	}
+    }
+    sol_deref_and_unlock(solll);
+}
+
+static int ipmisol_do_flush(struct gensio_ll *ll, int val)
+{
+    struct sol_ll *solll = ll_to_sol(ll);
+    int rv;
+
+    switch(val) {
+    case SERGIO_FLUSH_RCV_BUFFER:
+	val = IPMI_SOL_BMC_RECEIVE_QUEUE;
+	break;
+
+    case SERGIO_FLUSH_XMIT_BUFFER:
+	val = IPMI_SOL_BMC_TRANSMIT_QUEUE;
+	break;
+
+    case SERGIO_FLUSH_RCV_XMIT_BUFFERS:
+	return GE_NOTSUP;
+
+    default:
+	return GE_INVAL;
+    }
+
+    sol_lock(solll);
+    rv = ipmi_sol_flush(solll->sol, val, ipmisol_flush_done, solll);
+    if (!rv) {
+	sol_ref(solll);
+    } else if (rv == EAGAIN) {
+	solll->pending_flush |= val;
+    } else if (rv == IPMI_SOL_ERR_VAL(IPMI_SOL_UNCONFIRMABLE_OPERATION)) {
+	rv = 0; /* Operation done, but won't get a callback. */
+    } else {
+	rv = sol_xlat_ipmi_err(solll->o, rv);
+    }
+    sol_unlock(solll);
+
+    return rv;
+}
+
+static void
+ipmisol_break_done(ipmi_sol_conn_t *conn, int err, void *cb_data)
+{
+    struct sol_ll *solll = cb_data;
+    int rv;
+
+    sol_lock(solll);
+    if (solll->state == SOL_OPEN && solll->pending_break) {
+	/* A flush came in while one was pending, do it. */
+	rv = ipmi_sol_send_break(solll->sol, ipmisol_break_done, solll);
+	if (!rv) {
+	    solll->pending_break = 0;
+	    sol_ref(solll);
+	}
+    }
+    sol_deref_and_unlock(solll);
+}
+
+static int ipmisol_do_break(struct gensio_ll *ll)
+{
+    struct sol_ll *solll = ll_to_sol(ll);
+    int rv;
+
+    sol_lock(solll);
+    rv = ipmi_sol_send_break(solll->sol, ipmisol_break_done, solll);
+    if (!rv) {
+	sol_ref(solll);
+    } else if (rv == EAGAIN) {
+	solll->pending_break = 1;
+    } else if (rv == IPMI_SOL_ERR_VAL(IPMI_SOL_UNCONFIRMABLE_OPERATION)) {
+	rv = 0; /* Operation done, but won't get a callback. */
+    } else {
+	rv = sol_xlat_ipmi_err(solll->o, rv);
+    }
+    sol_unlock(solll);
+
+    return rv;
+}
+
+static void
+ipmisol_op_done(ipmi_sol_conn_t *conn, int err, void *icb_data)
+{
+    struct sol_op_done **op_done = icb_data;
+    struct sol_ll *solll = (*op_done)->solll;
+
+    sol_lock(solll);
+    sol_op_done(solll, err, op_done);
+    sol_deref_and_unlock(solll);
+}
+
+static int
+sol_start_op(struct sol_ll *solll, struct sol_op_done *op,
+	     struct sol_op_done **op_done)
+{
+    int rv;
+
+    rv = op->func(solll->sol, op->val, ipmisol_op_done, op_done);
+    switch (rv) {
+    case 0:
+	op->started = true;
+	sol_ref(solll);
+	break;
+
+    case IPMI_SOL_ERR_VAL(IPMI_SOL_UNCONFIRMABLE_OPERATION):
+	op->started = true;
+	op->use_runner = true;
+	rv = 0; /* Operation done, but won't get a callback. */
+	/* Schedule the callback in the runner. */
+	sol_sched_deferred_op(solll);
+	break;
+
+    case EAGAIN:
+	/* Should not happen. */
+	rv = GE_INUSE;
+	break;
+
+    default:
+	rv = sol_xlat_ipmi_err(solll->o, rv);
+	break;
+    }
+
+    return rv;
+}
+
+static void
+sol_op_done(struct sol_ll *solll, int err, struct sol_op_done **op_done)
+{
+    struct gensio_os_funcs *o = solll->o;
+    struct sol_op_done *op = *op_done;
+    sergensio_done cb;
+    void *cb_data;
+    int val;
+
+ restart:
+    if (err)
+	err = sol_xlat_ipmi_err(solll->o, err);
+
+    cb = op->cb;
+    cb_data = op->cb_data;
+    val = op->done_val;
+    *op_done = op->next;
+    o->free(o, op);
+    if (cb) {
+	sol_unlock(solll);
+	cb(solll->sio, err, val, cb_data);
+	sol_lock(solll);
+    }
+    op = *op_done;
+    if (op && !op->started) {
+	err = sol_start_op(solll, op, op_done);
+	if (err)
+	    goto restart;
+    }
+}
+
+static int
+sol_do_op(struct sol_ll *solll, struct sol_op_done **op_done,
+	  int (*func)(ipmi_sol_conn_t *, int, ipmi_sol_transmit_complete_cb,
+		      void *),
+	  int val, int done_val, sergensio_done done, void *cb_data)
+{
+    struct gensio_os_funcs *o = solll->o;
+    struct sol_op_done *op, *op2;
+    int rv = 0;
+
+    op = o->zalloc(o, sizeof(*op));
+    if (!op)
+	return GE_NOMEM;
+
+    op->use_runner = false;
+    op->solll = solll;
+    op->cb = done;
+    op->cb_data = cb_data;
+    op->val = val;
+    op->done_val = done_val;
+    op->func = func;
+    op->next = NULL;
+
+    if (*op_done) {
+	/* Something already in progress, just queue it. */
+	op2 = *op_done;
+	while (op2->next)
+	    op2 = op2->next;
+	op2->next = op;
+    } else {
+	rv = sol_start_op(solll, op, op_done);
+	if (rv)
+	    o->free(o, op);
+	else
+	    *op_done = op;
+    }
+
+    return rv;
+}
+
+static int ipmisol_do_cts(struct gensio_ll *ll, int ival,
+			  sergensio_done done, void *cb_data)
+{
+    struct sol_ll *solll = ll_to_sol(ll);
+    int rv, val;
+
+    sol_lock(solll);
+    switch (ival) {
+    case SERGENSIO_CTS_AUTO:
+	val = 1;
+	break;
+    case SERGENSIO_CTS_OFF:
+	val = 0;
+	break;
+    default:
+	rv = GE_INVAL;
+	goto out_unlock;
+    }
+    rv = sol_do_op(solll, &solll->cts_done, ipmi_sol_set_CTS_assertable,
+		   val, ival, done, cb_data);
+ out_unlock:
+    sol_unlock(solll);
+
+    return rv;
+}
+
+static int ipmisol_do_dcd_dsr(struct gensio_ll *ll, int ival,
+			      sergensio_done done, void *cb_data)
+{
+    struct sol_ll *solll = ll_to_sol(ll);
+    int rv, val;
+
+    sol_lock(solll);
+    switch (ival) {
+    case SERGENSIO_DCD_DSR_ON:
+	val = 1;
+	break;
+    case SERGENSIO_DCD_DSR_OFF:
+	val = 0;
+	break;
+    default:
+	rv = GE_INVAL;
+	goto out_unlock;
+    }
+    rv = sol_do_op(solll, &solll->dcd_dsr_done, ipmi_sol_set_DCD_DSR_asserted,
+		   val, ival, done, cb_data);
+ out_unlock:
+    sol_unlock(solll);
+
+    return rv;
+}
+
+static int ipmisol_do_ri(struct gensio_ll *ll, int ival,
+			  sergensio_done done, void *cb_data)
+{
+    struct sol_ll *solll = ll_to_sol(ll);
+    int rv, val;
+
+    sol_lock(solll);
+    switch (ival) {
+    case SERGENSIO_RI_ON:
+	val = 1;
+	break;
+    case SERGENSIO_RI_OFF:
+	val = 0;
+	break;
+    default:
+	rv = GE_INVAL;
+	goto out_unlock;
+    }
+    rv = sol_do_op(solll, &solll->ri_done, ipmi_sol_set_RI_asserted,
+		   val, ival, done, cb_data);
+ out_unlock:
+    sol_unlock(solll);
+
+    return rv;
+}
+
 static int
 ipmisol_ser_ops(struct gensio_ll *ll, int op,
 		int val, char *buf,
@@ -1284,10 +1738,19 @@ ipmisol_ser_ops(struct gensio_ll *ll, int op,
 {
     switch (op) {
     case SERGENSIO_FUNC_FLUSH:
-    break;
+	return ipmisol_do_flush(ll, val);
 
     case SERGENSIO_FUNC_SEND_BREAK:
-    break;
+	return ipmisol_do_break(ll);
+
+    case SERGENSIO_FUNC_CTS:
+	return ipmisol_do_cts(ll, val, done, cb_data);
+
+    case SERGENSIO_FUNC_DCD_DSR:
+	return ipmisol_do_dcd_dsr(ll, val, done, cb_data);
+
+    case SERGENSIO_FUNC_RI:
+	return ipmisol_do_ri(ll, val, done, cb_data);
 
     /* You really can't set much on a SOL connection once it's up. */
     case SERGENSIO_FUNC_BAUD:
@@ -1306,8 +1769,6 @@ ipmisol_ser_ops(struct gensio_ll *ll, int op,
     default:
 	return GE_NOTSUP;
     }
-
-    return 0;
 }
 
 static int
@@ -1375,7 +1836,7 @@ sol_get_defaults(struct sol_ll *solll)
 			     GENSIO_DEFAULT_INT, NULL, &ival);
     if (err)
 	return err;
-    ival = solll->shared_serial_alert_behavior;
+    solll->shared_serial_alert_behavior = ival;
     err = gensio_get_default(o, "sol", "deassert-CTS-DCD-DSR-on-connect", false,
 			     GENSIO_DEFAULT_BOOL, NULL, &ival);
     if (err)
@@ -1399,16 +1860,12 @@ sol_process_parm(struct sol_ll *solll, char *arg)
     } else if (strncmp(arg, "115200", 6) == 0) {
 	solll->speed = IPMI_SOL_BIT_RATE_115200;
     } else if (gensio_check_keybool(arg, "nobreak", &solll->disablebreak) > 0) {
-	;
     } else if (gensio_check_keybool(arg, "authenticated",
-				    &solll->authenticated) == 0) {
-	 ;
+				    &solll->authenticated) > 0) {
     } else if (gensio_check_keybool(arg, "encrypted", &solll->encrypted) > 0) {
-	 ;
     } else if (gensio_check_keybool(arg, "deassert-CTS-DCD-DSR-on-connect",
 				    &solll->deassert_CTS_DCD_DSR_on_connect)
 	       > 0) {
-	 ;
     } else if (strcasecmp(arg, "shared-serial-alert-fail") == 0) {
 	solll->shared_serial_alert_behavior = ipmi_sol_serial_alerts_fail;
     } else if (strcmp(arg, "shared-serial-alert-deferred") == 0) {
@@ -1418,10 +1875,8 @@ sol_process_parm(struct sol_ll *solll, char *arg)
     } else if (gensio_check_keyuint(arg, "ack-timeout",
 				    &solll->ack_timeout) > 0) {
 
-	;
     } else if (gensio_check_keyuint(arg, "ack-retries",
 				    &solll->ack_retries) > 0) {
-	;
 
     /* The rest of the ones below are deprecated. */
     } else if (strcmp(arg, "-NOBREAK") == 0) {
@@ -1468,7 +1923,8 @@ sol_process_parms(struct sol_ll *solll)
     return 0;
 }
 
-struct gensio_once gensio_ipmi_initialized;
+static struct gensio_once gensio_ipmi_initialized;
+static int ipmi_init_err;
 
 static void
 gensio_ipmi_init(void *cb_data)
@@ -1478,7 +1934,7 @@ gensio_ipmi_init(void *cb_data)
     gensio_os_handler = gio_alloc(o);
     if (!gensio_os_handler)
 	abort();
-    ipmi_init(gensio_os_handler);
+    ipmi_init_err = ipmi_init(gensio_os_handler);
 }
 
 int
@@ -1497,6 +1953,9 @@ ipmisol_gensio_ll_alloc(struct gensio_os_funcs *o,
 
     o->call_once(o, &gensio_ipmi_initialized, gensio_ipmi_init, o);
 
+    if (ipmi_init_err)
+	return sol_xlat_ipmi_err(o, ipmi_init_err);
+
     solll = o->zalloc(o, sizeof(*solll));
     if (!solll)
 	return GE_NOMEM;
@@ -1504,6 +1963,7 @@ ipmisol_gensio_ll_alloc(struct gensio_os_funcs *o,
     solll->o = o;
     solll->refcount = 1;
     solll->state = SOL_CLOSED;
+    solll->last_any_port_up = -1;
 
     solll->devname = gensio_strdup(o, devname);
     if (!solll->devname)
@@ -1546,21 +2006,30 @@ ipmisol_gensio_ll_alloc(struct gensio_os_funcs *o,
     if (!solll->lock)
 	goto out_nomem;
 
+    solll->xmit_done_lock = o->alloc_lock(o);
+    if (!solll->xmit_done_lock)
+	goto out_nomem;
+    gensio_list_init(&solll->xmit_dones);
+
+    solll->xmit_done_runner = o->alloc_runner(o, handle_xmit_dones, solll);
+    if (!solll->xmit_done_runner)
+	goto out_nomem;
+
     solll->read_data.maxsize = max_read_size;
     solll->read_data.buf = o->zalloc(o, max_read_size);
     if (!solll->read_data.buf)
 	goto out_nomem;
 
-    /* Don't set these until here lest a failure call the free operation. */
-    solll->ser_cbs = ser_cbs;
-    solll->ser_cbs_data = ser_cbs_data;
-
     solll->max_write_size = max_write_size;
 
     solll->ll = gensio_ll_alloc_data(o, gensio_ll_sol_func, solll);
     if (!solll->ll)
 	goto out_nomem;
 
+    /* Don't set these until here lest a failure call the free operation. */
+    solll->ser_cbs = ser_cbs;
+    solll->ser_cbs_data = ser_cbs_data;
+
     *rops = ipmisol_ser_ops;
     *rll = solll->ll;
     return 0;
@@ -1572,6 +2041,20 @@ ipmisol_gensio_ll_alloc(struct gensio_os_funcs *o,
     return err;
 }
 
+void
+ipmisol_gensio_ll_set_sio(struct gensio_ll *ll, struct sergensio *sio)
+{
+    struct sol_ll *solll = ll_to_sol(ll);
+
+    solll->sio = sio;
+}
+
+void
+gensio_sol_cleanup_mem(void)
+{
+    ipmi_shutdown();
+}
+
 #else
 
 int
@@ -1587,4 +2070,14 @@ ipmisol_gensio_ll_alloc(struct gensio_os_funcs *o,
     return GE_NOTSUP;
 }
 
+void
+gensio_sol_cleanup_mem(void)
+{
+}
+
+void
+ipmisol_gensio_ll_set_sio(struct gensio_ll *ll, struct sergensio *sio)
+{
+}
+
 #endif
