@@ -481,6 +481,14 @@ struct mux_data {
     struct gensio_list openchans;
     unsigned int opencount;
 
+    /*
+     * Used when doing an error close and we get a normal close in the
+     * mean time.  This will cause the error close to invoke the
+     * normal close.
+     */
+    bool do_normal_close;
+    struct mux_inst *do_normal_close_chan;
+
     /*
      * All the channels in the mux.  The channels are kept in id
      * order.
@@ -526,8 +534,9 @@ i_mux_add_trace(struct mux_data *mux, struct mux_inst *chan,
     else
 	mux->trace_pos = pos + 1;
 }
-#define mux_add_trace(mux, chan, state, new_state, cstate, new_cstate) \
-    i_mux_add_trace(mux, chan, state, new_state, cstsate, new_cstate, __LINE__)
+#define mux_add_trace(mux, chan, new_state, new_cstate) \
+    i_mux_add_trace(mux, chan, mux->state, new_state, chan->state, \
+		    new_cstate, __LINE__)
 
 #define mux_set_state(mux, nstate) do {					\
 	i_mux_add_trace(mux, NULL, mux->state, nstate, 99, 99, __LINE__);\
@@ -542,7 +551,7 @@ i_mux_add_trace(struct mux_data *mux, struct mux_inst *chan,
 
 #else
 #define i_mux_add_trace(mux, chan, state, cstate, new_state, new_cstate, line)
-#define mux_add_trace(mux, chan, state, cstate, new_state, new_cstate)
+#define mux_add_trace(mux, chan, cstate, new_cstate)
 #define mux_set_state i_mux_set_state
 #define muxc_set_state i_muxc_set_state
 #endif
@@ -779,10 +788,17 @@ mux_channel_set_closed(struct mux_inst *chan, gensio_done close_done,
     int err;
 
     muxc_set_state(chan, MUX_INST_CLOSED);
+    chan_deref(chan);
     assert(muxdata->nr_not_closed > 0);
     muxdata->nr_not_closed--;
+    mux_add_trace(muxdata, chan, 150, muxdata->nr_not_closed);
     if (muxdata->nr_not_closed == 0) {
 	/* There are no open instances, shut the mux down. */
+	if (muxdata->state == MUX_IN_CLOSE) {
+	    muxdata->do_normal_close = true;
+	    muxdata->do_normal_close_chan = chan;
+	    return true;
+	}
 	mux_set_state(muxdata, MUX_IN_CLOSE);
 	err = gensio_close(muxdata->child, close_done, close_data);
 	if (!err)
@@ -808,7 +824,13 @@ finish_close(struct mux_inst *chan)
 	close_done(chan->io, chan->close_data);
 	mux_lock(muxdata);
     }
-    chan_deref(chan);
+}
+
+static void
+i_finish_close_close_done(struct mux_inst *chan, struct mux_data *muxdata)
+{
+    mux_set_state(muxdata, MUX_CLOSED);
+    finish_close(chan);
 }
 
 static void
@@ -817,10 +839,9 @@ finish_close_close_done(struct gensio *child, void *close_data)
     struct mux_inst *chan = close_data;
     struct mux_data *muxdata = chan->mux;
 
-    mux_lock(muxdata);
-    finish_close(chan);
-    mux_set_state(muxdata, MUX_CLOSED);
-    mux_deref_and_unlock(muxdata); /* Lose the open ref. */
+    mux_lock_and_ref(muxdata);
+    i_finish_close_close_done(chan, muxdata);
+    mux_deref_and_unlock(muxdata);
 }
 
 static void
@@ -868,8 +889,8 @@ chan_check_send_more(struct mux_inst *chan)
 	return;
     chan->in_write_ready = true;
 
-    /* Need at least 3 bytes to write a message. */
-    while (chan->write_data_len + 3 < chan->max_write_size &&
+    /* Need at least 4 bytes to write a message. */
+    while (chan->max_write_size - chan->write_data_len >= 4 &&
 	   chan->write_ready_enabled && chan->state == MUX_INST_OPEN) {
 	chan_ref(chan);
 	mux_unlock(chan->mux);
@@ -1246,8 +1267,14 @@ muxc_free(struct mux_inst *chan)
     if (--chan->freeref > 0)
 	goto out_unlock;
     switch (chan->state) {
-    case MUX_INST_IN_CLOSE:
     case MUX_INST_IN_REM_CLOSE:
+	/*
+	 * We don't call finish_close() in this case, so we need to do
+	 * the deref here instead.
+	 */
+	chan_deref(chan);
+	/* Fallthrough */
+    case MUX_INST_IN_CLOSE:
     case MUX_INST_IN_CLOSE_FINAL:
     case MUX_INST_IN_OPEN_CLOSE:
 	chan->close_done = NULL;
@@ -1341,7 +1368,6 @@ mux_new_channel(struct mux_data *muxdata, gensio_event cb, void *user_data,
      * through the entries until we find an empty spot.
      */
     if (gensio_list_empty(&muxdata->chans)) {
-	id = 0; /* This is always the automatic channel. */
 	gensio_list_add_tail(&muxdata->chans, &chan->link);
 	/* Note that we do not claim a ref here, there is already one. */
     } else {
@@ -1377,7 +1403,6 @@ mux_new_channel(struct mux_data *muxdata, gensio_event cb, void *user_data,
 	return err;
 
     found:
-	tchan = gensio_container_of(p, struct mux_inst, link);
 	chan->id = id;
 	muxdata->last_id = id;
 	gensio_list_add_next(&muxdata->chans, p, &chan->link);
@@ -1559,6 +1584,8 @@ mux_call_open_done(struct mux_data *muxdata, struct mux_inst *chan, int err)
 	open_done(chan->io, err, open_data);
 	mux_lock(muxdata);
     }
+    if (err)
+	chan_deref(chan);
 }
 
 static void
@@ -1572,7 +1599,6 @@ mux_child_open_done(struct gensio *child, int err, void *open_data)
     if (err) {
 	mux_shutdown_channels(muxdata, err);
 	muxdata->nr_not_closed = 0;
-	mux_deref(muxdata); /* Lose the child open ref. */
     } else if (chan->state != MUX_INST_IN_OPEN) {
 	/* A close was requested, handle it. */	
 	muxc_set_state(chan, MUX_INST_CLOSED);
@@ -1623,6 +1649,8 @@ muxc_open(struct mux_inst *chan, gensio_done_err open_done, void *open_data,
 	muxdata->in_hdr = true;
 	muxdata->hdr_pos = 0;
 	muxdata->hdr_size = 0;
+	muxdata->exit_err = 0;
+	muxdata->do_normal_close = false;
 	muxc_reinit(chan);
 	if (muxdata->is_client) {
 	    if (!chan->in_open_chan) {
@@ -1657,8 +1685,6 @@ muxc_open(struct mux_inst *chan, gensio_done_err open_done, void *open_data,
 	    gensio_set_read_callback_enable(muxdata->child, true);
 	    err = 0;
 	}
-	if (!err)
-	    mux_ref(muxdata); /* Claim the ref for the channel. */
     } else {
 	if (!do_child) {
 	    err = GE_INVAL;
@@ -1750,7 +1776,7 @@ muxc_gensio_handler(struct gensio *io, int func, gensiods *count,
 	return muxc_write(chan, count, cbuf, buflen, auxdata);
 
     case GENSIO_FUNC_CLOSE:
-	return muxc_close(chan, cbuf, buf);
+	return muxc_close(chan, (void *) cbuf, buf);
 
     case GENSIO_FUNC_FREE:
 	muxc_free(chan);
@@ -1775,10 +1801,10 @@ muxc_gensio_handler(struct gensio *io, int func, gensiods *count,
 	return muxc_alloc_channel(chan->mux, buf);
 
     case GENSIO_FUNC_OPEN:
-	return muxc_open(chan, cbuf, buf, true);
+	return muxc_open(chan, (void *) cbuf, buf, true);
 
     case GENSIO_FUNC_OPEN_NOCHILD:
-	return muxc_open(chan, cbuf, buf, false);
+	return muxc_open(chan, (void *) cbuf, buf, false);
 
     case GENSIO_FUNC_CONTROL:
 	return muxc_control(chan, *((bool *) cbuf), buflen, buf, count);
@@ -1835,24 +1861,26 @@ mux_shutdown_channels(struct mux_data *muxdata, int err)
 	case MUX_INST_PENDING_OPEN:
 	    muxc_set_state(chan, MUX_INST_CLOSED);
 	    mux_call_open_done(muxdata, chan, err);
-	    chan_deref(chan); /* Will free it. */
 	    break;
 
 	case MUX_INST_IN_OPEN:
 	    muxc_set_state(chan, MUX_INST_CLOSED);
 	    mux_call_open_done(muxdata, chan, err);
-	    chan_deref(chan); /* Lose our open ref. */
 	    break;
 
 	case MUX_INST_IN_OPEN_CLOSE:
 	    muxc_set_state(chan, MUX_INST_CLOSED);
+	    chan_ref(chan); /* Don't let the open call free us. */
 	    mux_call_open_done(muxdata, chan, err);
 	    finish_close(chan);
+	    chan_deref(chan);
 	    break;
 
 	case MUX_INST_IN_CLOSE:
 	    /* Just close it. */
+	    muxc_set_state(chan, MUX_INST_CLOSED);
 	    finish_close(chan);
+	    chan_deref(chan);
 	    break;
 
 	case MUX_INST_OPEN:
@@ -1960,8 +1988,11 @@ mux_on_err_close(struct gensio *child, void *close_data)
 {
     struct mux_data *muxdata = close_data;
 
-    mux_lock(muxdata);
-    mux_shutdown_channels(muxdata, muxdata->exit_err);
+    mux_lock_and_ref(muxdata);
+    if (muxdata->do_normal_close)
+	i_finish_close_close_done(muxdata->do_normal_close_chan, muxdata);
+    else
+	mux_shutdown_channels(muxdata, muxdata->exit_err);
     mux_deref_and_unlock(muxdata); /* Lose the open ref. */
 }
 
@@ -2190,7 +2221,7 @@ mux_child_read(struct mux_data *muxdata, int ierr,
 		    goto protocol_err;
 		}
 		muxdata->msgid = *buf >> 4;
-		if (muxdata->msgid == 0 || muxdata->msgid > MUX_MAX_MSG_NUM) {
+		if (muxdata->msgid <= 0 || muxdata->msgid > MUX_MAX_MSG_NUM) {
 		    proto_err_str = "msgid out of range";
 		    goto protocol_err;
 		}
@@ -2238,9 +2269,9 @@ mux_child_read(struct mux_data *muxdata, int ierr,
 	    switch (muxdata->msgid) {
 	    case MUX_NEW_CHANNEL: {
 		unsigned int remote_id = gensio_buf_to_u16(muxdata->hdr + 2);
-		if (muxdata->state == MUX_WAITING_OPEN)
+		if (muxdata->state == MUX_WAITING_OPEN) {
 		    chan = mux_chan0(muxdata);
-		else {
+		} else {
 		    int err;
 
 		    if (mux_find_remote_id(muxdata, remote_id)) {
@@ -2434,6 +2465,7 @@ mux_child_read(struct mux_data *muxdata, int ierr,
 		    break;
 
 		case MUX_DATA:
+		    assert(chan);
 		    if (muxdata->data_size == 0)
 			goto handle_read_no_data;
 		    if (chan_rdbufleft(chan) < muxdata->data_size + 3) {
@@ -2497,8 +2529,14 @@ mux_child_read(struct mux_data *muxdata, int ierr,
 					       muxdata->acc_open_data);
 			mux_lock(muxdata);
 		    } else {
-			chan_ref(chan);
-			mux_call_open_done(muxdata, chan, err);
+			mux_call_open_done(muxdata, chan, 0);
+			/*
+			 * In this case we got a new channel after we
+			 * sent a new channel on initialization.  We
+			 * will have already claimed a ref for the
+			 * open, undo the ref above since it is not
+			 * required.
+			 */
 			chan_deref(chan);
 		    }
 		} else {
@@ -2526,6 +2564,7 @@ mux_child_read(struct mux_data *muxdata, int ierr,
 		goto more_data;
 
 	    case MUX_DATA:
+		assert(chan);
 		if (buflen + muxdata->data_pos < muxdata->data_size + 2) {
 		    /* Not all data received yet. */
 		    chan_addrdbuf(chan, buf, buflen);
@@ -2673,8 +2712,8 @@ mux_gensio_alloc(struct gensio *child, const char *const args[],
 	return GE_NOTSUP;
 
     memset(&data, 0, sizeof(data));
-    data.max_read_size = GENSIO_DEFAULT_BUF_SIZE;
-    data.max_write_size = GENSIO_DEFAULT_BUF_SIZE;
+    data.max_read_size = GENSIO_DEFAULT_BUF_SIZE * 16;
+    data.max_write_size = GENSIO_DEFAULT_BUF_SIZE * 2;
     data.max_channels = 1000;
     err = gensio_get_default(o, "mux", "max-channels", false,
 			     GENSIO_DEFAULT_INT, NULL, &ival);
@@ -2734,7 +2773,6 @@ struct muxna_data {
 static void
 muxna_free(struct muxna_data *nadata)
 {
-
     gensio_mux_config_cleanup(&nadata->data);
     nadata->o->free(nadata->o, nadata);
 }
@@ -2759,7 +2797,6 @@ muxna_new_child(struct muxna_data *nadata, void **finish_data,
     if (!err) {
 	mux_lock(muxdata);
 	chan = mux_chan0(muxdata);
-	mux_ref(muxdata);
 	ncio->new_io = chan->io;
 	mux_set_state(muxdata, MUX_UNINITIALIZED);
 	muxdata->acc_open_done = ncio->open_done;
