@@ -51,48 +51,65 @@ struct net_data {
     int oob_char;
 };
 
-static int net_check_open(void *handler_data, int fd)
+static int net_check_open(void *handler_data, struct gensio_iod *iod)
 {
     struct net_data *tdata = handler_data;
 
-    tdata->last_err = gensio_os_check_socket_open(tdata->o, fd);
+    tdata->last_err = tdata->o->sock_control(iod, GENSIO_SOCKCTL_CHECK_OPEN,
+					     NULL, NULL);
     return tdata->last_err;
 }
 
 static int
-net_try_open(struct net_data *tdata, int *fd)
+net_try_open(struct net_data *tdata, struct gensio_iod **iod)
 {
-    int new_fd = -1, err = GE_INUSE;
+    struct gensio_iod *new_iod = NULL;
+    int err = GE_INUSE;
     int protocol = tdata->istcp ? GENSIO_NET_PROTOCOL_TCP
 				: GENSIO_NET_PROTOCOL_UNIX;
+    unsigned int setup = (GENSIO_SET_OPENSOCK_REUSEADDR |
+			  GENSIO_OPENSOCK_REUSEADDR |
+			  GENSIO_SET_OPENSOCK_KEEPALIVE |
+			  GENSIO_SET_OPENSOCK_NODELAY);
 
-    err = gensio_os_socket_open(tdata->o, tdata->ai, protocol, &new_fd);
-    if (err)
-	goto out;
-
-    err = gensio_os_socket_setup(tdata->o, new_fd, protocol, tdata->istcp,
-				 tdata->nodelay, GENSIO_OPENSOCK_REUSEADDR,
-				 tdata->lai);
-    if (err)
-	goto out;
-
+    if (tdata->istcp)
+	setup |= GENSIO_OPENSOCK_KEEPALIVE;
+    if (tdata->nodelay)
+	setup |= GENSIO_OPENSOCK_NODELAY;
  retry:
-    err = gensio_os_connect(tdata->o, new_fd, tdata->ai);
+    err = tdata->o->socket_open(tdata->o, tdata->ai, protocol, &new_iod);
+    if (err)
+	goto out;
+
+    err = tdata->o->socket_set_setup(new_iod, setup, tdata->lai);
+    if (err)
+	goto out;
+
+    err = tdata->o->connect(new_iod, tdata->ai);
     if (err == GE_INPROGRESS) {
-	*fd = new_fd;
+	*iod = new_iod;
 	goto out_return;
     }
 
-    if (err) {
-	if (gensio_addr_next(tdata->ai))
+    /*
+     * The GE_NOMEM check is strange here, but it really has more to
+     * do with testing.  connect() is not going to return GE_NOMEM
+     * unless it's an error trigger failure, and we really want to
+     * fail in that case or we will get a "error triggered but no
+     * failure" in the test.
+     */
+    if (err && err != GE_NOMEM) {
+	if (gensio_addr_next(tdata->ai)) {
+	    tdata->o->close(&new_iod);
 	    goto retry;
+	}
     }
  out:
     if (err) {
-	if (new_fd != -1)
-	    gensio_os_close(tdata->o, &new_fd);
+	if (new_iod)
+	    tdata->o->close(&new_iod);
     } else {
-	*fd = new_fd;
+	*iod = new_iod;
     }
 
  out_return:
@@ -100,22 +117,22 @@ net_try_open(struct net_data *tdata, int *fd)
 }
 
 static int
-net_retry_open(void *handler_data, int *fd)
+net_retry_open(void *handler_data, struct gensio_iod **iod)
 {
     struct net_data *tdata = handler_data;
 
     if (!gensio_addr_next(tdata->ai))
 	return tdata->last_err;
-    return net_try_open(tdata, fd);
+    return net_try_open(tdata, iod);
 }
 
 static int
-net_sub_open(void *handler_data, int *fd)
+net_sub_open(void *handler_data, struct gensio_iod **iod)
 {
     struct net_data *tdata = handler_data;
 
     gensio_addr_rewind(tdata->ai);
-    return net_try_open(tdata, fd);
+    return net_try_open(tdata, iod);
 }
 
 static void
@@ -131,13 +148,13 @@ net_free(void *handler_data)
 }
 
 static int
-net_control(void *handler_data, int fd, bool get, unsigned int option,
-	    char *data, gensiods *datalen)
+net_control(void *handler_data, struct gensio_iod *iod, bool get,
+	    unsigned int option, char *data, gensiods *datalen)
 {
     struct net_data *tdata = handler_data;
     int rv, val;
-    unsigned int i;
-    gensiods pos;
+    unsigned int i, setup;
+    gensiods pos, size;
     struct gensio_addr *addr;
 
     switch (option) {
@@ -145,20 +162,23 @@ net_control(void *handler_data, int fd, bool get, unsigned int option,
 	if (!tdata->istcp)
 	    return GE_NOTSUP;
 	if (get) {
-	    if (fd != -1) {
-		rv = gensio_os_get_nodelay(tdata->o, fd,
-					   GENSIO_NET_PROTOCOL_TCP, &val);
+	    if (iod) {
+		setup = GENSIO_SET_OPENSOCK_NODELAY;
+		rv = tdata->o->socket_get_setup(iod, &setup);
 		if (rv)
 		    return rv;
+		val = !!(setup & GENSIO_OPENSOCK_NODELAY);
 	    } else {
 		val = tdata->nodelay;
 	    }
 	    *datalen = snprintf(data, *datalen, "%d", val);
 	} else {
 	    val = strtoul(data, NULL, 0);
-	    if (fd != -1) {
-		rv = gensio_os_set_nodelay(tdata->o, fd,
-					   GENSIO_NET_PROTOCOL_TCP, val);
+	    if (iod) {
+		setup = GENSIO_SET_OPENSOCK_NODELAY;
+		if (val)
+		    setup |= GENSIO_OPENSOCK_NODELAY;
+		rv = tdata->o->socket_set_setup(iod, val, NULL);
 		if (rv)
 		    return rv;
 	    }
@@ -173,7 +193,8 @@ net_control(void *handler_data, int fd, bool get, unsigned int option,
 	if (strtoul(data, NULL, 0) > 0)
 	    return GE_NOTFOUND;
 
-	rv = gensio_os_getsockname(tdata->o, fd, &addr);
+	rv = tdata->o->sock_control(iod, GENSIO_SOCKCTL_GET_SOCKNAME,
+				    &addr, NULL);
 	if (rv)
 	    return rv;
 
@@ -208,7 +229,8 @@ net_control(void *handler_data, int fd, bool get, unsigned int option,
 	return 0;
 
     case GENSIO_CONTROL_LPORT:
-	rv = gensio_os_socket_get_port(tdata->o, fd, &i);
+	size = sizeof(unsigned int);
+	rv = tdata->o->sock_control(iod, GENSIO_SOCKCTL_GET_PORT, &i, &size);
 	if (rv)
 	    return rv;
 	*datalen = snprintf(data, *datalen, "%d", i);
@@ -220,8 +242,8 @@ net_control(void *handler_data, int fd, bool get, unsigned int option,
 }
 
 static int
-net_read(int fd, void *data, gensiods count, gensiods *rcount,
-	 const char ***auxdata, void *cb_data)
+net_except_read(struct gensio_iod *iod, void *data, gensiods count,
+		gensiods *rcount, const char ***auxdata, void *cb_data)
 {
     struct net_data *tdata = cb_data;
     static const char *argv[3] = { "oob", "oobtcp", NULL };
@@ -238,19 +260,11 @@ net_read(int fd, void *data, gensiods count, gensiods *rcount,
 	return 0;
     }
 
-    return gensio_os_recv(tdata->o, fd, data, count, rcount, 0);
-}
-
-static void
-net_read_ready(void *handler_data, int fd)
-{
-    struct net_data *tdata = handler_data;
-
-    gensio_fd_ll_handle_incoming(tdata->ll, net_read, NULL, tdata);
+    return tdata->o->recv(iod, data, count, rcount, 0);
 }
 
 static int
-net_except_ready(void *handler_data, int fd)
+net_except_ready(void *handler_data, struct gensio_iod *iod)
 {
     struct net_data *tdata = handler_data;
     unsigned char urgdata;
@@ -260,17 +274,17 @@ net_except_ready(void *handler_data, int fd)
     if (!tdata->istcp)
 	return GE_NOTSUP;
 
-    rv = gensio_os_recv(tdata->o, fd, &urgdata, 1, &rcount, GENSIO_MSG_OOB);
+    rv = tdata->o->recv(iod, &urgdata, 1, &rcount, GENSIO_MSG_OOB);
     if (rv || rcount == 0)
 	return GE_NOTSUP;
 
     tdata->oob_char = urgdata;
-    net_read_ready(tdata, fd);
+    gensio_fd_ll_handle_incoming(tdata->ll, net_except_read, NULL, tdata);
     return 0;
 }
 
 static int
-net_write(void *handler_data, int fd, gensiods *rcount,
+net_write(void *handler_data, struct gensio_iod *iod, gensiods *rcount,
 	  const struct gensio_sg *sg, gensiods sglen,
 	  const char *const *auxdata)
 {
@@ -290,7 +304,26 @@ net_write(void *handler_data, int fd, gensiods *rcount,
 	}
     }
 
-    return gensio_os_send(tdata->o, fd, sg, sglen, rcount, flags);
+    return tdata->o->send(iod, sg, sglen, rcount, flags);
+}
+
+static int
+net_check_close(void *handler_data, struct gensio_iod *iod,
+		enum gensio_ll_close_state state,
+		gensio_time *timeout)
+{
+    struct net_data *tdata = handler_data;
+    int err;
+
+    if (state == GENSIO_LL_CLOSE_STATE_START)
+	return 0;
+
+    err = tdata->o->graceful_close(&iod);
+    if (err == GE_INPROGRESS && timeout) {
+	timeout->secs = 0;
+	timeout->nsecs = 1000000;
+    }
+    return err;
 }
 
 static const struct gensio_fd_ll_ops net_fd_ll_ops = {
@@ -300,7 +333,8 @@ static const struct gensio_fd_ll_ops net_fd_ll_ops = {
     .free = net_free,
     .control = net_control,
     .except_ready = net_except_ready,
-    .write = net_write
+    .write = net_write,
+    .check_close = net_check_close
 };
 
 static int
@@ -372,8 +406,8 @@ net_gensio_alloc(const struct gensio_addr *iai, const char * const args[],
     tdata->o = o;
     tdata->nodelay = nodelay;
 
-    tdata->ll = fd_gensio_ll_alloc(o, -1, &net_fd_ll_ops, tdata, max_read_size,
-				   false);
+    tdata->ll = fd_gensio_ll_alloc(o, NULL, &net_fd_ll_ops, tdata,
+				   max_read_size, false);
     if (!tdata->ll)
 	goto out_nomem;
 
@@ -483,7 +517,7 @@ struct netna_data {
     gensio_acc_done cb_en_done;
 
     struct gensio_addr *ai;		/* The address list for the portname. */
-    struct opensocks   *acceptfds;	/* The file descriptor used to
+    struct gensio_opensocks *acceptfds;	/* The file descriptor used to
 					   accept connections on the
 					   NET port. */
 #if HAVE_UNIX
@@ -514,17 +548,17 @@ static const struct gensio_fd_ll_ops net_server_fd_ll_ops = {
 };
 
 static void
-netna_fd_cleared(int fd, void *cbdata)
+netna_fd_cleared(struct gensio_iod *iod, void *cbdata)
 {
     struct netna_data *nadata = cbdata;
     unsigned int num_left, i;
 
     for (i = 0; i < nadata->nr_acceptfds; i++) {
-	if (fd == nadata->acceptfds[i].fd)
+	if (iod == nadata->acceptfds[i].iod)
 	    break;
     }
     assert(i < nadata->nr_acceptfds);
-    gensio_os_close(nadata->o, &nadata->acceptfds[i].fd);
+    nadata->o->close(&nadata->acceptfds[i].iod);
 
     nadata->o->lock(nadata->lock);
     assert(nadata->nr_accept_close_waiting > 0);
@@ -545,7 +579,7 @@ netna_set_fd_enables(struct netna_data *nadata, bool enable)
     unsigned int i;
 
     for (i = 0; i < nadata->nr_acceptfds; i++)
-	nadata->o->set_read_handler(nadata->o, nadata->acceptfds[i].fd, enable);
+	nadata->o->set_read_handler(nadata->acceptfds[i].iod, enable);
 }
 
 static void
@@ -557,18 +591,20 @@ netna_finish_server_open(struct gensio *net, int err, void *cb_data)
 }
 
 static void
-netna_readhandler(int fd, void *cbdata)
+netna_readhandler(struct gensio_iod *iod, void *cbdata)
 {
     struct netna_data *nadata = cbdata;
-    int new_fd = -1;
+    struct gensio_iod *new_iod = NULL;
     struct gensio_addr *raddr;
     struct net_data *tdata = NULL;
     struct gensio *io = NULL;
-    int protocol = nadata->istcp ? GENSIO_NET_PROTOCOL_TCP
-				 : GENSIO_NET_PROTOCOL_UNIX;
+    unsigned int setup = (GENSIO_SET_OPENSOCK_REUSEADDR |
+			  GENSIO_OPENSOCK_REUSEADDR |
+			  GENSIO_SET_OPENSOCK_KEEPALIVE |
+			  GENSIO_SET_OPENSOCK_NODELAY);
     int err;
 
-    err = gensio_os_accept(nadata->o, fd, &raddr, &new_fd);
+    err = nadata->o->accept(iod, &raddr, &new_iod);
     if (err) {
 	if (err != GE_NODATA)
 	    /* FIXME - maybe shut down the socket I/O? */
@@ -581,19 +617,19 @@ netna_readhandler(int fd, void *cbdata)
     err = base_gensio_accepter_new_child_start(nadata->acc);
     if (err) {
 	gensio_addr_free(raddr);
-	gensio_os_close(nadata->o, &new_fd);
+	nadata->o->close(&new_iod);
 	return;
     }
 
 #ifdef HAVE_TCPD_H
     if (nadata->istcp && nadata->tcpd != GENSIO_TCPD_OFF) {
-	const char *msg = gensio_os_check_tcpd_ok(new_fd,
+	const char *msg = gensio_os_check_tcpd_ok(new_iod,
 						  nadata->tcpd_progname);
 
 	if (msg) {
 	    if (nadata->tcpd == GENSIO_TCPD_PRINT) {
 		struct gensio_sg sg[1] = { { msg, strlen(msg) } };
-		gensio_os_send(nadata->o, new_fd, sg, 1, NULL, 0);
+		nadata->o->send(new_iod, sg, 1, NULL, 0);
 	    }
 	    gensio_acc_log(nadata->acc, GENSIO_LOG_INFO,
 			   "Error accepting net gensio: tcpd check failed");
@@ -615,18 +651,21 @@ netna_readhandler(int fd, void *cbdata)
     tdata->oob_char = -1;
     tdata->ai = raddr;
     tdata->istcp = nadata->istcp;
+    tdata->nodelay = nadata->nodelay;
     raddr = NULL;
     
-    err = gensio_os_socket_setup(tdata->o, new_fd, protocol, tdata->istcp,
-				 tdata->nodelay, GENSIO_OPENSOCK_REUSEADDR,
-				 tdata->lai);
+    if (tdata->istcp)
+	setup |= GENSIO_OPENSOCK_KEEPALIVE;
+    if (tdata->nodelay)
+	setup |= GENSIO_OPENSOCK_NODELAY;
+    err = tdata->o->socket_set_setup(new_iod, setup, NULL);
     if (err) {
 	gensio_acc_log(nadata->acc, GENSIO_LOG_ERR,
 		       "Error setting up net port: %s", gensio_err_to_str(err));
 	goto out_err;
     }
 
-    tdata->ll = fd_gensio_ll_alloc(nadata->o, new_fd, &net_server_fd_ll_ops,
+    tdata->ll = fd_gensio_ll_alloc(nadata->o, new_iod, &net_server_fd_ll_ops,
 				   tdata, nadata->max_read_size, false);
     if (!tdata->ll) {
 	gensio_acc_log(nadata->acc, GENSIO_LOG_ERR,
@@ -655,21 +694,21 @@ netna_readhandler(int fd, void *cbdata)
     base_gensio_accepter_new_child_end(nadata->acc, NULL, err);
     if (io) {
 	gensio_free(io);
-    } else if (tdata) {
-	if (tdata->ll)
-	    gensio_ll_free(tdata->ll);
-	else {
-	    /* gensio_ll_free() frees it otherwise. */
-	    net_free(tdata);
-	    if (new_fd != -1)
-		gensio_os_close(nadata->o, &new_fd);
-	}
-    } else {
-	if (raddr)
-	    gensio_addr_free(raddr);
-	if (new_fd != -1)
-	    gensio_os_close(nadata->o, &new_fd);
+	return;
     }
+    if (tdata) {
+	if (tdata->ll) {
+	    gensio_ll_free(tdata->ll);
+	    return;
+	}
+
+	/* gensio_ll_free() frees it otherwise. */
+	net_free(tdata);
+    }
+    if (raddr)
+	gensio_addr_free(raddr);
+    if (new_iod)
+	nadata->o->close(&new_iod);
 }
 
 #if HAVE_UNIX
@@ -706,18 +745,21 @@ netna_rm_unix_socket(struct gensio_addr *addr)
 }
 
 static int
-netna_b4_listen(int fd, void *data)
+netna_b4_listen(struct gensio_iod *iod, void *data)
 {
     struct netna_data *nadata = data;
+#if HAVE_UNIX
     char pwbuf[16384];
     uid_t ownerid = -1;
     uid_t groupid = -1;
     int err;
     char unpath[MAX_UNIX_ADDR_PATH];
+#endif
 
     if (nadata->istcp)
 	return 0;
 
+#if HAVE_UNIX
     get_unix_addr_path(nadata->ai, unpath);
 
     /* Set up perms for Unix domain sockets. */
@@ -765,6 +807,9 @@ netna_b4_listen(int fd, void *data)
     err = errno;
  out_err:
     return gensio_os_err_to_err(nadata->o, err);
+#else
+    return GE_NOTSUP;
+#endif
 }
 
 static int
@@ -772,7 +817,8 @@ netna_startup(struct gensio_accepter *accepter, struct netna_data *nadata)
 {
     int rv;
 
-    rv = gensio_os_open_socket(nadata->o, nadata->ai, netna_readhandler,
+    rv = gensio_os_open_listen_sockets(nadata->o, nadata->ai,
+			       netna_readhandler,
 			       NULL, netna_fd_cleared, netna_b4_listen, nadata,
 			       nadata->opensock_flags,
 			       &nadata->acceptfds, &nadata->nr_acceptfds);
@@ -791,7 +837,7 @@ netna_shutdown(struct gensio_accepter *accepter,
     nadata->shutdown_done = shutdown_done;
     nadata->nr_accept_close_waiting = nadata->nr_acceptfds;
     for (i = 0; i < nadata->nr_acceptfds; i++)
-	nadata->o->clear_fd_handlers(nadata->o, nadata->acceptfds[i].fd);
+	nadata->o->clear_fd_handlers(nadata->acceptfds[i].iod);
 
     if (!nadata->istcp)
 	/* Remove the socket. */
@@ -818,8 +864,7 @@ netna_set_accept_callback_enable(struct gensio_accepter *accepter,
 
     nadata->cb_en_done = done;
     for (i = 0; i < nadata->nr_acceptfds; i++)
-	nadata->o->set_read_handler(nadata->o, nadata->acceptfds[i].fd,
-				    enabled);
+	nadata->o->set_read_handler(nadata->acceptfds[i].iod, enabled);
 
     if (done)
 	nadata->o->run(nadata->cb_en_done_runner);
@@ -936,7 +981,9 @@ netna_control_laddr(struct netna_data *nadata, bool get,
     if (i >= nadata->nr_acceptfds)
 	return GE_NOTFOUND;
 
-    rv = gensio_os_getsockname(nadata->o, nadata->acceptfds[i].fd, &addr);
+    rv = nadata->o->sock_control(nadata->acceptfds[i].iod,
+				 GENSIO_SOCKCTL_GET_SOCKNAME,
+				 &addr, NULL);
     if (rv)
 	return rv;
 
@@ -1012,10 +1059,9 @@ netna_disable(struct gensio_accepter *accepter, struct netna_data *nadata)
     unsigned int i;
 
     for (i = 0; i < nadata->nr_acceptfds; i++)
-	nadata->o->clear_fd_handlers_norpt(nadata->o,
-					   nadata->acceptfds[i].fd);
+	nadata->o->clear_fd_handlers_norpt(nadata->acceptfds[i].iod);
     for (i = 0; i < nadata->nr_acceptfds; i++)
-	gensio_os_close(nadata->o, &nadata->acceptfds[i].fd);
+	nadata->o->close(&nadata->acceptfds[i].iod);
 }
 
 static int
