@@ -8,8 +8,13 @@
 /* This code handles running a child process using a pty. */
 
 #include "config.h"
+
+#if HAVE_PTY
+
+#ifdef linux
 #define _XOPEN_SOURCE 600 /* Get posix_openpt() and friends. */
 #define _GNU_SOURCE /* Get ptsname_r(). */
+#endif
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -38,8 +43,11 @@ struct pty_data {
 
     struct gensio_ll *ll;
 
+    struct gensio_lock *lock;
+
     pid_t pid;
     int ptym;
+    struct gensio_iod *iod;
     const char **argv;
     const char **env;
 
@@ -60,7 +68,7 @@ struct pty_data {
     bool exit_code_set;
 };
 
-static int pty_check_open(void *handler_data, int fd)
+static int pty_check_open(void *handler_data, struct gensio_iod *iod)
 {
     return 0;
 }
@@ -102,11 +110,12 @@ gensio_setup_child_on_pty(struct pty_data *tdata)
     struct gensio_os_funcs *o = tdata->o;
     pid_t pid = -1;
     int ptym, err = 0;
+    struct gensio_iod *iod = NULL;
     uid_t ownerid = -1;
     uid_t groupid = -1;
     const char *pgm;
     bool link_created = false;
-#ifdef HAVE_PTSNAME_R
+#if HAVE_PTSNAME_R
     char ptsstr[PATH_MAX];
     char pwbuf[16384];
 #endif
@@ -115,10 +124,15 @@ gensio_setup_child_on_pty(struct pty_data *tdata)
     if (ptym == -1)
 	return gensio_os_err_to_err(o, errno);
 
-    if (fcntl(ptym, F_SETFL, O_NONBLOCK) == -1)
-	goto out_errno;
+    err = o->add_iod(o, GENSIO_IOD_DEV, ptym, &iod);
+    if (err)
+	goto out_err_noconv;
 
-#ifdef HAVE_PTSNAME_R
+    err = o->set_non_blocking(iod);
+    if (err)
+	goto out_err_noconv;
+
+#if HAVE_PTSNAME_R
     err = ptsname_r(ptym, ptsstr, sizeof(ptsstr));
     if (err)
 	goto out_errno;
@@ -211,20 +225,6 @@ gensio_setup_child_on_pty(struct pty_data *tdata)
 	if (grantpt(ptym) < 0)
 	    exit(1);
 
-	/* setsid() does this, but just in case... */
-	fd = open("/dev/tty", O_RDWR);
-	if (fd != -1) {
-	    ioctl(fd, TIOCNOTTY, NULL);
-	    close(fd);
-
-	    fd = open("/dev/tty", O_RDWR);
-	    if (fd != -1) {
-		fprintf(stderr, "pty fork: failed to drop control term: %s\r\n",
-			strerror(errno));
-		exit(1);
-	    }
-	}
-
 	if (setsid() == -1) {
 	    fprintf(stderr, "pty fork: failed to start new session: %s\r\n",
 		    strerror(errno));
@@ -233,6 +233,8 @@ gensio_setup_child_on_pty(struct pty_data *tdata)
 
 #if 0 /* FIXME = do we need this? */
 	if (setpgid(0, 0) == -1) {
+	    fprintf(stderr, "pty fork: failed setpgid: %s\r\n",
+		    strerror(errno));
 	    exit(1);
 	}
 #endif
@@ -244,6 +246,15 @@ gensio_setup_child_on_pty(struct pty_data *tdata)
 	    exit(1);
 	}
 
+#if defined(TIOCSCTTY) && !defined(linux)
+	/* Linux sets the first opened TTY to the controlling terminal. */
+	if (ioctl(fd, TIOCSCTTY, NULL) == -1) {
+	    fprintf(stderr, "pty fork: failed to set controlling tty: %s\r\n",
+		    strerror(errno));
+	    exit(1);
+	}
+#endif
+
 	/* fd will be closed by the loop to close everything. */
 	if (open("/dev/tty", O_RDWR) == -1) {
 	    fprintf(stderr, "pty fork: failed to set control term: %s\r\n",
@@ -270,7 +281,7 @@ gensio_setup_child_on_pty(struct pty_data *tdata)
 	for (i = 3; i < openfiles; i++)
 	    close(i);
 
-	err = gensio_os_setupnewprog();
+	err = gensio_unix_os_setupnewprog();
 	if (err) {
 	    fprintf(stderr, "Unable to set groups or user: %s\r\n",
 		    strerror(err));
@@ -291,35 +302,68 @@ gensio_setup_child_on_pty(struct pty_data *tdata)
  skip_child:
     tdata->pid = pid;
     tdata->ptym = ptym;
+    tdata->iod = iod;
     return 0;
  out_errno:
     err = errno;
  out_err:
+    err = gensio_os_err_to_err(o, err);
+ out_err_noconv:
     if (link_created)
 	unlink(tdata->link);
+    if (iod)
+	o->close(&iod);
     close(ptym);
-    return gensio_os_err_to_err(o, err);
+    return err;
 }
 
 static int
-pty_sub_open(void *handler_data, int *fd)
+pty_sub_open(void *handler_data, struct gensio_iod **riod)
 {
     struct pty_data *tdata = handler_data;
     int err;
 
     err = gensio_setup_child_on_pty(tdata);
     if (!err)
-	*fd = tdata->ptym;
+	*riod = tdata->iod;
 
     return err;
 }
 
 static int
-pty_check_close(void *handler_data, enum gensio_ll_close_state state,
+pty_check_exit_code(struct pty_data *tdata)
+{
+    struct gensio_os_funcs *o = tdata->o;
+    pid_t rv;
+    int err;
+
+    err = 0;
+    o->lock(tdata->lock);
+    if (tdata->pid != -1) {
+	rv = waitpid(tdata->pid, &tdata->exit_code, WNOHANG);
+	if (rv < 0) {
+	    err = gensio_os_err_to_err(o, errno);
+	    goto out_unlock;
+	}
+	if (rv == 0) {
+	    err = GE_INPROGRESS;
+	    goto out_unlock;
+	}
+	tdata->exit_code_set = true;
+	tdata->pid = -1;
+    }
+ out_unlock:
+    o->unlock(tdata->lock);
+    return err;
+}
+
+static int
+pty_check_close(void *handler_data, struct gensio_iod *iod,
+		enum gensio_ll_close_state state,
 		gensio_time *timeout)
 {
     struct pty_data *tdata = handler_data;
-    pid_t rv;
+    int err;
 
     if (state != GENSIO_LL_CLOSE_STATE_DONE)
 	return 0;
@@ -332,46 +376,42 @@ pty_check_close(void *handler_data, enum gensio_ll_close_state state,
 	    unlink(tdata->link);
     }
 
-    if (tdata->pid != -1) {
-	rv = waitpid(tdata->pid, &tdata->exit_code, WNOHANG);
-	if (rv < 0)
-	    return gensio_os_err_to_err(tdata->o, errno);
-	if (rv == 0) {
-	    timeout->secs = 0;
-	    timeout->nsecs = 10000000;
-	    return GE_INPROGRESS;
-	}
-	tdata->exit_code_set = true;
-	tdata->pid = -1;
+    err = pty_check_exit_code(tdata);
+    if (err == GE_INPROGRESS) {
+	timeout->secs = 0;
+	timeout->nsecs = 10000000;
     }
-    return 0;
+
+    return err;
 }
 
 static void
 pty_free(void *handler_data)
 {
     struct pty_data *tdata = handler_data;
+    struct gensio_os_funcs *o = tdata->o;
 
     if (tdata->link)
-	tdata->o->free(tdata->o, tdata->link);
+	o->free(o, tdata->link);
     if (tdata->owner)
-	tdata->o->free(tdata->o, tdata->owner);
+	o->free(o, tdata->owner);
     if (tdata->group)
-	tdata->o->free(tdata->o, tdata->group);
+	o->free(o, tdata->group);
     if (tdata->argv)
-	gensio_argv_free(tdata->o, tdata->argv);
+	gensio_argv_free(o, tdata->argv);
     if (tdata->env)
-	gensio_argv_free(tdata->o, tdata->env);
-    tdata->o->free(tdata->o, tdata);
+	gensio_argv_free(o, tdata->env);
+    if (tdata->lock)
+	o->free_lock(tdata->lock);
+    o->free(o, tdata);
 }
 
 static int
-pty_write(void *handler_data, int fd, gensiods *rcount,
+pty_write(void *handler_data, struct gensio_iod *iod, gensiods *rcount,
 	  const struct gensio_sg *sg, gensiods sglen,
 	  const char *const *auxdata)
 {
-    struct pty_data *tdata = handler_data;
-    int rv = gensio_os_write(tdata->o, fd, sg, sglen, rcount);
+    int rv = iod->f->write(iod, sg, sglen, rcount);
 
     if (rv && rv == GE_IOERR)
 	return GE_REMCLOSE; /* We don't seem to get EPIPE from ptys */
@@ -379,11 +419,10 @@ pty_write(void *handler_data, int fd, gensiods *rcount,
 }
 
 static int
-pty_do_read(int fd, void *data, gensiods count, gensiods *rcount,
-	    const char ***auxdata, void *cb_data)
+pty_do_read(struct gensio_iod *iod, void *data, gensiods count,
+	    gensiods *rcount, const char ***auxdata, void *cb_data)
 {
-    struct pty_data *tdata = cb_data;
-    int rv = gensio_os_read(tdata->o, fd, data, count, rcount);
+    int rv = iod->f->read(iod, data, count, rcount);
 
     if (rv && rv == GE_IOERR)
 	return GE_REMCLOSE; /* We don't seem to get EPIPE from ptys */
@@ -391,7 +430,7 @@ pty_do_read(int fd, void *data, gensiods count, gensiods *rcount,
 }
 
 static void
-pty_read_ready(void *handler_data, int fd)
+pty_read_ready(void *handler_data, struct gensio_iod *iod)
 {
     struct pty_data *tdata = handler_data;
 
@@ -399,12 +438,13 @@ pty_read_ready(void *handler_data, int fd)
 }
 
 static int
-pty_control(void *handler_data, int fd, bool get, unsigned int option,
-	    char *data, gensiods *datalen)
+pty_control(void *handler_data, struct gensio_iod *iod, bool get,
+	    unsigned int option, char *data, gensiods *datalen)
 {
     struct pty_data *tdata = handler_data;
+    struct gensio_os_funcs *o = tdata->o;
     const char **env, **argv;
-    int err, status;
+    int err, val;
 
     switch (option) {
     case GENSIO_CONTROL_ENVIRONMENT:
@@ -436,23 +476,40 @@ pty_control(void *handler_data, int fd, bool get, unsigned int option,
     case GENSIO_CONTROL_EXIT_CODE:
 	if (!get)
 	    return GE_NOTSUP;
+	err = 0;
+	o->lock(tdata->lock);
 	if (!tdata->exit_code_set)
-	    return GE_NOTREADY;
-	*datalen = snprintf(data, *datalen, "%d", tdata->exit_code);
-	return 0;
+	    err = GE_NOTREADY;
+	o->unlock(tdata->lock);
+	if (!err)
+	    *datalen = snprintf(data, *datalen, "%d", tdata->exit_code);
+	return err;
+
+    case GENSIO_CONTROL_KILL_TASK:
+	if (get)
+	    return GE_NOTSUP;
+	o->lock(tdata->lock);
+	if (tdata->pid == -1) {
+	    err = GE_NOTREADY;
+	} else {
+	    val = strtoul(data, NULL, 0);
+	    err = kill(tdata->pid, val ? SIGKILL : SIGTERM);
+	    if (err)
+		err = gensio_os_err_to_err(o, errno);
+	}
+	o->unlock(tdata->lock);
+	return err;
 
     case GENSIO_CONTROL_WAIT_TASK:
 	if (!get)
 	    return GE_NOTSUP;
-	if (tdata->pid == -1)
-	    return GE_NOTREADY;
-	err = waitpid(tdata->pid, &status, WNOHANG | WNOWAIT);
-	if (err <= 0)
-	    return GE_NOTREADY;
-	*datalen = snprintf(data, *datalen, "%d", status);
+	err = pty_check_exit_code(tdata);
+	if (err)
+	    return err;
+	*datalen = snprintf(data, *datalen, "%d", tdata->exit_code);
 	return 0;
 
-#ifdef HAVE_PTSNAME_R
+#if HAVE_PTSNAME_R
     case GENSIO_CONTROL_LADDR:
     case GENSIO_CONTROL_LPORT:
     {
@@ -469,7 +526,7 @@ pty_control(void *handler_data, int fd, bool get, unsigned int option,
 	    err = gensio_os_err_to_err(tdata->o, errno);
 	else
 	    *datalen = snprintf(data, *datalen, "%s", ptsstr);
-	return 0;
+	return err;
     }
 #endif
 
@@ -530,7 +587,7 @@ pty_gensio_alloc(const char * const argv[], const char * const args[],
     for (i = 0; args && args[i]; i++) {
 	if (gensio_check_keyds(args[i], "readbuf", &max_read_size) > 0)
 	    continue;
-#ifdef HAVE_PTSNAME_R
+#if HAVE_PTSNAME_R
 	if (gensio_check_keyvalue(args[i], "link", &link))
 	    continue;
 	if (gensio_check_keybool(args[i], "forcelink", &forcelink) > 0)
@@ -570,6 +627,11 @@ pty_gensio_alloc(const char * const argv[], const char * const args[],
 
     tdata->o = o;
     tdata->ptym = -1;
+
+    tdata->lock = o->alloc_lock(o);
+    if (!tdata->lock)
+	goto out_nomem;
+
     if (link) {
 	tdata->link = gensio_strdup(o, link);
 	if (!tdata->link)
@@ -602,8 +664,8 @@ pty_gensio_alloc(const char * const argv[], const char * const args[],
 	    goto out_nomem;
     }
 
-    tdata->ll = fd_gensio_ll_alloc(o, -1, &pty_fd_ll_ops, tdata, max_read_size,
-				   false);
+    tdata->ll = fd_gensio_ll_alloc(o, NULL, &pty_fd_ll_ops, tdata,
+				   max_read_size, false);
     if (!tdata->ll)
 	goto out_nomem;
 
@@ -643,3 +705,28 @@ str_to_pty_gensio(const char *str, const char * const args[],
 
     return err;
 }
+
+#else
+
+#include <gensio/gensio_class.h>
+#include <gensio/gensio_builtins.h>
+
+int
+pty_gensio_alloc(const char * const argv[], const char * const args[],
+		 struct gensio_os_funcs *o,
+		 gensio_event cb, void *user_data,
+		 struct gensio **new_gensio)
+{
+    return GE_NOTSUP;
+}
+
+int
+str_to_pty_gensio(const char *str, const char * const args[],
+		  struct gensio_os_funcs *o,
+		  gensio_event cb, void *user_data,
+		  struct gensio **new_gensio)
+{
+    return GE_NOTSUP;
+}
+
+#endif
