@@ -32,7 +32,7 @@ struct sctp_data {
 
     struct gensio_ll *ll;
 
-    int fd;
+    struct gensio_iod *iod;
 
     struct gensio_addr *addr;
     struct gensio_addr *laddr; /* Local address, NULL if not set. */
@@ -54,17 +54,12 @@ sctp_setup(struct sctp_data *tdata)
 {
     struct gensio_os_funcs *o = tdata->o;
     struct sctp_status status;
-    socklen_t stat_size = sizeof(status);
     unsigned int i;
+    int err;
 
-    if (getsockopt(tdata->fd, IPPROTO_SCTP, SCTP_STATUS, &status,
-		   &stat_size) == -1) {
-	/*
-	 * If the remote end closes, this fails with EINVAL.  Just
-	 * assume the remote end closed on error.
-	 */
-	return GE_REMCLOSE;
-    }
+    err = o->sctp_get_socket_status(tdata->iod, &status);
+    if (err)
+	return err;
 
     tdata->instreams = status.sstat_instrms;
     tdata->ostreams = status.sstat_outstrms;
@@ -83,12 +78,12 @@ sctp_setup(struct sctp_data *tdata)
     return 0;
 }
 
-static int sctp_check_open(void *handler_data, int fd)
+static int sctp_check_open(void *handler_data, struct gensio_iod *iod)
 {
     struct sctp_data *tdata = handler_data;
     int err;
 
-    err = gensio_os_check_socket_open(tdata->o, fd);
+    err = tdata->o->sock_control(iod, GENSIO_SOCKCTL_CHECK_OPEN, NULL, NULL);
     if (!err)
 	err = sctp_setup(tdata);
 
@@ -96,51 +91,42 @@ static int sctp_check_open(void *handler_data, int fd)
 }
 
 static int
-sctp_socket_setup(struct sctp_data *tdata, int fd)
+sctp_socket_setup(struct sctp_data *tdata, struct gensio_iod *iod)
 {
-    struct sctp_event_subscribe event_sub;
     int err;
+    unsigned int setup = (GENSIO_SET_OPENSOCK_REUSEADDR |
+			  GENSIO_OPENSOCK_REUSEADDR |
+			  GENSIO_SET_OPENSOCK_KEEPALIVE |
+			  GENSIO_OPENSOCK_KEEPALIVE |
+			  GENSIO_SET_OPENSOCK_NODELAY);
 
-    err = gensio_os_socket_setup(tdata->o, fd, GENSIO_NET_PROTOCOL_SCTP,
-				 true, tdata->nodelay,
-				 GENSIO_OPENSOCK_REUSEADDR, tdata->laddr);
+    if (tdata->nodelay)
+	setup |= GENSIO_OPENSOCK_NODELAY;
+    err = tdata->o->socket_set_setup(iod, setup, tdata->laddr);
     if (err)
 	return err;
 
-    if (setsockopt(fd, IPPROTO_SCTP, SCTP_INITMSG, &tdata->initmsg,
-		   sizeof(tdata->initmsg)) == -1)
-	return gensio_os_err_to_err(tdata->o, errno);
-
-    if (setsockopt(fd, IPPROTO_SCTP, SCTP_DELAYED_SACK, &tdata->sackinfo,
-		   sizeof(tdata->sackinfo)) == -1)
-	return gensio_os_err_to_err(tdata->o, errno);
-
-    memset(&event_sub, 0, sizeof(event_sub));
-    event_sub.sctp_data_io_event = 1;
-    if (setsockopt(fd, IPPROTO_SCTP, SCTP_EVENTS, &event_sub,
-		   sizeof(event_sub)) == -1)
-	return gensio_os_err_to_err(tdata->o, errno);
-
-    return 0;
+    return tdata->o->sctp_socket_setup(iod, true, &tdata->initmsg,
+				       &tdata->sackinfo);
 }
 
 static int
-sctp_try_open(struct sctp_data *tdata, int *fd)
+sctp_try_open(struct sctp_data *tdata, struct gensio_iod **iod)
 {
     int err = GE_INUSE;
 
-    err = gensio_os_socket_open(tdata->o, tdata->addr,
-				GENSIO_NET_PROTOCOL_SCTP, &tdata->fd);
+    err = tdata->o->socket_open(tdata->o, tdata->addr,
+				GENSIO_NET_PROTOCOL_SCTP, &tdata->iod);
     if (err)
 	goto out;
 
-    err = sctp_socket_setup(tdata, tdata->fd);
+    err = sctp_socket_setup(tdata, tdata->iod);
     if (err)
 	goto out;
 
-    err = gensio_os_sctp_connectx(tdata->o, tdata->fd, tdata->addr);
+    err = tdata->o->sctp_connectx(tdata->iod, tdata->addr);
     if (err == GE_INPROGRESS) {
-	*fd = tdata->fd;
+	*iod = tdata->iod;
 	goto out_return;
     } else if (err) {
 	goto out;
@@ -150,10 +136,10 @@ sctp_try_open(struct sctp_data *tdata, int *fd)
 
  out:
     if (err) {
-	if (tdata->fd != -1)
-	    gensio_os_close(tdata->o, &tdata->fd);
+	if (tdata->iod)
+	    tdata->o->close(&tdata->iod);
     } else {
-	*fd = tdata->fd;
+	*iod = tdata->iod;
     }
 
  out_return:
@@ -161,11 +147,11 @@ sctp_try_open(struct sctp_data *tdata, int *fd)
 }
 
 static int
-sctp_sub_open(void *handler_data, int *fd)
+sctp_sub_open(void *handler_data, struct gensio_iod **iod)
 {
     struct sctp_data *tdata = handler_data;
 
-    return sctp_try_open(tdata, fd);
+    return sctp_try_open(tdata, iod);
 }
 
 static void
@@ -190,32 +176,35 @@ sctp_free(void *handler_data)
 }
 
 static int
-sctp_control(void *handler_data, int fd, bool get, unsigned int option,
+sctp_control(void *handler_data, struct gensio_iod *iod, bool get, unsigned int option,
 	     char *data, gensiods *datalen)
 {
     struct sctp_data *tdata = handler_data;
     int rv, val;
-    unsigned int i;
-    gensiods pos;
+    unsigned int i, setup;
+    gensiods pos, size;
     struct gensio_addr *addr;
 
     switch (option) {
     case GENSIO_CONTROL_NODELAY:
 	if (get) {
-	    if (fd != -1) {
-		rv = gensio_os_get_nodelay(tdata->o, fd,
-					   GENSIO_NET_PROTOCOL_SCTP, &val);
+	    if (iod) {
+		setup = GENSIO_SET_OPENSOCK_NODELAY;
+		rv = tdata->o->socket_get_setup(iod, &setup);
 		if (rv)
 		    return rv;
+		val = !!(setup & GENSIO_OPENSOCK_NODELAY);
 	    } else {
 		val = tdata->nodelay;
 	    }
 	    *datalen = snprintf(data, *datalen, "%d", val);
 	} else {
 	    val = strtoul(data, NULL, 0);
-	    if (fd != -1) {
-		rv = gensio_os_set_nodelay(tdata->o, fd,
-					   GENSIO_NET_PROTOCOL_SCTP, val);
+	    if (iod) {
+		setup = GENSIO_SET_OPENSOCK_NODELAY;
+		if (val)
+		    setup |= GENSIO_OPENSOCK_NODELAY;
+		rv = tdata->o->socket_set_setup(iod, val, NULL);
 		if (rv)
 		    return rv;
 	    }
@@ -239,7 +228,8 @@ sctp_control(void *handler_data, int fd, bool get, unsigned int option,
 	if (i > 0)
 	    return GE_NOTFOUND;
 
-	rv = gensio_os_sctp_getladdrs(tdata->o, fd, &addr);
+	rv = tdata->o->sock_control(iod, GENSIO_SOCKCTL_GET_SOCKNAME,
+				    &addr, NULL);
 	if (rv)
 	    return rv;
 
@@ -260,7 +250,8 @@ sctp_control(void *handler_data, int fd, bool get, unsigned int option,
 	if (i > 0)
 	    return GE_NOTFOUND;
 
-	rv = gensio_os_sctp_getpaddrs(tdata->o, fd, &addr);
+	rv = tdata->o->sock_control(iod, GENSIO_SOCKCTL_GET_PEERNAME,
+				    &addr, NULL);
 	if (rv)
 	    return rv;
 
@@ -276,10 +267,12 @@ sctp_control(void *handler_data, int fd, bool get, unsigned int option,
     case GENSIO_CONTROL_RADDR_BIN:
 	if (!get)
 	    return GE_NOTSUP;
-	return gensio_os_sctp_getraddr(tdata->o, tdata->fd, data, datalen);
+	return tdata->o->sock_control(tdata->iod, GENSIO_SOCKCTL_GET_PEERRAW,
+				      data, datalen);
 
     case GENSIO_CONTROL_LPORT:
-	rv = gensio_os_socket_get_port(tdata->o, fd, &i);
+	size = sizeof(unsigned int);
+	rv = tdata->o->sock_control(iod, GENSIO_SOCKCTL_GET_PORT, &i, &size);
 	if (rv)
 	    return rv;
 	*datalen = snprintf(data, *datalen, "%d", i);
@@ -301,7 +294,7 @@ sctp_control(void *handler_data, int fd, bool get, unsigned int option,
 }
 
 static int
-sctp_write(void *handler_data, int fd, gensiods *rcount,
+sctp_write(void *handler_data, struct gensio_iod *iod, gensiods *rcount,
 	  const struct gensio_sg *sg, gensiods sglen,
 	  const char *const *auxdata)
 {
@@ -324,12 +317,11 @@ sctp_write(void *handler_data, int fd, gensiods *rcount,
     }
 
     sinfo.sinfo_stream = stream;
-    return gensio_os_sctp_send(tdata->o,
-			       tdata->fd, sg, sglen, rcount, &sinfo, 0);
+    return tdata->o->sctp_send(tdata->iod, sg, sglen, rcount, &sinfo, 0);
 }
 
 static int
-sctp_do_read(int fd, void *data, gensiods count, gensiods *rcount,
+sctp_do_read(struct gensio_iod *iod, void *data, gensiods count, gensiods *rcount,
 	     const char ***auxdata, void *cb_data)
 {
     struct sctp_data *tdata = cb_data;
@@ -339,8 +331,7 @@ sctp_do_read(int fd, void *data, gensiods count, gensiods *rcount,
     unsigned int stream;
     unsigned int i = 0;
 
-    rv = gensio_os_sctp_recvmsg(tdata->o,
-				fd, data, count, rcount, &sinfo, &flags);
+    rv = tdata->o->sctp_recvmsg(iod, data, count, rcount, &sinfo, &flags);
     /* If the data length is zero, we won't have any info. */
     if (rv || *rcount == 0)
 	return rv;
@@ -361,7 +352,7 @@ sctp_do_read(int fd, void *data, gensiods count, gensiods *rcount,
 }
 
 static void
-sctp_read_ready(void *handler_data, int fd)
+sctp_read_ready(void *handler_data, struct gensio_iod *iod)
 {
     struct sctp_data *tdata = handler_data;
 
@@ -469,10 +460,9 @@ sctp_gensio_alloc(const struct gensio_addr *iai, const char * const args[],
     tdata->initmsg.sinit_num_ostreams = ostreams;
     tdata->sackinfo.sack_freq = sack_freq;
     tdata->sackinfo.sack_delay = sack_delay;
-    tdata->fd = -1;
     tdata->nodelay = nodelay;
 
-    tdata->ll = fd_gensio_ll_alloc(o, -1, &sctp_fd_ll_ops, tdata,
+    tdata->ll = fd_gensio_ll_alloc(o, NULL, &sctp_fd_ll_ops, tdata,
 				   max_read_size, false);
     if (!tdata->ll)
 	goto out_nomem;
@@ -546,7 +536,7 @@ struct sctpna_data {
     gensio_acc_done cb_en_done;
 
     struct gensio_addr *ai;
-    struct opensocks *acceptfds;
+    struct gensio_opensocks *acceptfds;
     unsigned int nr_acceptfds;
 
     unsigned int nr_accept_close_waiting;
@@ -563,17 +553,17 @@ static const struct gensio_fd_ll_ops sctp_server_fd_ll_ops = {
 };
 
 static void
-sctpna_fd_cleared(int fd, void *cbdata)
+sctpna_fd_cleared(struct gensio_iod *iod, void *cbdata)
 {
     struct sctpna_data *nadata = cbdata;
     unsigned int num_left, i;
 
     for (i = 0; i < nadata->nr_acceptfds; i++) {
-	if (fd == nadata->acceptfds[i].fd)
+	if (iod == nadata->acceptfds[i].iod)
 	    break;
     }
     assert(i < nadata->nr_acceptfds);
-    gensio_os_close(nadata->o, &nadata->acceptfds[i].fd);
+    nadata->o->close(&nadata->acceptfds[i].iod);
 
     nadata->o->lock(nadata->lock);
     assert(nadata->nr_accept_close_waiting > 0);
@@ -594,7 +584,7 @@ sctpna_set_fd_enables(struct sctpna_data *nadata, bool enable)
     unsigned int i;
 
     for (i = 0; i < nadata->nr_acceptfds; i++)
-	nadata->o->set_read_handler(nadata->o, nadata->acceptfds[i].fd, enable);
+	nadata->o->set_read_handler(nadata->acceptfds[i].iod, enable);
 }
 
 static void
@@ -606,15 +596,15 @@ sctpna_finish_server_open(struct gensio *net, int err, void *cb_data)
 }
 
 static void
-sctpna_readhandler(int fd, void *cbdata)
+sctpna_readhandler(struct gensio_iod *iod, void *cbdata)
 {
     struct sctpna_data *nadata = cbdata;
-    int new_fd = -1;
+    struct gensio_iod *new_iod = NULL;
     struct sctp_data *tdata = NULL;
     struct gensio *io = NULL;
     int err;
 
-    err = gensio_os_accept(nadata->o, fd, NULL, &new_fd);
+    err = nadata->o->accept(iod, NULL, &new_iod);
     if (err) {
 	if (err != GE_NODATA)
 	    /* FIXME - maybe shut down the socket I/O? */
@@ -626,7 +616,7 @@ sctpna_readhandler(int fd, void *cbdata)
 
     err = base_gensio_accepter_new_child_start(nadata->acc);
     if (err) {
-	gensio_os_close(nadata->o, &new_fd);
+	nadata->o->close(&new_iod);
 	return;
     }
 
@@ -639,10 +629,12 @@ sctpna_readhandler(int fd, void *cbdata)
     }
 
     tdata->o = nadata->o;
-    tdata->fd = new_fd;
+    tdata->iod = new_iod;
     tdata->nodelay = nadata->nodelay;
+    tdata->initmsg = nadata->initmsg;
+    tdata->sackinfo = nadata->sackinfo;
 
-    err = sctp_socket_setup(tdata, new_fd);
+    err = sctp_socket_setup(tdata, new_iod);
     if (!err)
 	err = sctp_setup(tdata);
     if (err) {
@@ -652,7 +644,7 @@ sctpna_readhandler(int fd, void *cbdata)
 	goto out_err;
     }
 
-    tdata->ll = fd_gensio_ll_alloc(nadata->o, new_fd, &sctp_server_fd_ll_ops,
+    tdata->ll = fd_gensio_ll_alloc(nadata->o, new_iod, &sctp_server_fd_ll_ops,
 				   tdata, nadata->max_read_size, false);
     if (!tdata->ll) {
 	gensio_acc_log(nadata->acc, GENSIO_LOG_ERR,
@@ -680,31 +672,26 @@ sctpna_readhandler(int fd, void *cbdata)
     base_gensio_accepter_new_child_end(nadata->acc, NULL, err);
     if (io) {
 	gensio_free(io);
-    } else if (tdata) {
-	if (tdata->ll)
-	    gensio_ll_free(tdata->ll);
-	else {
-	    sctp_free(tdata);
-	    if (new_fd != -1)
-		gensio_os_close(nadata->o, &new_fd);
-	}
+	return;
     }
+    if (tdata) {
+	if (tdata->ll) {
+	    gensio_ll_free(tdata->ll);
+	    return;
+	}
+	sctp_free(tdata);
+    }
+    if (new_iod)
+	nadata->o->close(&new_iod);
 }
 
 static int
-sctpna_setup_socket(int fd, void *data)
+sctpna_setup_socket(struct gensio_iod *iod, void *data)
 {
     struct sctpna_data *nadata = data;
 
-    if (setsockopt(fd, IPPROTO_SCTP, SCTP_INITMSG, &nadata->initmsg,
-		   sizeof(nadata->initmsg)) == -1)
-	return gensio_os_err_to_err(nadata->o, errno);
-
-    if (setsockopt(fd, IPPROTO_SCTP, SCTP_DELAYED_SACK, &nadata->sackinfo,
-		   sizeof(nadata->sackinfo)) == -1)
-	return gensio_os_err_to_err(nadata->o, errno);
-
-    return 0;
+    return iod->f->sctp_socket_setup(iod, false,
+				     &nadata->initmsg, &nadata->sackinfo);
 }
 
 static int
@@ -712,7 +699,7 @@ sctpna_startup(struct gensio_accepter *accepter, struct sctpna_data *nadata)
 {
     int rv;
 
-    rv = gensio_os_sctp_open_socket(nadata->o, nadata->ai,
+    rv = gensio_os_open_listen_sockets(nadata->o, nadata->ai,
 				    sctpna_readhandler, NULL, sctpna_fd_cleared,
 				    sctpna_setup_socket, nadata,
 				    nadata->opensock_flags,
@@ -733,7 +720,7 @@ sctpna_shutdown(struct gensio_accepter *accepter,
     nadata->shutdown_done = shutdown_done;
     nadata->nr_accept_close_waiting = nadata->nr_acceptfds;
     for (i = 0; i < nadata->nr_acceptfds; i++)
-	nadata->o->clear_fd_handlers(nadata->o, nadata->acceptfds[i].fd);
+	nadata->o->clear_fd_handlers(nadata->acceptfds[i].iod);
     return 0;
 }
 
@@ -876,7 +863,9 @@ sctpna_control_laddr(struct sctpna_data *nadata, bool get,
     if (i >= nadata->nr_acceptfds)
 	return GE_NOTFOUND;
 
-    rv = gensio_os_sctp_getladdrs(nadata->o, nadata->acceptfds[i].fd, &addrs);
+    rv = nadata->o->sock_control(nadata->acceptfds[i].iod,
+				 GENSIO_SOCKCTL_GET_SOCKNAME,
+				 &addrs, NULL);
     if (rv)
 	return rv;
 
@@ -932,10 +921,9 @@ sctpna_disable(struct gensio_accepter *accepter, struct sctpna_data *nadata)
     unsigned int i;
 
     for (i = 0; i < nadata->nr_acceptfds; i++)
-	nadata->o->clear_fd_handlers_norpt(nadata->o,
-					   nadata->acceptfds[i].fd);
+	nadata->o->clear_fd_handlers_norpt(nadata->acceptfds[i].iod);
     for (i = 0; i < nadata->nr_acceptfds; i++)
-	gensio_os_close(nadata->o, &nadata->acceptfds[i].fd);
+	nadata->o->close(&nadata->acceptfds[i].iod);
 }
 
 static int
