@@ -8,22 +8,17 @@
 /* This code handles stdio stream I/O. */
 
 #include "config.h"
+
 #include <stdlib.h>
-#include <unistd.h>
 #include <string.h>
 #include <errno.h>
-#include <fcntl.h>
 #include <stdio.h>
 #include <assert.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <unistd.h>
-#include <sys/wait.h>
 
 #include <gensio/gensio.h>
-#include <gensio/gensio_class.h>
 #include <gensio/gensio_osops.h>
 #include <gensio/gensio_builtins.h>
+#include <gensio/gensio_class.h>
 
 #include "utils.h"
 
@@ -38,17 +33,13 @@ struct stdion_channel {
 
     int ll_err; /* Set if an error occurs reading or writing. */
 
-    int infd;
-    int outfd;
+    struct gensio_iod *in_iod;
+    struct gensio_iod *out_iod;
 
     /* Are the above fds registered with set_fd_handlers()? */
     bool in_handler_set;
     bool out_handler_set;
 
-    /* Mark if these file are regular file, needing special handling. */
-    bool infd_regfile;
-    bool outfd_regfile;
-
     unsigned int refcount;
 
     struct gensio *io;
@@ -94,6 +85,7 @@ struct stdiona_data {
 
     struct gensio_os_funcs *o;
 
+    bool raw;
     bool stderr_to_stdout;
     bool noredir_stderr;
 
@@ -108,11 +100,6 @@ struct stdiona_data {
 
     struct gensio_timer *waitpid_timer;
 
-    int old_flags_ostdin;
-    int old_flags_ostdout;
-    bool old_flags_ostdin_set;
-    bool old_flags_ostdout_set;
-
     /* For the accepter only. */
     bool in_free;
     bool in_shutdown;
@@ -130,7 +117,7 @@ struct stdiona_data {
      * If non-zero, this is the PID of the other process and we are
      * in client mode.
      */
-    pid_t opid;
+    intptr_t opid;
 
     struct stdion_channel io; /* stdin, stdout */
     struct stdion_channel err; /* stderr */
@@ -141,9 +128,6 @@ struct stdiona_data {
     struct gensio_accepter *acc;
 };
 
-static void i_stdion_fd_cleared(int fd, struct stdiona_data *nadata,
-				struct stdion_channel *schan);
-
 static void
 stdiona_lock(struct stdiona_data *nadata)
 {
@@ -159,6 +143,12 @@ stdiona_unlock(struct stdiona_data *nadata)
 static void
 stdiona_finish_free(struct stdiona_data *nadata)
 {
+    struct gensio_os_funcs *o = nadata->o;
+
+    if (nadata->io.out_iod)
+	o->release_iod(nadata->io.out_iod);
+    if (nadata->io.in_iod)
+	o->release_iod(nadata->io.in_iod);
     if (nadata->argv)
 	gensio_argv_free(nadata->o, nadata->argv);
     if (nadata->env)
@@ -224,7 +214,7 @@ stdion_write(struct gensio *io, gensiods *count,
     if (schan->ll_err) {
 	rv = schan->ll_err;
     } else {
-	rv = gensio_os_write(nadata->o, schan->infd, sg, sglen, count);
+	rv = nadata->o->write(schan->in_iod, sg, sglen, count);
 	if (rv)
 	    schan->ll_err = rv;
     }
@@ -241,22 +231,21 @@ stdion_finish_read(struct stdion_channel *schan, int err)
     struct gensio *io = schan->io;
     gensiods count;
 
-    if (err) {
-	/* Do this here so the user can modify it. */
-	schan->read_enabled = false;
-	schan->data_pending_len = 0;
+    if (err && !schan->ll_err && schan->out_iod) {
 	schan->ll_err = err;
-	if (!schan->outfd_regfile)
-	    nadata->o->set_read_handler(nadata->o, schan->outfd, false);
-	do {
-	    stdiona_unlock(nadata);
-	    gensio_cb(io, GENSIO_EVENT_READ, err, NULL, NULL, NULL);
-	    stdiona_lock(nadata);
-	} while (!schan->closed && schan->read_enabled);
-    } else {
-	while (schan->data_pending_len && !schan->closed &&
+	nadata->o->set_read_handler(schan->out_iod, false);
+	nadata->o->set_except_handler(schan->out_iod, false);
+    }
+
+    while ((schan->data_pending_len || schan->ll_err) &&
 	       schan->read_enabled) {
-	    count = schan->data_pending_len;
+	count = schan->data_pending_len;
+	if (schan->ll_err && schan->data_pending_len == 0) {
+	    schan->read_enabled = false;
+	    stdiona_unlock(nadata);
+	    gensio_cb(io, GENSIO_EVENT_READ, schan->ll_err, NULL, NULL, NULL);
+	    stdiona_lock(nadata);
+	} else {
 	    stdiona_unlock(nadata);
 	    gensio_cb(io, GENSIO_EVENT_READ, 0,
 		      schan->read_data + schan->data_pos, &count, NULL);
@@ -273,39 +262,65 @@ stdion_finish_read(struct stdion_channel *schan, int err)
 
     schan->in_read = false;
 
-    if (schan->read_enabled && !schan->outfd_regfile)
-	nadata->o->set_read_handler(nadata->o, schan->outfd, true);
+    if (schan->out_iod) {
+	if (schan->closed) {
+	    nadata->o->set_read_handler(schan->out_iod, false);
+	    nadata->o->set_except_handler(schan->out_iod, false);
+	} else if (schan->read_enabled) {
+	    nadata->o->set_read_handler(schan->out_iod, true);
+	    nadata->o->set_except_handler(schan->out_iod, true);
+	}
+    }
 }
 
+/* FIXME - This should probably be configurable. */
+#define NUM_WAIT_RETRIES 1000
 static void
 check_waitpid(struct stdion_channel *schan)
 {
     struct stdiona_data *nadata = schan->nadata;
+    struct gensio_os_funcs *o = nadata->o;
+    int rv;
+    gensiods count = 0;
+    gensio_time timeout = { 0, 10000000 };
 
     if (nadata->closing_chan)
 	schan = nadata->closing_chan;
+
+    /* Wait for the output buffer to clear for half our allotted time. */
+    if (schan->out_iod) {
+	o->bufcount(schan->out_iod, GENSIO_OUT_BUF, &count);
+	if (count > 0 && nadata->waitpid_retries < NUM_WAIT_RETRIES / 2)
+	    goto try_again;
+    }
+
+    if (schan->in_iod)
+	o->close(&schan->in_iod);
+    if (schan->out_iod)
+	o->close(&schan->out_iod);
+
     if (nadata->opid != -1 && !nadata->io.out_handler_set &&
 		!nadata->io.in_handler_set && !nadata->err.out_handler_set) {
-	pid_t rv;
+	rv = o->wait_subprog(o, nadata->opid, &nadata->exit_code);
+	if (rv == GE_INPROGRESS) {
+	    goto try_again;
+	} else {
+	    if (rv)
+		/* FIXME = no real way to report this. */
+		;
 
-	rv = waitpid(nadata->opid, &nadata->exit_code, WNOHANG);
-	if (rv < 0)
-	    /* FIXME = no real way to report this. */
-	    ;
-
-	if (rv == 0) {
-	    gensio_time timeout = { 0, 10000000 };
-
-	    nadata->waitpid_retries++;
-	    /* The sub-process has not died, wait a bit and try again. */
-	    stdiona_ref(nadata);
-	    nadata->o->start_timer(nadata->waitpid_timer, &timeout);
-	    nadata->closing_chan = schan;
-	    return;
+	    nadata->exit_code_set = true;
+	    nadata->opid = -1;
 	}
+    }
 
-	nadata->exit_code_set = true;
-	nadata->opid = -1;
+ close_anyway:
+    if (schan->in_iod)
+	o->close(&schan->in_iod);
+    if (schan->out_iod) {
+	if (count > 0)
+	    o->flush(schan->out_iod, GENSIO_OUT_BUF);
+	o->close(&schan->out_iod);
     }
 
     if (schan->close_done) {
@@ -318,12 +333,26 @@ check_waitpid(struct stdion_channel *schan)
 	stdiona_unlock(nadata);
 	close_done(schan->io, close_data);
 	stdiona_lock(nadata);
-
-	if (schan->in_free && schan->io) {
-	    gensio_data_free(schan->io);
-	    schan->io = NULL;
-	}
     }
+
+    if (schan->in_free && schan->io) {
+	gensio_data_free(schan->io);
+	schan->io = NULL;
+	stdiona_deref(nadata);
+    }
+    return;
+
+ try_again:
+    /* The sub-process has not died or buffer is not clear, wait a
+       bit and try again. */
+
+    if (nadata->waitpid_retries >= NUM_WAIT_RETRIES)
+	goto close_anyway;
+    nadata->waitpid_retries++;
+    stdiona_ref(nadata);
+    rv = o->start_timer(nadata->waitpid_timer, &timeout);
+    assert(rv == 0);
+    nadata->closing_chan = schan;
 }
 
 static void
@@ -348,17 +377,9 @@ stdion_start_close(struct stdion_channel *schan)
 
     schan->read_enabled = false;
     schan->xmit_enabled = false;
-    if (schan->outfd_regfile)
-	i_stdion_fd_cleared(schan->outfd, nadata, schan);
-    else
-	nadata->o->clear_fd_handlers(nadata->o, schan->outfd);
-
-    if (schan->infd != -1) {
-	if (schan->infd_regfile)
-	    i_stdion_fd_cleared(schan->infd, nadata, schan);
-	else
-	    nadata->o->clear_fd_handlers(nadata->o, schan->infd);
-    }
+    nadata->o->clear_fd_handlers(schan->out_iod);
+    if (schan->in_iod)
+	nadata->o->clear_fd_handlers(schan->in_iod);
 }
 
 static int
@@ -367,8 +388,8 @@ stdion_do_read(struct stdiona_data *nadata, struct stdion_channel *schan)
     int rv;
     gensiods count;
 
-    rv = gensio_os_read(nadata->o, schan->outfd, schan->read_data,
-			schan->max_read_size, &count);
+    rv = nadata->o->read(schan->out_iod, schan->read_data,
+			 schan->max_read_size, &count);
     if (!rv) {
 	schan->data_pending_len = count;
 	schan->data_pos = 0;
@@ -393,46 +414,19 @@ stdion_deferred_op(struct gensio_runner *runner, void *cbdata)
 	    stdiona_lock(nadata);
 	}
 	schan->in_open = false;
-	if (schan->outfd_regfile) {
-	    if (schan->read_enabled) {
-		schan->in_read = true;
-		schan->deferred_read = true;
-	    }
-	} else {
-	    nadata->o->set_read_handler(nadata->o, schan->outfd,
-					schan->read_enabled);
-	}
-	if (schan->infd_regfile) {
-	    schan->deferred_write = schan->xmit_enabled;
-	} else if (schan->infd != -1) {
-	    nadata->o->set_write_handler(nadata->o, schan->infd,
-					 schan->xmit_enabled);
-	    nadata->o->set_except_handler(nadata->o, schan->infd,
-					  schan->xmit_enabled);
-	}
-    }
-
-    if (schan->deferred_write) {
-	/* This can only happen if infd_regfile == true. */
-	schan->deferred_write = false;
-	while (schan->xmit_enabled) {
-	    stdiona_unlock(nadata);
-	    gensio_cb(schan->io, GENSIO_EVENT_WRITE_READY, 0, NULL, NULL, NULL);
-	    stdiona_lock(nadata);
+	nadata->o->set_read_handler(schan->out_iod, schan->read_enabled);
+	nadata->o->set_except_handler(schan->out_iod, schan->read_enabled);
+	if (schan->in_iod) {
+	    nadata->o->set_write_handler(schan->in_iod, schan->xmit_enabled);
+	    nadata->o->set_except_handler(schan->in_iod, schan->xmit_enabled);
 	}
     }
 
     if (schan->deferred_read) {
-	int err;
-
 	schan->deferred_read = false;
-    redo_read:
-	err = schan->ll_err;
-	if (!err && schan->outfd_regfile && !schan->data_pending_len)
-	    err = stdion_do_read(nadata, schan);
-	stdion_finish_read(schan, err);
-	if (!err && schan->outfd_regfile && schan->read_enabled)
-	    goto redo_read;
+	while (schan->read_enabled &&
+	       (schan->ll_err || schan->data_pending_len))
+	    stdion_finish_read(schan, 0);
     }
 
     if (schan->deferred_close) {
@@ -460,41 +454,37 @@ stdion_start_deferred_op(struct stdion_channel *schan)
 }
 
 static void
-i_stdion_fd_cleared(int fd, struct stdiona_data *nadata,
+i_stdion_fd_cleared(struct gensio_iod *iod, struct stdiona_data *nadata,
 		    struct stdion_channel *schan)
 {
-    if (fd == schan->infd) {
+    if (iod == schan->in_iod) {
 	schan->in_handler_set = false;
-	schan->infd = -1;
-    } else if (fd == schan->outfd) {
+	nadata->o->close(&schan->in_iod);
+    } else if (iod == schan->out_iod) {
 	schan->out_handler_set = false;
-	schan->outfd = -1;
+	nadata->o->close(&schan->out_iod);
     } else {
 	assert(false);
     }
 
-    if (fd > 2) /* Don't close stdin, stdout, or stderr. */
-	gensio_os_close(nadata->o, &fd);
-
     if (schan->in_close && !schan->in_handler_set && !schan->out_handler_set) {
 	if (schan == &nadata->io && !nadata->err.out_handler_set &&
-		nadata->err.outfd != -1) {
+		nadata->err.out_iod) {
 	    /* The stderr channel is not open, so close the fd. */
-	    gensio_os_close(nadata->o, &nadata->err.outfd);
-	    nadata->err.outfd = -1;
+	    nadata->o->close(&nadata->err.out_iod);
 	}
 	check_waitpid(schan);
     }
 }
 
 static void
-stdion_fd_cleared(int fd, void *cbdata)
+stdion_fd_cleared(struct gensio_iod *iod, void *cbdata)
 {
     struct stdion_channel *schan = cbdata;
     struct stdiona_data *nadata = schan->nadata;
 
     stdiona_lock(nadata);
-    i_stdion_fd_cleared(fd, nadata, schan);
+    i_stdion_fd_cleared(iod, nadata, schan);
     stdiona_deref_and_unlock(nadata);
 }
 
@@ -505,20 +495,19 @@ stdion_set_read_callback_enable(struct gensio *io, bool enabled)
     struct stdiona_data *nadata = schan->nadata;
 
     stdiona_lock(nadata);
-    if (schan->closed || !schan->io)
+    if ((!schan->in_close && schan->closed) || !schan->io)
 	goto out_unlock;
     schan->read_enabled = enabled;
     if (schan->in_read || schan->in_open ||
 			(schan->data_pending_len && !enabled)) {
 	/* Nothing to do, let the read handling wake things up. */
-    } else if (schan->data_pending_len || (enabled && schan->outfd_regfile)) {
+    } else if (schan->data_pending_len) {
 	schan->deferred_read = true;
 	schan->in_read = true;
 	stdion_start_deferred_op(schan);
-    } else if (schan->outfd_regfile) {
-	/* Nothing to do here. */
-    } else {
-	nadata->o->set_read_handler(nadata->o, schan->outfd, enabled);
+    } else if (schan->out_iod) { /* Could be in the middle of close. */
+	nadata->o->set_read_handler(schan->out_iod, enabled);
+	nadata->o->set_except_handler(schan->out_iod, enabled);
     }
  out_unlock:
     stdiona_unlock(nadata);
@@ -531,41 +520,34 @@ stdion_set_write_callback_enable(struct gensio *io, bool enabled)
     struct stdiona_data *nadata = schan->nadata;
 
     stdiona_lock(nadata);
-    if (schan->closed || schan->infd == -1)
+    if ((!schan->in_close && schan->closed) || !schan->in_iod)
 	goto out_unlock;
     if (schan->xmit_enabled == enabled)
 	goto out_unlock;
     schan->xmit_enabled = enabled;
     if (schan->in_open)
 	goto out_unlock;
-    if (schan->infd_regfile) {
-	if (enabled) {
-	    schan->deferred_write = true;
-	    stdion_start_deferred_op(schan);
-	} else {
-	    schan->deferred_write = false;
-	}
-    } else {
-	    nadata->o->set_write_handler(nadata->o, schan->infd, enabled);
-	    nadata->o->set_except_handler(nadata->o, schan->infd, enabled);
+    if (schan->in_iod) { /* Could be in the middle of close. */
+	nadata->o->set_write_handler(schan->in_iod, enabled);
+	nadata->o->set_except_handler(schan->in_iod, enabled);
     }
  out_unlock:
     stdiona_unlock(nadata);
 }
 
 static void
-stdion_read_ready(int fd, void *cbdata)
+stdion_read_ready(struct gensio_iod *iod, void *cbdata)
 {
     struct stdion_channel *schan = cbdata;
     struct stdiona_data *nadata = schan->nadata;
 
     stdiona_lock(nadata);
+    nadata->o->set_read_handler(schan->out_iod, false);
+    nadata->o->set_except_handler(schan->out_iod, false);
     if (!schan->read_enabled || schan->in_read || schan->data_pending_len) {
 	stdiona_unlock(nadata);
 	return;
     }
-    if (!schan->outfd_regfile)
-	nadata->o->set_read_handler(nadata->o, schan->outfd, false);
     if (!schan->ll_err) {
 	schan->in_read = true;
 	stdion_finish_read(schan, stdion_do_read(nadata, schan));
@@ -574,7 +556,13 @@ stdion_read_ready(int fd, void *cbdata)
 }
 
 static void
-stdion_write_ready(int fd, void *cbdata)
+stdion_read_except_ready(struct gensio_iod *iod, void *cbdata)
+{
+    stdion_read_ready(iod, cbdata);
+}
+
+static void
+stdion_write_ready(struct gensio_iod *iod, void *cbdata)
 {
     struct stdion_channel *schan = cbdata;
     struct stdiona_data *nadata = schan->nadata;
@@ -600,200 +588,61 @@ stdion_write_ready(int fd, void *cbdata)
 }
 
 static void
-stdion_write_except_ready(int fd, void *cbdata)
+stdion_write_except_ready(struct gensio_iod *iod, void *cbdata)
 {
-    stdion_write_ready(fd, cbdata);
+    stdion_write_ready(iod, cbdata);
 }
 
-extern char **environ;
-
 static int
 setup_child_proc(struct stdiona_data *nadata)
 {
-    int err;
-    int stdinpipe[2] = {-1, -1};
-    int stdoutpipe[2] = {-1, -1};
-    int stderrpipe[2] = {-1, -1};
+    struct gensio_os_funcs *o = nadata->o;
+    int rv;
 
-    err = pipe(stdinpipe);
-    if (err) {
-	err = errno;
-	goto out_err;
-    }
-
-    err = pipe(stdoutpipe);
-    if (err) {
-	err = errno;
-	goto out_err;
-    }
-
-    nadata->io.infd = stdinpipe[1];
-    nadata->io.outfd = stdoutpipe[0];
-    if (fcntl(nadata->io.infd, F_SETFL, O_NONBLOCK) == -1) {
-	err = errno;
-	goto out_err;
-    }
-    if (fcntl(nadata->io.outfd, F_SETFL, O_NONBLOCK) == -1) {
-	err = errno;
-	goto out_err;
-    }
-
-    nadata->err.infd = -1;
-
-    if (nadata->stderr_to_stdout) {
-	nadata->err.outfd = stdoutpipe[0];
-	stderrpipe[0] = stdoutpipe[0];
-	stderrpipe[1] = stdoutpipe[1];
-    } else if (nadata->noredir_stderr) {
-	nadata->err.outfd = -1;
-    } else {
-	err = pipe(stderrpipe);
-	if (err) {
-	    err = errno;
-	    goto out_err;
-	}
-	nadata->err.outfd = stderrpipe[0];
-	if (fcntl(nadata->err.outfd, F_SETFL, O_NONBLOCK) == -1) {
-	    err = errno;
-	    goto out_err;
-	}
-    }
-
-    nadata->opid = fork();
-    if (nadata->opid < 0) {
-	err = errno;
-	goto out_err;
-    }
-    if (nadata->opid == 0) {
-	int i, openfiles = sysconf(_SC_OPEN_MAX);
-
-	dup2(stdinpipe[0], 0);
-	dup2(stdoutpipe[1], 1);
-	if (!nadata->noredir_stderr)
-	    dup2(stderrpipe[1], 2);
-
-	/* Close everything but stdio. */
-	for (i = 3; i < openfiles; i++)
-	    close(i);
-
-	err = gensio_os_setupnewprog();
-	if (err) {
-	    fprintf(stderr, "Unable to set groups or user: %s\r\n",
-		    strerror(err));
-	    exit(1);
-	}
-
-	if (nadata->env)
-	    environ = (char **) nadata->env;
-
-	execvp(nadata->argv[0], (char * const *) nadata->argv);
-	fprintf(stderr, "Err: %s %s\r\n", nadata->argv[0], strerror(errno));
-	exit(1); /* Only reached on error. */
-    }
-
-    close(stdinpipe[0]);
-    close(stdoutpipe[1]);
-    if (stdoutpipe[1] != stderrpipe[1])
-	close(stderrpipe[1]);
-    return 0;
-
- out_err:
-    if (stdinpipe[0] != -1)
-	close(stdinpipe[0]);
-    if (stdinpipe[1] != -1)
-	close(stdinpipe[1]);
-    if (stdoutpipe[0] != -1)
-	close(stdoutpipe[0]);
-    if (stdoutpipe[1] != -1)
-	close(stdoutpipe[1]);
-    if (stderrpipe[0] != -1 && stderrpipe[0] != stdoutpipe[0])
-	close(stderrpipe[0]);
-    if (stderrpipe[1] != -1 && stderrpipe[1] != stdoutpipe[1])
-	close(stderrpipe[1]);
-
-    return gensio_os_err_to_err(nadata->o, err);
+    rv = o->exec_subprog(o, nadata->argv, nadata->env, nadata->stderr_to_stdout,
+			 &nadata->opid, &nadata->io.in_iod, &nadata->io.out_iod,
+			 nadata->noredir_stderr ? NULL : &nadata->err.out_iod);
+    return rv;
 }
 
 static int
 setup_io_self(struct stdiona_data *nadata)
 {
+    struct gensio_os_funcs *o = nadata->o;
     int rv;
-    struct stat statb;
 
-    /*
-     * Figure out if the files are regular files.  If they are, they
-     * are handled differently.
-     */
-    rv = fstat(nadata->io.infd, &statb);
-    if (rv == -1) {
-	rv = gensio_os_err_to_err(nadata->o, errno);
-	goto out_err;
+    if (nadata->raw) {
+	rv = o->makeraw(nadata->io.in_iod);
+	if (rv)
+	    return rv;
+	rv = o->makeraw(nadata->io.out_iod);
+	if (rv)
+	    return rv;
     }
-    nadata->io.infd_regfile = (statb.st_mode & S_IFMT) == S_IFREG;
-
-    rv = fstat(nadata->io.outfd, &statb);
-    if (rv == -1) {
-	rv = gensio_os_err_to_err(nadata->o, errno);
-	goto out_err;
-    }
-    nadata->io.outfd_regfile = (statb.st_mode & S_IFMT) == S_IFREG;
 
     /*
      * If these are not regular files, save off the old flags and turn
      * on non-blocking.
      */
-    if (!nadata->io.infd_regfile) {
-	rv = fcntl(nadata->io.infd, F_GETFL, 0);
-	if (rv == -1) {
-	    rv = gensio_os_err_to_err(nadata->o, errno);
-	    goto out_err;
-	}
-	nadata->old_flags_ostdin = rv;
-	if (fcntl(nadata->io.infd, F_SETFL, O_NONBLOCK) == -1) {
-	    rv = gensio_os_err_to_err(nadata->o, errno);
-	    goto out_err;
-	}
-	nadata->old_flags_ostdin_set = true;
-    }
+    rv = o->set_non_blocking(nadata->io.in_iod);
+    if (rv)
+	return rv;
 
-    if (!nadata->io.outfd_regfile) {
-	rv = fcntl(nadata->io.outfd, F_GETFL, 0);
-	if (rv == -1) {
-	    rv = gensio_os_err_to_err(nadata->o, errno);
-	    goto out_err;
-	}
-	nadata->old_flags_ostdout = rv;
+    rv = o->set_non_blocking(nadata->io.out_iod);
+    if (rv)
+	return rv;
 
-	if (fcntl(nadata->io.outfd, F_SETFL, O_NONBLOCK) == -1) {
-	    rv = gensio_os_err_to_err(nadata->o, errno);
-	    goto out_err;
-	}
-	nadata->old_flags_ostdout_set = true;
-    }
-    rv = 0;
-
- out_err:
-    if (rv && nadata->old_flags_ostdin_set) {
-	fcntl(nadata->io.infd, F_SETFL, nadata->old_flags_ostdin);
-	nadata->old_flags_ostdin_set = false;
-    }
-    return rv;
+    return 0;
 }
 
 static void
 cleanup_io_self(struct stdiona_data *nadata)
 {
-    if (nadata->old_flags_ostdin_set)
-	fcntl(nadata->io.infd, F_SETFL, nadata->old_flags_ostdin);
-    nadata->old_flags_ostdin_set = false;
-    if (nadata->old_flags_ostdout_set)
-	fcntl(nadata->io.outfd, F_SETFL, nadata->old_flags_ostdout);
-    nadata->old_flags_ostdout_set = false;
     if (nadata->io.out_handler_set)
-	nadata->o->clear_fd_handlers_norpt(nadata->o, nadata->io.outfd);
+	nadata->o->clear_fd_handlers_norpt(nadata->io.out_iod);
     nadata->io.out_handler_set = false;
     if (nadata->io.in_handler_set)
-	nadata->o->clear_fd_handlers_norpt(nadata->o, nadata->io.infd);
+	nadata->o->clear_fd_handlers_norpt(nadata->io.in_iod);
     nadata->io.in_handler_set = false;
 }
 
@@ -820,22 +669,21 @@ stdion_open(struct gensio *io, gensio_done_err open_done, void *open_data)
 	    goto out_err;
     }
 
-    if (!schan->outfd_regfile) {
-	err = nadata->o->set_fd_handlers(nadata->o, schan->outfd, schan,
-					 stdion_read_ready, NULL, NULL,
-					 stdion_fd_cleared);
-	if (err)
-	    goto out_err;
-	schan->out_handler_set = true;
-	stdiona_ref(nadata);
-    }
+    err = nadata->o->set_fd_handlers(schan->out_iod, schan,
+				     stdion_read_ready, NULL,
+				     stdion_read_except_ready,
+				     stdion_fd_cleared);
+    if (err)
+	goto out_err;
+    schan->out_handler_set = true;
+    stdiona_ref(nadata);
 
-    if (!schan->infd_regfile && schan->infd != -1) {
+    if (schan->in_iod) {
 	/*
 	 * On the write side we send an exception to the write ready
 	 * operation.
 	 */
-	err = nadata->o->set_fd_handlers(nadata->o, schan->infd, schan,
+	err = nadata->o->set_fd_handlers(schan->in_iod, schan,
 					 NULL, stdion_write_ready,
 					 stdion_write_except_ready,
 					 stdion_fd_cleared);
@@ -859,16 +707,12 @@ stdion_open(struct gensio *io, gensio_done_err open_done, void *open_data)
     stdiona_deref(nadata);
  out_err:
     cleanup_io_self(nadata);
-    if (nadata->io.infd != -1 && nadata->io.infd != 1)
-	gensio_os_close(nadata->o, &nadata->io.infd);
-    nadata->io.infd = -1;
-    if (nadata->err.outfd != -1 && nadata->err.outfd != 2 &&
-		nadata->err.outfd != nadata->io.outfd)
-	gensio_os_close(nadata->o, &nadata->err.outfd);
-    nadata->err.outfd = -1;
-    if (nadata->io.outfd != -1 && nadata->io.outfd != 0)
-	gensio_os_close(nadata->o, &nadata->io.outfd);
-    nadata->io.outfd = -1;
+    if (nadata->io.in_iod)
+	nadata->o->close(&nadata->io.in_iod);
+    if (nadata->err.out_iod)
+	nadata->o->close(&nadata->err.out_iod);
+    if (nadata->io.out_iod)
+	nadata->o->close(&nadata->io.out_iod);
  out_unlock:
     stdiona_unlock(nadata);
 
@@ -877,8 +721,8 @@ stdion_open(struct gensio *io, gensio_done_err open_done, void *open_data)
 
 static int
 stdion_alloc_channel(struct gensio *io, const char * const args[],
-		    gensio_event cb, void *user_data,
-		    struct gensio **new_io)
+		     gensio_event cb, void *user_data,
+		     struct gensio **new_io)
 {
     struct stdion_channel *schan = gensio_get_gensio_data(io);
     struct stdiona_data *nadata = schan->nadata;
@@ -886,7 +730,7 @@ stdion_alloc_channel(struct gensio *io, const char * const args[],
     unsigned int i;
     gensiods max_read_size = nadata->io.max_read_size;
 
-    if (nadata->stderr_to_stdout || nadata->noredir_stderr)
+    if (!nadata->err.out_iod || io != nadata->io.io)
 	return GE_INVAL;
 
     for (i = 0; args && args[i]; i++) {
@@ -896,14 +740,6 @@ stdion_alloc_channel(struct gensio *io, const char * const args[],
     }
 
     stdiona_lock(nadata);
-    if (io != nadata->io.io) {
-	rv = GE_INVAL;
-	goto out_err;
-    }
-    if (nadata->err.outfd == -1) {
-	rv = GE_NOTFOUND;
-	goto out_err;
-    }
     if (nadata->err.io) {
 	rv = GE_INUSE;
 	goto out_err;
@@ -1025,25 +861,19 @@ stdion_disable(struct gensio *io)
     schan->in_close = false;
     schan->in_open = false;
     schan->close_done = NULL;
-    if (nadata->io.out_handler_set) {
-	nadata->o->clear_fd_handlers_norpt(nadata->o, nadata->io.outfd);
-	if (nadata->io.outfd != 0)
-	    gensio_os_close(nadata->o, &nadata->io.outfd);
-	nadata->io.outfd = -1;
-    }
-    if (nadata->io.in_handler_set) {
-	nadata->o->clear_fd_handlers_norpt(nadata->o, nadata->io.infd);
-	if (nadata->io.infd != 1)
-	    gensio_os_close(nadata->o, &nadata->io.infd);
-	nadata->io.infd = -1;
-    }
-    if (nadata->err.out_handler_set) {
-	nadata->o->clear_fd_handlers_norpt(nadata->o, nadata->err.outfd);
-	if (nadata->err.outfd != 2)
-	    gensio_os_close(nadata->o, &nadata->err.outfd);
-	nadata->err.outfd = -1;
-    }
-    stdiona_deref_and_unlock(nadata); /* unlocks */
+    if (nadata->io.out_handler_set)
+	nadata->o->clear_fd_handlers_norpt(nadata->io.out_iod);
+    if (nadata->io.out_iod)
+	nadata->o->close(&nadata->io.out_iod);
+    if (nadata->io.in_handler_set)
+	nadata->o->clear_fd_handlers_norpt(nadata->io.in_iod);
+    if (nadata->io.in_iod)
+	nadata->o->close(&nadata->io.in_iod);
+    if (nadata->err.out_handler_set)
+	nadata->o->clear_fd_handlers_norpt(nadata->err.out_iod);
+    if (nadata->err.out_iod)
+	nadata->o->close(&nadata->err.out_iod);
+    stdiona_deref_and_unlock(nadata);
     return 0;
 }
 
@@ -1053,63 +883,85 @@ stdion_control(struct gensio *io, bool get, unsigned int option,
 {
     struct stdion_channel *schan = gensio_get_gensio_data(io);
     struct stdiona_data *nadata = schan->nadata;
+    struct gensio_os_funcs *o = nadata->o;
     const char **env, **argv;
-    int err, status;
+    int err, status, val;
     gensiods pos;
 
     switch (option) {
     case GENSIO_CONTROL_ENVIRONMENT:
 	if (get)
 	    return GE_NOTSUP;
-	err = gensio_argv_copy(nadata->o, (const char **) data, NULL, &env);
+	err = gensio_argv_copy(o, (const char **) data, NULL, &env);
 	if (err)
 	    return err;
 	if (nadata->env)
-	    gensio_argv_free(nadata->o, nadata->env);
+	    gensio_argv_free(o, nadata->env);
 	nadata->env = env;
 	return 0;
 
     case GENSIO_CONTROL_ARGS:
 	if (get)
 	    return GE_NOTSUP;
-	err = gensio_argv_copy(nadata->o, (const char **) data, NULL, &argv);
+	err = gensio_argv_copy(o, (const char **) data, NULL, &argv);
 	if (err)
 	    return err;
 	if (nadata->argv)
-	    gensio_argv_free(nadata->o, nadata->argv);
+	    gensio_argv_free(o, nadata->argv);
 	nadata->argv = argv;
 	return 0;
 
     case GENSIO_CONTROL_EXIT_CODE:
 	if (!get)
 	    return GE_NOTSUP;
+	err = 0;
+	stdiona_lock(nadata);
 	if (!nadata->exit_code_set)
-	    return GE_NOTREADY;
-	*datalen = snprintf(data, *datalen, "%d", nadata->exit_code);
-	return 0;
+	    err = GE_NOTREADY;
+	stdiona_unlock(nadata);
+	if (!err)
+	    *datalen = snprintf(data, *datalen, "%d", nadata->exit_code);
+	return err;
 
     case GENSIO_CONTROL_WAIT_TASK:
 	if (!get)
 	    return GE_NOTSUP;
+	stdiona_lock(nadata);
 	if (nadata->opid == -1)
-	    return GE_NOTREADY;
-	err = waitpid(nadata->opid, &status, WNOHANG | WNOWAIT);
-	if (err <= 0)
-	    return GE_NOTREADY;
-	*datalen = snprintf(data, *datalen, "%d", status);
+	    err = GE_NOTREADY;
+	else
+	    err = o->wait_subprog(o, nadata->opid, &status);
+	if (!err) {
+	    nadata->opid = -1;
+	    nadata->exit_code = status;
+	}
+	stdiona_unlock(nadata);
+	if (!err)
+	    *datalen = snprintf(data, *datalen, "%d", status);
 	return 0;
 
+    case GENSIO_CONTROL_KILL_TASK:
+	if (get)
+	    return GE_NOTSUP;
+	stdiona_lock(nadata);
+	if (nadata->opid == -1) {
+	    err = GE_NOTREADY;
+	} else {
+	    val = strtoul(data, NULL, 0);
+	    err = o->kill_subprog(o, nadata->opid, val);
+	}
+	stdiona_unlock(nadata);
+	return err;
+
     case GENSIO_CONTROL_CLOSE_OUTPUT:
 	if (get)
 	    return GE_NOTSUP;
 	err = 0;
 	stdiona_lock(nadata);
-	if (schan->infd == -1) {
+	if (!schan->in_iod)
 	    err = GE_NOTREADY;
-	} else {
-	    nadata->o->clear_fd_handlers(nadata->o, schan->infd);
-	    schan->infd = -1;
-	}
+	else
+	    o->clear_fd_handlers(schan->in_iod);
 	stdiona_unlock(nadata);
 	return err;
 
@@ -1133,7 +985,8 @@ stdion_control(struct gensio *io, bool get, unsigned int option,
     case GENSIO_CONTROL_REMOTE_ID:
 	if (!get)
 	    return GE_NOTSUP;
-	*datalen = snprintf(data, *datalen, "%d", nadata->opid);
+	*datalen = snprintf(data, *datalen, "%llu",
+			    (unsigned long long) nadata->opid);
 	return 0;
     }
 
@@ -1150,10 +1003,10 @@ gensio_stdio_func(struct gensio *io, int func, gensiods *count,
 	return stdion_write(io, count, cbuf, buflen);
 
     case GENSIO_FUNC_OPEN:
-	return stdion_open(io, cbuf, buf);
+	return stdion_open(io, (void *) cbuf, buf);
 
     case GENSIO_FUNC_CLOSE:
-	return stdion_close(io, cbuf, buf);
+	return stdion_close(io, (void *) cbuf, buf);
 
     case GENSIO_FUNC_FREE:
 	stdion_free(io);
@@ -1191,7 +1044,7 @@ gensio_stdio_func(struct gensio *io, int func, gensiods *count,
 
 static int
 stdio_nadata_setup(struct gensio_os_funcs *o, gensiods max_read_size,
-		   struct stdiona_data **new_nadata)
+		   bool raw, struct stdiona_data **new_nadata)
 {
     struct stdiona_data *nadata;
 
@@ -1206,10 +1059,6 @@ stdio_nadata_setup(struct gensio_os_funcs *o, gensiods max_read_size,
     nadata->err.closed = true;
     nadata->io.nadata = nadata;
     nadata->err.nadata = nadata;
-    nadata->io.infd = -1;
-    nadata->io.outfd = -1;
-    nadata->io.infd = -1;
-    nadata->io.outfd = -1;
     nadata->opid = -1;
 
     nadata->waitpid_timer = o->alloc_timer(o, check_waitpid_timeout,
@@ -1217,6 +1066,7 @@ stdio_nadata_setup(struct gensio_os_funcs *o, gensiods max_read_size,
     if (!nadata->waitpid_timer)
 	goto out_nomem;
 
+    nadata->raw = raw;
     nadata->io.max_read_size = max_read_size;
     nadata->io.read_data = o->zalloc(o, max_read_size);
     if (!nadata->io.read_data)
@@ -1246,6 +1096,27 @@ stdio_nadata_setup(struct gensio_os_funcs *o, gensiods max_read_size,
     return GE_NOMEM;
 }
 
+static int
+setup_self(struct stdiona_data *nadata, bool console)
+{
+    struct gensio_os_funcs *o = nadata->o;
+    int err;
+    enum gensio_iod_type type;
+
+    if (console)
+	type = GENSIO_IOD_CONSOLE;
+    else
+	type = GENSIO_IOD_STDIO;
+    err = o->add_iod(o, type, 1, &nadata->io.in_iod);
+    if (err)
+	return err;
+    err = o->add_iod(o, type, 0, &nadata->io.out_iod);
+    if (err)
+	return err;
+
+    return 0;
+}
+
 int
 stdio_gensio_alloc(const char * const argv[], const char * const args[],
 		   struct gensio_os_funcs *o,
@@ -1256,38 +1127,51 @@ stdio_gensio_alloc(const char * const argv[], const char * const args[],
     struct stdiona_data *nadata = NULL;
     int i;
     gensiods max_read_size = GENSIO_DEFAULT_BUF_SIZE;
-    bool self= false;
+    bool self = false;
+    bool console = false;
     bool stderr_to_stdout = false;
     bool noredir_stderr = false;
+    bool raw = false;
 
     for (i = 0; args && args[i]; i++) {
 	if (gensio_check_keyds(args[i], "readbuf", &max_read_size) > 0)
 	    continue;
+	if (gensio_check_keybool(args[i], "console", &console) > 0)
+	    continue;
 	if (gensio_check_keybool(args[i], "self", &self) > 0)
 	    continue;
-	if (gensio_check_keybool(args[i], "stderr-to-stdout",
-				 &stderr_to_stdout) > 0)
+	if (gensio_check_keybool(args[i], "raw", &raw) > 0)
 	    continue;
+	if (gensio_check_keybool(args[i], "stderr-to-stdout",
+				 &stderr_to_stdout) > 0) {
+	    /* We don't want to setup stderr here. */
+	    noredir_stderr = true;
+	    continue;
+	}
 	if (gensio_check_keybool(args[i], "noredir-stderr",
 				 &noredir_stderr) > 0)
 	    continue;
 	return GE_INVAL;
     }
 
-    err = stdio_nadata_setup(o, max_read_size, &nadata);
+    if (raw && !(self || console))
+	return GE_INVAL;
+
+    err = stdio_nadata_setup(o, max_read_size, raw, &nadata);
     if (err)
 	return err;
 
     nadata->stderr_to_stdout = stderr_to_stdout;
     nadata->noredir_stderr = noredir_stderr;
 
-    if (self) {
-	nadata->io.infd = 1;
-	nadata->io.outfd = 0;
+    if (self || console) {
+	err = setup_self(nadata, console);
+	if (err)
+	    goto out_err;
     } else {
 	err = gensio_argv_copy(o, argv, NULL, &nadata->argv);
 	if (err)
-	    goto out_nomem;
+	    goto out_err;
     }
 
     nadata->io.io = gensio_data_alloc(nadata->o, cb, user_data,
@@ -1303,8 +1187,10 @@ stdio_gensio_alloc(const char * const argv[], const char * const args[],
     return 0;
 
  out_nomem:
+    err = GE_NOMEM;
+ out_err:
     stdiona_finish_free(nadata);
-    return GE_NOMEM;
+    return err;
 }
 
 int
@@ -1358,23 +1244,17 @@ stdiona_do_connect(struct gensio_runner *runner, void *cbdata)
  * fd cleared for a gensio from an acceptor only.
  */
 static void
-stdiona_fd_cleared(int fd, void *cbdata)
+stdiona_fd_cleared(struct gensio_iod *iod, void *cbdata)
 {
     struct stdion_channel *schan = cbdata;
     struct stdiona_data *nadata = schan->nadata;
 
     stdiona_lock(nadata);
-    if (fd == schan->infd)
+    if (iod == schan->in_iod)
 	schan->in_handler_set = false;
     else
 	schan->out_handler_set = false;
 
-    if (!nadata->io.in_handler_set && !nadata->io.out_handler_set) {
-	/* We came from an acceptor, set stdio back to original values. */
-	fcntl(nadata->io.infd, F_SETFL, nadata->old_flags_ostdin);
-	fcntl(nadata->io.outfd, F_SETFL, nadata->old_flags_ostdout);
-    }
-
     if (!schan->in_handler_set && !schan->out_handler_set && schan->in_close) {
 	schan->in_close = false;
 	if (schan->close_done) {
@@ -1414,7 +1294,7 @@ stdiona_startup(struct gensio_accepter *accepter)
     if (rv)
 	goto out_unlock;
 
-    rv = nadata->o->set_fd_handlers(nadata->o, nadata->io.infd,
+    rv = nadata->o->set_fd_handlers(nadata->io.in_iod,
 				    &nadata->io, NULL, stdion_write_ready, NULL,
 				    stdiona_fd_cleared);
     if (rv)
@@ -1422,7 +1302,7 @@ stdiona_startup(struct gensio_accepter *accepter)
     nadata->io.in_handler_set = true;
     stdiona_ref(nadata);
 
-    rv = nadata->o->set_fd_handlers(nadata->o, nadata->io.outfd,
+    rv = nadata->o->set_fd_handlers(nadata->io.out_iod,
 				    &nadata->io, stdion_read_ready, NULL, NULL,
 				    stdiona_fd_cleared);
     if (rv)
@@ -1576,15 +1456,18 @@ stdio_gensio_accepter_alloc(const char * const args[],
     int err;
     struct stdiona_data *nadata = NULL;
     gensiods max_read_size = GENSIO_DEFAULT_BUF_SIZE;
+    bool raw = false;
     int i;
 
     for (i = 0; args && args[i]; i++) {
 	if (gensio_check_keyds(args[i], "readbuf", &max_read_size) > 0)
 	    continue;
+	if (gensio_check_keybool(args[i], "raw", &raw) > 0)
+	    continue;
 	return GE_INVAL;
     }
 
-    err = stdio_nadata_setup(o, max_read_size, &nadata);
+    err = stdio_nadata_setup(o, max_read_size, raw, &nadata);
     if (err)
 	return err;
 
@@ -1594,8 +1477,17 @@ stdio_gensio_accepter_alloc(const char * const args[],
 	return GE_NOMEM;
     }
 
-    nadata->io.infd = 1;
-    nadata->io.outfd = 0;
+    err = setup_self(nadata, false);
+    if (err) {
+	stdiona_finish_free(nadata);
+	return err;
+    }
+
+    err = o->add_iod(o, GENSIO_IOD_STDIO, 0, &nadata->io.out_iod);
+    if (err) {
+	stdiona_finish_free(nadata);
+	return err;
+    }
 
     nadata->acc = gensio_acc_data_alloc(o, cb, user_data, gensio_acc_stdio_func,
 					NULL, "stdio", nadata);
