@@ -46,7 +46,8 @@ struct udpn_data {
 
     unsigned int refcount;
 
-    int myfd; /* fd the original request came in on, for sending. */
+    /* iod the original request came in on, for sending. */
+    struct gensio_iod *myiod;
 
     bool read_enabled;	/* Read callbacks are enabled. */
     bool write_enabled;	/* Write callbacks are enabled. */
@@ -114,6 +115,7 @@ struct udpna_data {
 
     unsigned char *read_data;
 
+    bool readhandler_read_disabled;
     gensiods data_pending_len;
     gensiods data_pos;
     struct udpn_data *pending_data_owner;
@@ -142,7 +144,7 @@ struct udpna_data {
     void *shutdown_data;
 
     struct gensio_addr *ai;		/* The address list for the portname. */
-    struct opensocks   *fds;		/* The file descriptor used for
+    struct gensio_opensocks *fds;	/* The file descriptor used for
 					   the UDP ports. */
     unsigned int   nr_fds;
     unsigned int opensock_flags;
@@ -324,7 +326,7 @@ udpna_enable_read(struct udpna_data *nadata)
 
     nadata->read_disabled = false;
     for (i = 0; i < nadata->nr_fds; i++)
-	nadata->o->set_read_handler(nadata->o, nadata->fds[i].fd, true);
+	nadata->o->set_read_handler(nadata->fds[i].iod, true);
 }
 
 static void
@@ -334,7 +336,7 @@ udpna_disable_read(struct udpna_data *nadata)
 
     nadata->read_disabled = true;
     for (i = 0; i < nadata->nr_fds; i++)
-	nadata->o->set_read_handler(nadata->o, nadata->fds[i].fd, false);
+	nadata->o->set_read_handler(nadata->fds[i].iod, false);
 }
 
 static void udpna_check_read_state(struct udpna_data *nadata)
@@ -370,7 +372,7 @@ udpna_disable_write(struct udpna_data *nadata)
     unsigned int i;
 
     for (i = 0; i < nadata->nr_fds; i++)
-	nadata->o->set_write_handler(nadata->o, nadata->fds[i].fd, false);
+	nadata->o->set_write_handler(nadata->fds[i].iod, false);
 }
 
 static void
@@ -388,7 +390,7 @@ udpna_enable_write(struct udpna_data *nadata)
     unsigned int i;
 
     for (i = 0; i < nadata->nr_fds; i++)
-	nadata->o->set_write_handler(nadata->o, nadata->fds[i].fd, true);
+	nadata->o->set_write_handler(nadata->fds[i].iod, true);
 }
 
 static void
@@ -405,8 +407,8 @@ udpna_do_free(struct udpna_data *nadata)
     unsigned int i;
 
     for (i = 0; i < nadata->nr_fds; i++) {
-	if (nadata->fds[i].fd != -1)
-	    gensio_os_close(nadata->o, &nadata->fds[i].fd);
+	if (nadata->fds && nadata->fds[i].iod)
+	    nadata->o->close(&nadata->fds[i].iod);
     }
 
     if (nadata->deferred_op_runner)
@@ -427,7 +429,7 @@ udpna_do_free(struct udpna_data *nadata)
 }
 
 static void
-udpna_fd_cleared(int fd, void *cbdata)
+udpna_fd_cleared(struct gensio_iod *iod, void *cbdata)
 {
     struct udpna_data *nadata = cbdata;
 
@@ -450,7 +452,7 @@ udpna_check_finish_free(struct udpna_data *nadata)
     udpna_deref(nadata);
     for (i = 0; i < nadata->nr_fds; i++) {
 	udpna_ref(nadata);
-	nadata->o->clear_fd_handlers(nadata->o, nadata->fds[i].fd);
+	nadata->o->clear_fd_handlers(nadata->fds[i].iod);
     }
 }
 
@@ -506,7 +508,7 @@ udpn_write(struct gensio *io, gensiods *count,
     if (!addr)
 	addr = ndata->raddr;
 
-    err = gensio_os_sendto(ndata->o, ndata->myfd, sg, sglen, count, 0, addr);
+    err = ndata->o->sendto(ndata->myiod, sg, sglen, count, 0, addr);
     if (free_addr)
 	gensio_addr_free(addr);
     return err;
@@ -900,16 +902,17 @@ udpn_handle_write_incoming(struct udpna_data *nadata, struct udpn_data *ndata)
 }
 
 static void
-udpna_writehandler(int fd, void *cbdata)
+udpna_writehandler(struct gensio_iod *iod, void *cbdata)
 {
     struct udpna_data *nadata = cbdata;
     struct gensio_link *l;
 
     udpna_lock_and_ref(nadata);
-    if (nadata->in_write)
+    if (nadata->in_write) {
+	udpna_disable_write(nadata);
 	goto out_unlock;
+    }
 
-    udpna_disable_write(nadata);
     gensio_list_for_each(&nadata->udpns, l) {
 	struct udpn_data *ndata = gensio_link_to_ndata(l);
 
@@ -947,7 +950,9 @@ udpna_control_laddr(struct udpna_data *nadata, bool get,
     if (i >= nadata->nr_fds)
 	return GE_NOTFOUND;
 
-    rv = gensio_os_getsockname(nadata->o, nadata->fds[i].fd, &addr);
+    rv = nadata->o->sock_control(nadata->fds[i].iod,
+				 GENSIO_SOCKCTL_GET_SOCKNAME,
+				 &addr, NULL);
     if (rv)
 	return rv;
 
@@ -989,6 +994,7 @@ udpna_control_lport(struct udpna_data *nadata, bool get,
 {
     int rv;
     unsigned int i;
+    gensiods size;
 
     if (!get)
 	return GE_NOTSUP;
@@ -1000,7 +1006,9 @@ udpna_control_lport(struct udpna_data *nadata, bool get,
     if (i >= nadata->nr_fds)
 	return GE_NOTFOUND;
 
-    rv = gensio_os_socket_get_port(nadata->o, nadata->fds[i].fd, &i);
+    size = sizeof(unsigned int);
+    rv = nadata->o->sock_control(nadata->fds[i].iod, GENSIO_SOCKCTL_GET_PORT,
+				 &i, &size);
     if (rv)
 	return rv;
     *datalen = snprintf(data, *datalen, "%d", i);
@@ -1013,8 +1021,8 @@ udpn_control(struct gensio *io, bool get, int option,
 {
     struct udpn_data *ndata = gensio_get_gensio_data(io);
     struct udpna_data *nadata = ndata->nadata;
+    struct gensio_os_funcs *o = nadata->o;
     int err;
-    struct gensio_addr *addr;
 
     switch(option) {
     case GENSIO_CONTROL_MAX_WRITE_PACKET:
@@ -1038,27 +1046,75 @@ udpn_control(struct gensio *io, bool get, int option,
 	if (!get)
 	    return GE_NOTSUP;
 	gensio_addr_getaddr(ndata->raddr, data, datalen);
-	return 0;
+	break;
 
     case GENSIO_CONTROL_LPORT:
 	return udpna_control_lport(nadata, get, data, datalen);
 
     case GENSIO_CONTROL_ADD_MCAST:
-    case GENSIO_CONTROL_DEL_MCAST:
+    case GENSIO_CONTROL_DEL_MCAST: {
+	struct gensio_addr *addr;
+
 	err = gensio_scan_network_addr(nadata->o, data,
 				       GENSIO_NET_PROTOCOL_UDP, &addr);
 	if (err)
 	    return err;
 	if (option == GENSIO_CONTROL_ADD_MCAST)
-	    err = gensio_os_mcast_add(nadata->o, nadata->fds->fd, addr,
-				      0, false);
+	    err = nadata->o->mcast_add(nadata->fds->iod, addr, 0, false);
 	else
-	    err = gensio_os_mcast_del(nadata->o, nadata->fds->fd, addr,
-				      0, false);
+	    err = nadata->o->mcast_del(nadata->fds->iod, addr, 0, false);
 	gensio_addr_free(addr);
-	if (err)
-	    return err;
+	return err;
+    }
+
+    case GENSIO_CONTROL_MCAST_LOOP: {
+	bool mcast_loop;
+	gensiods size = sizeof(mcast_loop);
+	struct gensio_iod *iod = nadata->fds->iod;
+
+	if (get) {
+	    err = o->sock_control(iod, GENSIO_SOCKCTL_GET_MCAST_LOOP,
+				  &mcast_loop, &size);
+	    if (err)
+		return err;
+	    if (mcast_loop)
+		*datalen = snprintf(data, *datalen, "true");
+	    else
+		*datalen = snprintf(data, *datalen, "false");
+	} else {
+	    if (strncasecmp(data, "true", *datalen))
+		mcast_loop = true;
+	    else if (strncasecmp(data, "false", *datalen))
+		mcast_loop = false;
+	    else
+		return GE_INVAL;
+	    return o->sock_control(iod, GENSIO_SOCKCTL_SET_MCAST_LOOP,
+				   &mcast_loop, &size);
+	}
 	break;
+    }
+
+    case GENSIO_CONTROL_MCAST_TTL: {
+	unsigned int ttl;
+	gensiods size = sizeof(ttl);
+	struct gensio_iod *iod = nadata->fds->iod;
+
+	if (get) {
+	    err = o->sock_control(iod, GENSIO_SOCKCTL_SET_MCAST_TTL,
+				  &ttl, &size);
+	    if (err)
+		return err;
+	    *datalen = snprintf(data, *datalen, "%u", ttl);
+	} else {
+	    unsigned int ttl;
+	    gensiods size = sizeof(ttl);
+
+	    ttl = strtoul(data, NULL, 0);
+	    return o->sock_control(iod, GENSIO_SOCKCTL_SET_MCAST_TTL,
+				   &ttl, &size);
+	}
+	break;
+    }
 
     default:
 	return GE_NOTSUP;
@@ -1076,10 +1132,10 @@ gensio_udp_func(struct gensio *io, int func, gensiods *count,
 	return udpn_write(io, count, cbuf, buflen, auxdata);
 
     case GENSIO_FUNC_OPEN:
-	return udpn_open(io, cbuf, buf);
+	return udpn_open(io, (void *) cbuf, buf);
 
     case GENSIO_FUNC_CLOSE:
-	return udpn_close(io, cbuf, buf);
+	return udpn_close(io, (void *) cbuf, buf);
 
     case GENSIO_FUNC_FREE:
 	udpn_free(io);
@@ -1110,7 +1166,7 @@ gensio_udp_func(struct gensio *io, int func, gensiods *count,
 }
 
 static struct udpn_data *
-udp_alloc_gensio(struct udpna_data *nadata, int fd,
+udp_alloc_gensio(struct udpna_data *nadata, struct gensio_iod *iod,
 		 const struct gensio_addr *addr,
 		 gensio_event cb, void *user_data,
 		 struct gensio_list *starting_list)
@@ -1148,7 +1204,7 @@ udp_alloc_gensio(struct udpna_data *nadata, int fd,
     }
     gensio_set_is_packet(ndata->io, true);
 
-    ndata->myfd = fd;
+    ndata->myiod = iod;
 
     /* Stick it on the end of the list. */
     udpn_add_to_list(starting_list, ndata);
@@ -1158,7 +1214,7 @@ udp_alloc_gensio(struct udpna_data *nadata, int fd,
 }
 
 static void
-udpna_readhandler(int fd, void *cbdata)
+udpna_readhandler(struct gensio_iod *iod, void *cbdata)
 {
     struct udpna_data *nadata = cbdata;
     struct udpn_data *ndata;
@@ -1167,12 +1223,14 @@ udpna_readhandler(int fd, void *cbdata)
     int err;
 
     udpna_lock_and_ref(nadata);
-    if (nadata->data_pending_len)
+    if (nadata->data_pending_len) {
+	nadata->readhandler_read_disabled = true;
+	udpna_fd_read_disable(nadata);
 	goto out_unlock;
+    }
 
-    err = gensio_os_recvfrom(nadata->o,
-			     fd, nadata->read_data, nadata->max_read_size,
-			     &datalen, 0, nadata->curr_recvaddr);
+    err = nadata->o->recvfrom(iod, nadata->read_data, nadata->max_read_size,
+			      &datalen, 0, nadata->curr_recvaddr);
     if (err) {
 	if (!nadata->is_dummy)
 	    /* Don't log on dummy accepters. */
@@ -1184,8 +1242,6 @@ udpna_readhandler(int fd, void *cbdata)
     if (datalen == 0)
 	goto out_unlock;
 
-    udpna_fd_read_disable(nadata);
-
     nadata->data_pending_len = datalen;
     nadata->data_pos = 0;
 
@@ -1210,7 +1266,7 @@ udpna_readhandler(int fd, void *cbdata)
     }
 
     /* New connection. */
-    ndata = udp_alloc_gensio(nadata, fd, nadata->curr_recvaddr,
+    ndata = udp_alloc_gensio(nadata, iod, nadata->curr_recvaddr,
 			     NULL, NULL, &nadata->udpns);
     if (!ndata)
 	goto out_nomem;
@@ -1266,7 +1322,10 @@ udpna_readhandler(int fd, void *cbdata)
     gensio_acc_log(nadata->acc, GENSIO_LOG_ERR,
 		   "Out of memory allocating for udp port");
  out_unlock_enable:
-    udpna_fd_read_enable(nadata);
+    if (nadata->readhandler_read_disabled) {
+	nadata->readhandler_read_disabled = false;
+	udpna_fd_read_enable(nadata);
+    }
  out_unlock:
     udpna_deref_and_unlock(nadata);
 
@@ -1288,7 +1347,7 @@ udpna_startup(struct gensio_accepter *accepter)
 
     udpna_lock(nadata);
     if (!nadata->fds) {
-	rv = gensio_os_open_socket(nadata->o, nadata->ai,
+	rv = gensio_os_open_listen_sockets(nadata->o, nadata->ai,
 				   udpna_readhandler, udpna_writehandler,
 				   udpna_fd_cleared, NULL, nadata,
 				   nadata->opensock_flags,
@@ -1401,12 +1460,13 @@ udpna_free(struct gensio_accepter *accepter)
 	if (nadata->udpn_count == 0) {
 	    unsigned int i;
 
-	    for (i = 0; i < nadata->nr_fds; i++)
-		nadata->o->clear_fd_handlers_norpt(nadata->o,
-						   nadata->fds[i].fd);
 	    for (i = 0; i < nadata->nr_fds; i++) {
-		if (nadata->fds[i].fd != -1)
-		    gensio_os_close(nadata->o, &nadata->fds[i].fd);
+		if (nadata->fds[i].iod)
+		    nadata->o->clear_fd_handlers_norpt(nadata->fds[i].iod);
+	    }
+	    for (i = 0; i < nadata->nr_fds; i++) {
+		if (nadata->fds[i].iod)
+		    nadata->o->close(&nadata->fds[i].iod);
 	    }
 	}
     }
@@ -1471,7 +1531,7 @@ udpna_str_to_gensio(struct gensio_accepter *accepter, const char *addrstr,
 	goto out_err;
     }
 
-    ndata = udp_alloc_gensio(nadata, nadata->fds[fdi].fd, addr,
+    ndata = udp_alloc_gensio(nadata, nadata->fds[fdi].iod, addr,
 			     cb, user_data, &nadata->closed_udpns);
     if (!ndata) {
 	udpna_unlock(nadata);
@@ -1585,7 +1645,7 @@ i_udp_gensio_accepter_alloc(struct gensio_addr *iai, gensiods max_read_size,
     if (!nadata->lock)
 	goto out_nomem;
 
-    nadata->curr_recvaddr = gensio_addr_alloc_recvfrom(o);
+    nadata->curr_recvaddr = o->addr_alloc_recvfrom(o);
     if (!nadata->curr_recvaddr)
 	goto out_nomem;
 
@@ -1662,11 +1722,13 @@ udp_gensio_alloc(const struct gensio_addr *addr, const char * const args[],
     struct gensio_accepter *accepter;
     struct udpna_data *nadata = NULL;
     struct gensio_addr *laddr = NULL, *mcast = NULL, *tmpaddr, *tmpaddr2;
-    int err, new_fd, ival;
-    gensiods max_read_size = GENSIO_DEFAULT_UDP_BUF_SIZE;
-    unsigned int i;
+    int err, ival;
+    struct gensio_iod *new_iod;
+    gensiods max_read_size = GENSIO_DEFAULT_UDP_BUF_SIZE, size;
+    unsigned int i, setup;
     bool nocon = false, mcast_loop_set = false, mcast_loop = true;
     bool reuseaddr = false;
+    unsigned int mttl;
 
     err = gensio_get_defaultaddr(o, "udp", "laddr", false,
 				 GENSIO_NET_PROTOCOL_UDP, true, false, &laddr);
@@ -1680,6 +1742,11 @@ udp_gensio_alloc(const struct gensio_addr *addr, const char * const args[],
     if (err)
 	return err;
     reuseaddr = ival;
+    err = gensio_get_default(o, "udp", "mttl", false,
+			     GENSIO_DEFAULT_INT, NULL, &ival);
+    if (err)
+	return err;
+    mttl = ival;
 
     err = GE_INVAL;
     for (i = 0; args && args[i]; i++) {
@@ -1712,6 +1779,13 @@ udp_gensio_alloc(const struct gensio_addr *addr, const char * const args[],
 	}
 	if (gensio_check_keybool(args[i], "nocon", &nocon) > 0)
 	    continue;
+	if (gensio_check_keyuint(args[i], "mttl", &mttl) > 0) {
+	    if (mttl < 1 || mttl > 255) {
+		err = GE_INVAL;
+		goto parm_err;
+	    }
+	    continue;
+	}
 	if (gensio_check_keybool(args[i], "mloop", &mcast_loop) > 0) {
 	    mcast_loop_set = true;
 	    continue;
@@ -1726,7 +1800,7 @@ udp_gensio_alloc(const struct gensio_addr *addr, const char * const args[],
 	return err;
     }
 
-    err = gensio_os_socket_open(o, addr, GENSIO_NET_PROTOCOL_UDP, &new_fd);
+    err = o->socket_open(o, addr, GENSIO_NET_PROTOCOL_UDP, &new_iod);
     if (err) {
 	if (laddr)
 	    gensio_addr_free(laddr);
@@ -1735,12 +1809,12 @@ udp_gensio_alloc(const struct gensio_addr *addr, const char * const args[],
 	return err;
     }
 
-    err = gensio_os_socket_setup(o, new_fd, GENSIO_NET_PROTOCOL_UDP,
-				 false, false,
-				 reuseaddr ? GENSIO_OPENSOCK_REUSEADDR : 0,
-				 laddr);
+    setup = GENSIO_SET_OPENSOCK_REUSEADDR;
+    if (reuseaddr)
+	setup |= GENSIO_OPENSOCK_REUSEADDR;
+    err = o->socket_set_setup(new_iod, setup, laddr);
     if (err) {
-	gensio_os_close(o, &new_fd);
+	o->close(&new_iod);
 	if (laddr)
 	    gensio_addr_free(laddr);
 	if (mcast)
@@ -1754,19 +1828,31 @@ udp_gensio_alloc(const struct gensio_addr *addr, const char * const args[],
     }
 
     if (mcast) {
-	err = gensio_os_mcast_add(o, new_fd, mcast, 0, false);
+	err = o->mcast_add(new_iod, mcast, 0, false);
 	gensio_addr_free(mcast);
 	if (err) {
-	    gensio_os_close(o, &new_fd);
+	    o->close(&new_iod);
 	    return err;
 	}
 	mcast = NULL;
     }
 
     if (mcast_loop_set) {
-	err = gensio_os_set_mcast_loop(o, new_fd, addr, mcast_loop);
+	size = sizeof(mcast_loop);
+	err = o->sock_control(new_iod, GENSIO_SOCKCTL_SET_MCAST_LOOP,
+			      &mcast_loop, &size);
 	if (err) {
-	    gensio_os_close(o, &new_fd);
+	    o->close(&new_iod);
+	    return err;
+	}
+    }
+
+    if (mttl > 1) {
+	size = sizeof(mttl);
+	err = o->sock_control(new_iod, GENSIO_SOCKCTL_SET_MCAST_TTL,
+			      &mttl, &size);
+	if (err) {
+	    o->close(&new_iod);
 	    return err;
 	}
     }
@@ -1775,7 +1861,7 @@ udp_gensio_alloc(const struct gensio_addr *addr, const char * const args[],
     err = i_udp_gensio_accepter_alloc(NULL, max_read_size, reuseaddr, o,
 				      NULL, NULL, &accepter);
     if (err) {
-	gensio_os_close(o, &new_fd);
+	o->close(&new_iod);
 	return err;
     }
     nadata = gensio_acc_get_gensio_data(accepter);
@@ -1784,26 +1870,26 @@ udp_gensio_alloc(const struct gensio_addr *addr, const char * const args[],
 
     nadata->fds = o->zalloc(o, sizeof(*nadata->fds));
     if (!nadata->fds) {
-	gensio_os_close(o, &new_fd);
+	o->close(&new_iod);
 	udpna_do_free(nadata);
 	return GE_NOMEM;
     }
     nadata->fds->family = gensio_addr_get_nettype(addr);
-    nadata->fds->fd = new_fd;
+    nadata->fds->iod = new_iod;
     nadata->nr_fds = 1;
     /* fd belongs to udpn now, updn_do_free() will close it. */
 
     nadata->closed = true; /* Free nadata when ndata is freed. */
     nadata->freed = true;
 
-    ndata = udp_alloc_gensio(nadata, new_fd, addr,
+    ndata = udp_alloc_gensio(nadata, new_iod, addr,
 			     cb, user_data, &nadata->closed_udpns);
     if (!ndata) {
 	err = GE_NOMEM;
     } else {
 	gensio_set_is_client(ndata->io, true);
 	nadata->udpn_count = 1;
-	err = o->set_fd_handlers(o, new_fd, nadata,
+	err = o->set_fd_handlers(new_iod, nadata,
 				 udpna_readhandler, udpna_writehandler, NULL,
 				 udpna_fd_cleared);
     }
