@@ -375,6 +375,7 @@ free_service(struct gensio_os_funcs *o, struct gensio_mdns_service *s)
 	o->free(o, s->host);
     if (s->txt)
 	avahi_string_list_free(s->txt);
+    o->free(o, s);
 }
 
 static int
@@ -396,7 +397,7 @@ gensio_mdns_remove_service(struct gensio_mdns_service *s)
 
     gensio_avahi_lock(m->ap);
     err = i_gensio_mdns_remove_service(s);
-    gensio_avahi_unlock(m->ap);
+    gensio_mdns_deref_and_unlock(m);
 
     return err;
 }
@@ -433,6 +434,7 @@ gensio_mdns_add_service(struct gensio_mdns *m,
 	return GE_NOMEM;
 
     s->m = m;
+    gensio_mdns_ref(m);
     s->interface = interface;
     s->protocol = protocol;
     s->port = port;
@@ -467,7 +469,9 @@ gensio_mdns_add_service(struct gensio_mdns *m,
     return 0;
 
  out_err:
+    gensio_avahi_lock(m->ap);
     free_service(o, s);
+    gensio_mdns_deref_and_unlock(m);
     return err;
 }
 
@@ -593,6 +597,8 @@ struct gensio_mdns_watch {
 static void
 enqueue_callback(struct gensio_mdns *m, struct gensio_mdns_callback *c)
 {
+    if (c->remove)
+	return;
     if (!c->in_queue) {
 	gensio_list_add_tail(&m->callbacks, &c->link);
 	c->in_queue = true;
@@ -909,6 +915,9 @@ mdns_service_type_callback(AvahiServiceTypeBrowser *ab,
     struct gensio_link *l;
     struct gensio_mdns_watch_browser *b = NULL;
 
+    if (w->removed)
+	return;
+
     switch (event) {
     case AVAHI_BROWSER_NEW:
     case AVAHI_BROWSER_REMOVE:
@@ -1058,6 +1067,7 @@ gensio_mdns_add_watch(struct gensio_mdns *m,
 	return GE_NOMEM;
 
     w->m = m;
+    gensio_mdns_ref(m);
     w->cb = callback;
     w->callback_data.w = w;
     w->userdata = userdata;
@@ -1097,8 +1107,10 @@ gensio_mdns_add_watch(struct gensio_mdns *m,
     return 0;
 
  out_err:
+    gensio_avahi_lock(m->ap);
     watch_free(o, w);
-    return GE_NOMEM;
+    gensio_mdns_deref_and_unlock(m);
+    return err;
 }
 
 static int
@@ -1140,8 +1152,8 @@ i_gensio_mdns_remove_watch(struct gensio_mdns_watch *w,
 	gensio_list_rm(&w->browsers, &b->link);
 	browser_free(o, b);
     }
-    w->callback_data.remove = true;
     enqueue_callback(m, &w->callback_data);
+    w->callback_data.remove = true;
 
     return 0;
 }
@@ -1215,6 +1227,7 @@ static void mdns_runner(struct gensio_runner *runner, void *userdata)
 		gensio_avahi_lock(m->ap);
 	    }
 	    watch_free(o, w);
+	    gensio_mdns_deref(m);
 	} else {
 	    if (c->data) {
 		struct gensio_mdns_watch_data *d = c->data;
@@ -1305,6 +1318,9 @@ gensio_free_mdns(struct gensio_mdns *m, gensio_mdns_done done, void *userdata)
 	err = GE_INUSE;
 	goto out_unlock;
     }
+
+    gensio_avahi_poll_disable(m->ap);
+
     m->freed = true;
     m->free_done = done;
     m->free_userdata = userdata;
