@@ -5,8 +5,9 @@
  *  SPDX-License-Identifier: LGPL-2.1-only
  */
 
-typedef PyObject swig_cb;
-typedef PyObject swig_cb_val;
+#include <gensio/gensio_swig.h>
+#include "python_swig_internals.h"
+
 typedef struct swig_ref {
     PyObject *val;
 } swig_ref;
@@ -17,11 +18,9 @@ typedef struct swig_ref {
 #ifdef WITH_THREAD
 static void gensio_swig_init_lang(void)
 {
+    swig_waiter_wake = wake_curr_waiter;
     PyEval_InitThreads();
 }
-#define OI_PY_STATE PyGILState_STATE
-#define OI_PY_STATE_GET() PyGILState_Ensure()
-#define OI_PY_STATE_PUT(s) PyGILState_Release(s)
 
 /* We do need to work about blocking, though. */
 #define GENSIO_SWIG_C_BLOCK_ENTRY Py_BEGIN_ALLOW_THREADS
@@ -29,6 +28,7 @@ static void gensio_swig_init_lang(void)
 #else
 static void gensio_swig_init_lang(void)
 {
+    swig_waiter_wake = wake_curr_waiter;
 }
 #define OI_PY_STATE int
 #define OI_PY_STATE_GET() 0
@@ -62,16 +62,12 @@ OI_PI_AsBytesAndSize(PyObject *o, char **buf, my_ssize_t *len)
 }
 
 #define OI_PI_StringCheck PyUnicode_Check
-#define OI_PI_FromString PyUnicode_FromString
 #define OI_PI_FromStringAndSize PyUnicode_FromStringAndSize
-#define OI_PI_AsString PyUnicode_AsUTF8
 #else
 #define OI_PI_BytesCheck PyString_Check
 #define OI_PI_AsBytesAndSize PyString_AsStringAndSize
 #define OI_PI_StringCheck PyString_Check
-#define OI_PI_FromString PyString_FromString
 #define OI_PI_FromStringAndSize PyString_FromStringAndSize
-#define OI_PI_AsString PyString_AsString
 #endif
 
 static PyObject *
@@ -88,17 +84,7 @@ OI_PI_FromStringN(const char *s)
     return o;
 }
 
-static swig_cb_val *
-ref_swig_cb_i(swig_cb *cb)
-{
-    OI_PY_STATE gstate;
-
-    gstate = OI_PY_STATE_GET();
-    Py_INCREF(cb);
-    OI_PY_STATE_PUT(gstate);
-    return cb;
-}
-#define ref_swig_cb(cb, func) ref_swig_cb_i(cb)
+#define ref_swig_cb(cb, func) gensio_python_ref_swig_cb_i(cb)
 
 static swig_ref
 swig_make_ref_i(void *item, swig_type_info *class)
@@ -114,18 +100,7 @@ swig_make_ref_i(void *item, swig_type_info *class)
 #define swig_make_ref(item, name) \
 	swig_make_ref_i(item, SWIGTYPE_p_ ## name)
 
-static swig_cb_val *
-deref_swig_cb_val(swig_cb_val *cb)
-{
-    OI_PY_STATE gstate;
-
-    if (cb) {
-	gstate = OI_PY_STATE_GET();
-	Py_DECREF(cb);
-	OI_PY_STATE_PUT(gstate);
-    }
-    return cb;
-}
+#define deref_swig_cb_val(v) gensio_python_deref_swig_cb_val(v)
 
 /* No way to check the refcount in Python. */
 #define swig_free_ref_check(r, c) \
@@ -133,34 +108,6 @@ deref_swig_cb_val(swig_cb_val *cb)
 	    swig_free_ref(r);						\
 	} while(0)
 
-static PyObject *
-swig_finish_call_rv(swig_cb_val *cb, const char *method_name, PyObject *args,
-		    bool optional)
-{
-    PyObject *p, *o = NULL;
-
-    if (PyObject_HasAttrString(cb, method_name)) {
-	p = PyObject_GetAttrString(cb, method_name);
-	o = PyObject_CallObject(p, args);
-	Py_DECREF(p);
-	if (PyErr_Occurred())
-	    wake_curr_waiter();
-    } else if (!optional) {
-	PyObject *t = PyObject_GetAttrString(cb, "__class__");
-	PyObject *c = PyObject_GetAttrString(t, "__name__");
-	const char *class = OI_PI_AsString(c);
-
-	PyErr_Format(PyExc_RuntimeError,
-		     "gensio callback: Class '%s' has no method '%s'\n",
-		     class, method_name);
-	wake_curr_waiter();
-    }
-    if (args)
-	Py_DECREF(args);
-
-    return o;
-}
-
 static gensiods
 swig_finish_call_rv_gensiods(swig_cb_val *cb, const char *method_name,
 			     PyObject *args, bool optional)
@@ -224,14 +171,6 @@ swig_finish_call(swig_cb_val *cb, const char *method_name, PyObject *args,
 	Py_DECREF(o);
 }
 
-struct os_funcs_data {
-#ifdef USE_POSIX_THREADS
-    pthread_mutex_t lock;
-#endif
-    unsigned int refcount;
-    swig_cb_val *log_handler;
-};
-
 #ifdef USE_POSIX_THREADS
 static void os_funcs_lock(struct os_funcs_data *odata)
 {
@@ -260,23 +199,6 @@ os_funcs_ref(struct gensio_os_funcs *o)
     os_funcs_unlock(odata);
 }
 
-static void
-check_os_funcs_free(struct gensio_os_funcs *o)
-{
-    struct os_funcs_data *odata = o->other_data;
-
-    os_funcs_lock(odata);
-    if (--odata->refcount == 0) {
-	os_funcs_unlock(odata);
-	if (odata->log_handler)
-	    deref_swig_cb_val(odata->log_handler);
-	free(odata);
-	o->free_funcs(o);
-    } else {
-	os_funcs_unlock(odata);
-    }
-}
-
 struct gensio_data {
     bool tmpval; /* If true, just ignore this on destroy. */
     int refcount;
@@ -379,42 +301,6 @@ gensio_accepter_ref(struct gensio_accepter *acc)
     ref_gensio_data(data);
 }
 
-static void gensio_do_vlog(struct gensio_os_funcs *o,
-			   enum gensio_log_levels level,
-			   const char *fmt, va_list fmtargs)
-{
-    struct os_funcs_data *odata = o->other_data;
-    char *buf = NULL;
-    unsigned int len;
-    PyObject *args, *po;
-    va_list tmpva;
-    OI_PY_STATE gstate;
-
-    if (!odata->log_handler)
-	return;
-
-    gstate = OI_PY_STATE_GET();
-
-    va_copy(tmpva, fmtargs);
-    len = vsnprintf(buf, 0, fmt, tmpva);
-    va_end(tmpva);
-    buf = o->zalloc(o, len + 1);
-    if (!buf)
-	goto out;
-    vsnprintf(buf, len + 1, fmt, fmtargs);
-
-    args = PyTuple_New(2);
-    po = OI_PI_FromString(gensio_log_level_to_str(level));
-    PyTuple_SET_ITEM(args, 0, po);
-    po = OI_PI_FromString(buf);
-    PyTuple_SET_ITEM(args, 1, po);
-    o->free(o, buf);
-
-    swig_finish_call(odata->log_handler, "gensio_log", args, false);
- out:
-    OI_PY_STATE_PUT(gstate);
-}
-
 static void
 gensio_open_done(struct gensio *io, int err, void *cb_data) {
     swig_cb_val *cb = cb_data;
@@ -663,7 +549,7 @@ gensio_child_event(struct gensio *io, void *user_data, int event, int readerr,
 		   const char *const *auxdata)
 {
     struct gensio_data *data = user_data;
-    swig_ref io_ref = { .val = NULL }, new_con = { .val = NULL };
+    swig_ref io_ref, new_con;
     PyObject *args, *o;
     OI_PY_STATE gstate;
     int rv = 0;
@@ -827,6 +713,54 @@ gensio_child_event(struct gensio *io, void *user_data, int event, int readerr,
 	}
 	break;
 
+    case GENSIO_EVENT_2FA_VERIFY:
+	io_ref = swig_make_ref(io, gensio);
+	args = PyTuple_New(2);
+	ref_gensio_data(data);
+	PyTuple_SET_ITEM(args, 0, io_ref.val);
+	/*
+	 * FIXME - is there a way to make this a secure python string
+	 * that gets wiped on free?
+	 */
+	o = PyBytes_FromStringAndSize((const char *) buf, *buflen);
+	PyTuple_SET_ITEM(args, 1, o);
+
+	rv = swig_finish_call_rv_int(data->handler_val, "verify_2fa",
+				     args, true);
+	break;
+
+    case GENSIO_EVENT_REQUEST_2FA:
+	io_ref = swig_make_ref(io, gensio);
+	args = PyTuple_New(1);
+	ref_gensio_data(data);
+	PyTuple_SET_ITEM(args, 0, io_ref.val);
+	o = swig_finish_call_rv(data->handler_val, "request_2fa",
+				args, true);
+	rv = GE_NOTSUP;
+	if (o) {
+	    if (OI_PI_BytesCheck(o)) {
+		char *p;
+		unsigned char *p2;
+		my_ssize_t len;
+
+		rv = OI_PI_AsBytesAndSize(o, &p, &len);
+		if (!rv) {
+		    p2 = data->o->zalloc(data->o, len + 1);
+		    if (!p2) {
+			rv = GE_NOMEM;
+		    } else {
+			memcpy(p2, p, len);
+			*((unsigned char **) buf) = p2;
+			*buflen = len;
+		    }
+		}
+	    } else if (PyInt_Check(o)) {
+		rv = PyInt_AsLong(o);
+	    }
+	    Py_DecRef(o);
+	}
+	break;
+
     case GENSIO_EVENT_SER_MODEMSTATE:
 	sgensio_modemstate(io, *((unsigned int *) buf));
 	break;
@@ -1111,6 +1045,63 @@ gensio_acc_child_event(struct gensio_accepter *accepter, void *user_data,
 	    }
 	    Py_DecRef(o);
 	}
+	OI_PY_STATE_PUT(gstate);
+	return rv;
+
+    case GENSIO_ACC_EVENT_2FA_VERIFY:
+	pwvfy = (struct gensio_acc_password_verify_data *) cdata;
+	return gensio_acc_io_call_cb(accepter, pwvfy->io, "verify_2fa",
+				     -1, pwvfy->password, true);
+
+    case GENSIO_ACC_EVENT_REQUEST_2FA:
+	pwvfy = (struct gensio_acc_password_verify_data *) cdata;
+	io = pwvfy->io;
+
+	gstate = OI_PY_STATE_GET();
+
+	/*
+	 * This is a situation where the gensio has not been reported
+	 * to the upper layer yet and thus there is no user data.
+	 * Just create something to say that this isn't valid.
+	 */
+	old_user_data = gensio_get_user_data(io);
+	tmpdata.tmpval = true;
+	gensio_set_user_data(io, &tmpdata);
+
+	acc_ref = swig_make_ref(accepter, gensio_accepter);
+	gensio_accepter_ref(accepter);
+	io_ref = swig_make_ref(io, gensio);
+	args = PyTuple_New(2);
+	PyTuple_SET_ITEM(args, 0, acc_ref.val);
+	PyTuple_SET_ITEM(args, 1, io_ref.val);
+
+	o = swig_finish_call_rv(data->handler_val, "request_2fa",
+				args, true);
+	gensio_set_user_data(io, old_user_data);
+	rv = GE_NOTSUP;
+	if (o) {
+	    if (OI_PI_BytesCheck(o)) {
+		char *p;
+		unsigned char *p2;
+		my_ssize_t len;
+
+		rv = OI_PI_AsBytesAndSize(o, &p, &len);
+		if (!rv) {
+		    p2 = data->o->zalloc(data->o, len + 1);
+		    if (!p2) {
+			rv = GE_NOMEM;
+		    } else {
+			memcpy(p2, p, len);
+			*((unsigned char **) pwvfy->password) = p2;
+			pwvfy->password_len = len;
+		    }
+		}
+	    } else if (PyInt_Check(o)) {
+		rv = PyInt_AsLong(o);
+	    }
+	    Py_DecRef(o);
+	}
+	OI_PY_STATE_PUT(gstate);
 	return rv;
     }
 
@@ -1270,7 +1261,9 @@ static void gensio_mdns_remove_watch_done(struct gensio_mdns_watch *watch,
     if (w->free_on_close) {
 	o->unlock(w->lock);
 	o->free_lock(w->lock);
+	deref_swig_cb_val(w->cb_val);
 	o->free(o, w);
+	check_os_funcs_free(o);
     } else {
 	w->watch = NULL;
 	o->unlock(w->lock);
