@@ -25,7 +25,15 @@
 #include <errno.h>
 #include <signal.h>
 #include <gensio/gensio.h>
-#include <gensio/gensio_selector.h>
+#include <gensio/gensio_osops_env.h>
+#include <gensio/argvutils.h>
+#include <gensio/gensio_unix.h>
+#ifdef HAVE_GLIB
+#include <gensio/gensio_glib.h>
+#endif
+#ifdef HAVE_TCL
+#include <gensio/gensio_tcl.h>
+#endif
 #include "pthread_handler.h"
 
 struct oom_tests {
@@ -62,11 +70,68 @@ struct oom_tests {
     struct gensio_os_funcs *o;
 };
 
-static sigset_t waitsigs;
-bool verbose;
-bool debug;
+static bool verbose;
+static bool debug;
+
+static unsigned int num_extra_threads = 3;
+static bool use_glib = false;
+static bool use_tcl = false;
+static const char *os_func_str = "";
+
+struct gensio_os_proc_data *proc_data;
+
+#ifdef _WIN32
+#include <windows.h>
+
+#ifdef _MSC_VER
+typedef int ssize_t;
+#endif
+
+#define DEFAULT_ECHO_COMMPORT "COM1"
+
+bool
+file_is_accessible_dev(const char *filename)
+{
+    return true; /* FIXME - what to do here? */
+}
+
+#define sleep(n) Sleep(n * 1000)
+
+#define WIFEXITED(n) ((n) < 128)
+#define WEXITSTATUS(n) (n)
+#define WIFSIGNALED(n) ((n) >= 128)
+#define WTERMSIG(n) (n)
+const char *
+strsignal(int n)
+{
+    if (n == STATUS_ACCESS_VIOLATION)
+	return "access violation";
+    if (n == STATUS_ASSERTION_FAILURE)
+	return "assertion failure";
+    if (n == STATUS_CONTROL_C_EXIT)
+	return "control-C";
+    return "unknown";
+}
+
+
+static FILE *
+open_tempfile(char *name, unsigned int len, const char *pattern)
+{
+    DWORD pos;
+
+    pos = GetTempPathA(len, name);
+    if (pos == 0)
+	return NULL;
+
+    if (pos + strlen(pattern) + 7 >= len)
+	return NULL;
+
+    sprintf(name + pos, "%s%6.6ld", pattern, GetTickCount() % 1000000);
+    return fopen(name, "w");
+}
+
+#else /* _WIN32 */
 
-#if HAVE_SERIALDEV
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
@@ -92,6 +157,25 @@ file_is_accessible_dev(const char *filename)
 	return false;
     }
 }
+
+static FILE *
+open_tempfile(char *name, unsigned int len, const char *pattern)
+{
+    int fd;
+    FILE *f;
+
+    snprintf(name, len, "/tmp/%sXXXXXX", pattern);
+    fd = mkstemp(name);
+    if (fd == -1)
+	return NULL;
+    f = fdopen(fd, "w");
+    if (!f)
+	close(fd);
+    return f;
+}
+
+#define DEFAULT_ECHO_COMMPORT "/dev/ttyEcho0"
+
 #endif
 
 bool sleep_on_timeout_err;
@@ -140,22 +224,36 @@ static bool
 get_echo_dev(struct gensio_os_funcs *o, const char *testname,
 	     const char *str, char **newstr)
 {
-    const char *e = getenv("GENSIO_TEST_ECHO_DEV");
+    char *e, *te = NULL;
+    int rv;
 
-    if (e) {
-	if (strlen(e) == 0) {
+    rv = gensio_os_env_getalloc(o, "GENSIO_TEST_ECHO_DEV", &te);
+
+    if (rv == 0) {
+	if (strlen(te) == 0) {
 	    printf("Serial echo device disabled, skipping serialdev test\n");
+	    if (te)
+		o->free(o, te);
 	    return false;
 	}
+	e = te;
+    } else if (rv == GE_NOTFOUND) {
+	e = DEFAULT_ECHO_COMMPORT;
     } else {
-	e = "/dev/ttyEcho0";
+	fprintf(stderr, "Unable to get GENSIO_TEST_ECHO_DEV: %s",
+		gensio_err_to_str(rv));
+	return false;
     }
     if (!file_is_accessible_dev(e)) {
 	printf("Serial echo device '%s' doesn't exist or is not accessible,\n"
 	       "skipping %s test\n", e, testname);
+	if (te)
+	    o->free(o, te);
 	return false;
     }
     *newstr = gensio_alloc_sprintf(o, str, e);
+    if (te)
+	o->free(o, te);
     if (!*newstr) {
 	printf("Unable to allocate memory for echo device '%s',\n"
 	       "skipping %s test\n", e, testname);
@@ -216,6 +314,7 @@ ipmisim_cb(struct gensio *io, void *user_data, int event, int err,
 
 	if (verbose)
 	    printf("IPMISIM err: %s\n", gensio_err_to_str(err));
+	test->o->wake(test->waiter);
 	return 0;
     }
 
@@ -257,9 +356,13 @@ ipmisim_err_cb(struct gensio *io, void *user_data, int event, int err,
 {
     /* FIXME - Add debug I/O. */
     if (err) {
+	struct oom_tests *test = user_data;
+
+	test->err = err;
 	gensio_set_read_callback_enable(io, false);
 	if (verbose)
 	    printf("IPMISIM err err: %s\n", gensio_err_to_str(err));
+	test->o->wake(test->waiter);
 	return 0;
     }
 
@@ -325,9 +428,11 @@ ipmisim_start(struct gensio_os_funcs *o, struct oom_tests *test)
     gensio_set_read_callback_enable(test->io2, true);
 
  retry:
-    err = o->wait_intr_sigmask(test->waiter, 1, &timeout, &waitsigs);
+    err = o->wait_intr_sigmask(test->waiter, 1, &timeout, proc_data);
     if (err == GE_INTERRUPTED)
 	goto retry;
+    if (test->err)
+	err = test->err;
     if (err) {
 	printf("Error waiting for ipmi_sim started for %s: %s,\n"
 	       "skipping ipmisol test\n", test->args, gensio_err_to_str(err));
@@ -338,10 +443,12 @@ ipmisim_start(struct gensio_os_funcs *o, struct oom_tests *test)
 
  out_err:
     if (test->io) {
+	gensio_close_s(test->io);
 	gensio_free(test->io);
 	test->io = NULL;
     }
     if (test->io2) {
+	gensio_close_s(test->io2);
 	gensio_free(test->io2);
 	test->io2 = NULL;
     }
@@ -399,30 +506,40 @@ static bool
 check_ipmisim_present(struct gensio_os_funcs *o, struct oom_tests *test)
 {
     char *config = NULL;
-    const char *prog = getenv("IPMISIM_EXEC");
-    int fd;
+    char *prog, *tprog = NULL;
+    FILE *f;
     ssize_t rv, len;
 
-    if (!prog)
+    rv = gensio_os_env_getalloc(o, "IPMISIM_EXEC", &tprog);
+    if (rv == GE_NOTFOUND) {
 	prog = "ipmi_sim";
-
-    if (!get_echo_dev(o, "ipmisol", ipmisim_config, &config))
+    } else if (rv != 0) {
+	fprintf(stderr, "Unable to get IPMISIM_EXEC: %s\n",
+		gensio_err_to_str(rv));
 	return false;
+    } else {
+	prog = tprog;
+    }
 
-    strncpy(test->configname, "/tmp/gensio_oomtest_conf_XXXXXX",
-	    sizeof(test->configname));
-    fd = mkstemp(test->configname);
-    if (fd == -1) {
+    if (!get_echo_dev(o, "ipmisol", ipmisim_config, &config)) {
+	if (tprog)
+	    o->free(o, tprog);
+	return false;
+    }
+
+    f = open_tempfile(test->configname, sizeof(test->configname),
+		      "gensio_oomtest_conf_");
+    if (!f) {
 	printf("Unable to open ipmisim config file '%s',\n"
 	       "skipping ipmisol test\n", test->configname);
 	test->configname[0] = '\0';
 	goto out_err;
     }
     len = strlen(config);
-    rv = write(fd, config, len);
+    rv = fwrite(config, 1, len, f);
     o->free(o, config);
     config = NULL;
-    close(fd);
+    fclose(f);
     if (rv != len) {
 	unlink(test->configname);
 	printf("Unable to write ipmisim config file '%s',\n"
@@ -432,15 +549,16 @@ check_ipmisim_present(struct gensio_os_funcs *o, struct oom_tests *test)
 
     strncpy(test->emuname, "/tmp/gensio_oomtest_emu_XXXXXX",
 	    sizeof(test->emuname));
-    fd = mkstemp(test->emuname);
-    if (fd == -1) {
+    f = open_tempfile(test->emuname, sizeof(test->emuname),
+		      "gensio_oomtest_emu_");
+    if (!f) {
 	printf("Unable to open ipmisim emu file '%s',\n"
 	       "skipping ipmisol test\n", test->emuname);
 	test->emuname[0] = '\0';
 	goto out_err;
     }
-    rv = write(fd, ipmisim_emu, strlen(ipmisim_emu));
-    close(fd);
+    rv = fwrite(ipmisim_emu, 1, strlen(ipmisim_emu), f);
+    fclose(f);
     if (rv != strlen(ipmisim_emu)) {
 	printf("Unable to write ipmisim emu file '%s',\n"
 	       "skipping ipmisol test\n", test->emuname);
@@ -467,11 +585,17 @@ check_ipmisim_present(struct gensio_os_funcs *o, struct oom_tests *test)
     if (ipmisim_start(o, test))
 	goto out_err;
 
+    if (tprog)
+	o->free(o, tprog);
+
     return true;
 
  out_err:
+    if (tprog)
+	o->free(o, tprog);
     if (config)
 	o->free(o, config);
+    ipmisim_end(o, test);
     ipmisim_finish(o, test);
 
     return false;
@@ -480,14 +604,10 @@ check_ipmisim_present(struct gensio_os_funcs *o, struct oom_tests *test)
 static bool
 check_serialdev_present(struct gensio_os_funcs *o, struct oom_tests *test)
 {
-#if HAVE_SERIALDEV
     if (!get_echo_dev(o, "serialdev", test->connecter, &test->connecter))
 	return false;
     test->free_connecter = true;
     return true;
-#else
-    return false;
-#endif
 }
 
 static bool
@@ -541,7 +661,7 @@ struct oom_tests oom_tests[] = {
     { "tcp,localhost,", "tcp,0" },
     { "sctp,localhost,", "sctp,0",
       .check_if_present = check_sctp_present, .check_value = HAVE_LIBSCTP },
-    { "udp,localhost,", "udp,0",
+    { "udp,ipv4,localhost,", "udp,ipv4,0",
       /* In this tests some errors will not result in a failure. */
       .allow_no_err_on_trig = true
     },
@@ -557,7 +677,6 @@ struct oom_tests oom_tests[] = {
     { "telnet(rfc2217),tcp,localhost,", "telnet(rfc2217),tcp,0" },
     { "serialdev,%s,115200", NULL,
       .check_if_present = check_serialdev_present,
-      .check_value = HAVE_SERIALDEV,
       /*
        * The error injections cause this to take way to long with
        * large I/O sizes.  So limit it to a reasonable value.
@@ -585,35 +704,20 @@ struct oom_tests oom_tests[] = {
        *    It can still be run directly with the -t option.
        */
       .no_default_run = true,
-      .check_value = HAVE_PTY && HAVE_SERIALDEV },
+      .check_value = HAVE_PTY },
     { NULL }
 };
 
 static struct gensio_os_funcs *o;
 static char *gensiot;
-static bool got_sigchild;
 
 static void
-handle_sigchld(int sig)
-{
-    got_sigchild = true;
-}
-
-static void
-handle_sigusr1(int sig)
-{
-}
-
-#ifdef USE_PTHREADS
-static void *
 gensio_loop(void *info)
 {
     struct gensio_waiter *closewaiter = info;
 
     o->wait(closewaiter, 1, NULL);
-    return NULL;
 }
-#endif
 
 static void
 do_vlog(struct gensio_os_funcs *f, enum gensio_log_levels level,
@@ -686,6 +790,7 @@ struct oom_test_data {
     gensiods ccon_stderr_pos;
     struct gensio *ccon_stderr_io;
     bool stderr_expect_close;
+    bool stderr_rem_closed;
 
     bool stderr_open_done;
     bool stderr_closed;
@@ -799,7 +904,7 @@ ccon_stderr_closed(struct gensio *io, void *close_data)
 
     assert(!od->finished);
     od->stderr_closed = true;
-    rv = gensio_control(io, GENSIO_CONTROL_DEPTH_FIRST, true,
+    rv = gensio_control(io, GENSIO_CONTROL_DEPTH_FIRST, GENSIO_CONTROL_GET,
 			GENSIO_CONTROL_EXIT_CODE, intstr, &size);
     assert(!debug || !rv);
     OOMLOCK(&od->lock);
@@ -940,20 +1045,20 @@ con_cb(struct gensio *io, void *user_data,
 
 	    rv = gensio_write(io, &count, iodata + id->write_pos, wrsize, NULL);
 	    if (rv) {
-		gensio_set_write_callback_enable(io, false);
-		gensio_set_read_callback_enable(io, false);
 		if (rv == GE_SHUTDOWN || rv == GE_NOTREADY) {
 		    if (debug) {
 			printf("Write on shutdown or not ready socket\n");
 			fflush(stdout);
 		    }
 		} else {
-		    assert(!debug || !rv || rv == GE_REMCLOSE);
 		    if (debug) {
 			printf("con_cb error 2: %s\n", gensio_err_to_str(rv));
 			fflush(stdout);
 		    }
+		    assert(debug || !rv || rv == GE_REMCLOSE);
 		}
+		gensio_set_write_callback_enable(io, false);
+		gensio_set_read_callback_enable(io, false);
 		id->err = rv;
 		o->wake(od->waiter);
 	    } else {
@@ -982,8 +1087,8 @@ set_max_write(struct io_test_data *id, struct gensio *io)
     char databuf[20];
     gensiods dbsize = sizeof(databuf);
 
-    rv = gensio_control(io, 0, true, GENSIO_CONTROL_MAX_WRITE_PACKET,
-			databuf, &dbsize);
+    rv = gensio_control(io, 0, GENSIO_CONTROL_GET,
+			GENSIO_CONTROL_MAX_WRITE_PACKET, databuf, &dbsize);
     if (!rv)
 	id->max_write = strtoul(databuf, NULL, 0);
 }
@@ -1044,6 +1149,7 @@ ccon_stderr_cb(struct gensio *io, void *user_data,
 	OOMLOCK(&od->lock);
 	assert(!debug || err == GE_REMCLOSE);
 	gensio_set_read_callback_enable(io, false);
+	od->stderr_rem_closed = true;
 	if (!od->stderr_expect_close || err != GE_REMCLOSE)
 	    od->ccon.err = err;
 	o->wake(od->waiter);
@@ -1067,7 +1173,17 @@ ccon_stderr_cb(struct gensio *io, void *user_data,
 	    char *nl, *c, *s;
 
 	    od->ccon_stderr[od->ccon_stderr_pos] = '\0';
+#ifdef _WIN32 /* Stupid Windows newlines. */
+	    nl = strchr(od->ccon_stderr, '\r');
+	    if (nl) {
+		if (*(nl + 1) == '\n')
+		    *nl++ = '\0';
+		else
+		    nl = NULL; /* Haven't received the whole newline yet. */
+	    }
+#else
 	    nl = strchr(od->ccon_stderr, '\n');
+#endif
 	    if (nl) {
 		*nl = '\0';
 		if (strcmp(od->ccon_stderr, "Done") == 0) {
@@ -1273,7 +1389,7 @@ wait_for_data(struct oom_test_data *od, gensio_time *timeout)
 
     for (;;) {
 	OOMUNLOCK(&od->lock);
-	rv = o->wait_intr_sigmask(od->waiter, 1, timeout, &waitsigs);
+	rv = o->wait_intr_sigmask(od->waiter, 1, timeout, proc_data);
 	OOMLOCK(&od->lock);
 	if (rv == GE_INTERRUPTED)
 	    continue;
@@ -1319,7 +1435,7 @@ close_con(struct io_test_data *id, gensio_time *timeout)
     /* Make sure the open completes. */
     while (!id->open_done) {
 	OOMUNLOCK(&od->lock);
-	rv = o->wait_intr_sigmask(id->od->waiter, 1, timeout, &waitsigs);
+	rv = o->wait_intr_sigmask(id->od->waiter, 1, timeout, proc_data);
 	OOMLOCK(&od->lock);
 	if (rv == GE_TIMEDOUT) {
 	    printf("Waiting on timeout err A\n");
@@ -1356,6 +1472,23 @@ close_stderr(struct oom_test_data *od, gensio_time *timeout)
     if (!od->ccon_stderr_io)
 	return 0;
 
+    while (!od->stderr_rem_closed) {
+	OOMUNLOCK(&od->lock);
+	rv = o->wait_intr_sigmask(od->waiter, 1, timeout, proc_data);
+	OOMLOCK(&od->lock);
+	if (rv == GE_TIMEDOUT) {
+	    printf("Waiting on timeout err G1\n");
+	    fflush(stdout);
+	    handle_timeout_err();
+	}
+	if (rv == GE_INTERRUPTED)
+	    continue;
+	if (rv) {
+	    if (!err)
+		err = rv;
+	    break;
+	}
+    }
     rv = gensio_close(od->ccon_stderr_io, ccon_stderr_closed, od);
     assert(!debug || !rv || rv == GE_REMCLOSE);
     if (rv && !err) {
@@ -1366,7 +1499,7 @@ close_stderr(struct oom_test_data *od, gensio_time *timeout)
     od_ref(od); /* Ref for the close */
     while (od->ccon_stderr_io) {
 	OOMUNLOCK(&od->lock);
-	rv = o->wait_intr_sigmask(od->waiter, 1, timeout, &waitsigs);
+	rv = o->wait_intr_sigmask(od->waiter, 1, timeout, proc_data);
 	OOMLOCK(&od->lock);
 	if (rv == GE_TIMEDOUT) {
 	    printf("Waiting on timeout err G\n");
@@ -1401,7 +1534,7 @@ close_cons(struct oom_test_data *od, bool close_acc, gensio_time *timeout)
 
     while (!err && (od->ccon.io || od->scon.io)) {
 	OOMUNLOCK(&od->lock);
-	rv = o->wait_intr_sigmask(od->waiter, 1, timeout, &waitsigs);
+	rv = o->wait_intr_sigmask(od->waiter, 1, timeout, proc_data);
 	OOMLOCK(&od->lock);
 	if (rv == GE_TIMEDOUT) {
 	    printf("Waiting on timeout err B\n");
@@ -1424,8 +1557,15 @@ close_cons(struct oom_test_data *od, bool close_acc, gensio_time *timeout)
     return err;
 }
 
+struct env_info {
+    const char **argv;
+    gensiods args;
+    gensiods argc;
+};
+
 static int
-run_oom_test(struct oom_tests *test, long count, int *exitcode, bool close_acc)
+run_oom_test(struct oom_tests *test, long count, int *exitcode,
+	     bool close_acc, struct env_info *env)
 {
     struct oom_test_data *od;
     int rv, err = 0;
@@ -1439,16 +1579,18 @@ run_oom_test(struct oom_tests *test, long count, int *exitcode, bool close_acc)
 
     OOMLOCK(&od->lock);
     if (count < 0) {
-	rv = unsetenv("GENSIO_ERRTRIG_TEST");
+	rv = gensio_os_argvenv_set(o, &env->argv, &env->args, &env->argc,
+				   "GENSIO_ERRTRIG_TEST", NULL);
     } else {
 	snprintf(intstr, sizeof(intstr), "%ld ", count);
-	rv = setenv("GENSIO_ERRTRIG_TEST", intstr, 1);
+	rv = gensio_os_argvenv_set(o, &env->argv, &env->args, &env->argc,
+				   "GENSIO_ERRTRIG_TEST", intstr);
     }
     if (rv) {
-	rv = errno;
-	fprintf(stderr, "Unable to set environment properly\n");
+	fprintf(stderr, "Unable to set environment properly: %s\n",
+		gensio_err_to_str(rv));
 	od_deref_and_unlock(od);
-	return gensio_os_err_to_err(o, rv);
+	return rv;
     }
 
     if (test->accepter) {
@@ -1470,12 +1612,16 @@ run_oom_test(struct oom_tests *test, long count, int *exitcode, bool close_acc)
 	if (rv)
 	    goto out_err;
 
-	constr = gensio_alloc_sprintf(o, "stdio, %s%s -i 'stdio(self)' '%s%s'",
-				      gensiot, test->conacc ? " -a" : "",
-				      test->connecter, intstr);
+	constr = gensio_alloc_sprintf(o,
+			"stdio,%s%s%s -n %u -i 'stdio(self)' '%s%s'",
+			gensiot, test->conacc ? " -a" : "",
+			os_func_str, num_extra_threads, test->connecter,
+			intstr);
     } else {
-	constr = gensio_alloc_sprintf(o, "stdio, %s -i 'stdio(self)' '%s'",
-				      gensiot, test->connecter);
+	constr = gensio_alloc_sprintf(o,
+			"stdio,%s%s -n %u -i 'stdio(self)' '%s'",
+			gensiot, os_func_str, num_extra_threads,
+			test->connecter);
     }
     if (!constr) {
 	rv = GE_NOMEM;
@@ -1488,6 +1634,11 @@ run_oom_test(struct oom_tests *test, long count, int *exitcode, bool close_acc)
     if (rv)
 	goto out_err;
 
+    rv = gensio_control(od->ccon.io, 0, GENSIO_CONTROL_SET,
+			GENSIO_CONTROL_ENVIRONMENT, (char *) env->argv, NULL);
+    if (rv)
+	goto out_err;
+
     rv = gensio_open(od->ccon.io, ccon_open_done, od);
     assert(!debug || !rv);
     if (rv) {
@@ -1511,7 +1662,7 @@ run_oom_test(struct oom_tests *test, long count, int *exitcode, bool close_acc)
 	    od_ref(od); /* Ref for the close */
 	    while (od->acc) {
 		OOMUNLOCK(&od->lock);
-		rv = o->wait_intr_sigmask(od->waiter, 1, &timeout, &waitsigs);
+		rv = o->wait_intr_sigmask(od->waiter, 1, &timeout, proc_data);
 		OOMLOCK(&od->lock);
 		if (rv == GE_TIMEDOUT) {
 		    printf("Waiting on timeout err C\n");
@@ -1559,7 +1710,7 @@ run_oom_test(struct oom_tests *test, long count, int *exitcode, bool close_acc)
 
 static int
 run_oom_acc_test(struct oom_tests *test, long count, int *exitcode,
-		 bool close_acc)
+		 bool close_acc, struct env_info *env)
 {
     struct oom_test_data *od;
     int rv, err = 0;
@@ -1572,19 +1723,23 @@ run_oom_acc_test(struct oom_tests *test, long count, int *exitcode,
 
     OOMLOCK(&od->lock);
     if (count < 0) {
-	rv = unsetenv("GENSIO_ERRTRIG_TEST");
+	rv = gensio_os_argvenv_set(o, &env->argv, &env->args, &env->argc,
+				   "GENSIO_ERRTRIG_TEST", NULL);
     } else {
 	snprintf(intstr, sizeof(intstr), "%ld ", count);
-	rv = setenv("GENSIO_ERRTRIG_TEST", intstr, 1);
+	rv = gensio_os_argvenv_set(o, &env->argv, &env->args, &env->argc,
+				   "GENSIO_ERRTRIG_TEST", intstr);
     }
     if (rv) {
-	fprintf(stderr, "Unable to set environment properly\n");
+	fprintf(stderr, "Unable to set environment properly: %s\n",
+		gensio_err_to_str(rv));
 	od_deref_and_unlock(od);
-	return gensio_os_err_to_err(o, errno);
+	return rv;
     }
 
-    constr = gensio_alloc_sprintf(o, "stdio, %s -v -a -p -i 'stdio(self)' '%s'",
-				  gensiot, test->accepter);
+    constr = gensio_alloc_sprintf(o,
+		"stdio,%s%s -n %d -v -a -p -i 'stdio(self)' '%s'",
+		gensiot, os_func_str, num_extra_threads, test->accepter);
     if (!constr) {
 	err = GE_NOMEM;
 	goto out_err;
@@ -1607,7 +1762,7 @@ run_oom_acc_test(struct oom_tests *test, long count, int *exitcode,
 
     for (;;) {
 	OOMUNLOCK(&od->lock);
-	rv = o->wait_intr_sigmask(od->waiter, 1, &timeout, &waitsigs);
+	rv = o->wait_intr_sigmask(od->waiter, 1, &timeout, proc_data);
 	OOMLOCK(&od->lock);
 	if (debug && rv == GE_TIMEDOUT) {
 	    printf("Waiting on err E\n");
@@ -1718,8 +1873,9 @@ print_test(struct oom_tests *test, char *tstr, bool close_acc, long count)
 static unsigned long
 run_oom_tests(struct oom_tests *test, char *tstr,
 	      int (*tester)(struct oom_tests *test, long count,
-			    int *exitcode, bool close_acc),
-	      int start, int end)
+			    int *exitcode, bool close_acc,
+			    struct env_info *env),
+	      int start, int end, struct env_info *env)
 {
     long count, errcount = 0;
     int rv, exit_code = 1;
@@ -1734,7 +1890,7 @@ run_oom_tests(struct oom_tests *test, char *tstr,
 	    if (rv)
 		goto next;
 	}
-	rv = tester(test, count, &exit_code, close_acc);
+	rv = tester(test, count, &exit_code, close_acc, env);
 	if (test->end_test)
 	    test->end_test(o, test);
 	if (rv && rv != GE_REMCLOSE && rv != GE_NOTREADY && rv != GE_SHUTDOWN
@@ -1855,18 +2011,21 @@ fill_random(void *buf, size_t buflen)
 
 static void
 run_tests(struct oom_tests *test, int testnrstart, int testnrend,
-	  unsigned long *skipcount, unsigned long *errcount)
+	  unsigned long *skipcount, unsigned long *errcount,
+	  unsigned long *testcount, struct env_info *env)
 {
     if (!check_oom_test_present(o, test)) {
 	(*skipcount)++;
 	return;
     }
+    printf("Running test %s\n", test->connecter);
+    (*testcount)++;
     *errcount += run_oom_tests(test, "oom", run_oom_test,
-			       testnrstart, testnrend);
+			       testnrstart, testnrend, env);
     if (test->accepter && !test->conacc)
 	*errcount += run_oom_tests(test, "oom acc",
 				   run_oom_acc_test,
-				   testnrstart, testnrend);
+				   testnrstart, testnrend, env);
     if (test->end_test_suite)
 	test->end_test_suite(o, test);
 }
@@ -1875,20 +2034,19 @@ int
 main(int argc, char *argv[])
 {
     int rv;
-#ifdef USE_PTHREADS
-    pthread_t loopth[3];
+    struct gensio_thread *loopth[3];
     struct gensio_waiter *loopwaiter[3];
-    unsigned int num_extra_threads = 3;
-#endif
     unsigned int i, j;
-    unsigned long errcount = 0;
-    unsigned long skipcount = 0;
+    unsigned long errcount = 0, skipcount = 0, testcount = 0;
     unsigned int repeat_count = 1;
-    struct sigaction sigdo;
-    sigset_t sigs;
     int testnr = -1, numtests = 0, testnrstart = -1, testnrend = MAX_LOOPS;
     gensio_time zerotime = { 0, 0 };
     struct oom_tests user_test;
+    bool list_tests = false;
+    char oshstr[20];
+    gensiods len = sizeof(oshstr);
+    char *s;
+    struct env_info env;
 
     memset(&user_test, 0, sizeof(user_test));
 
@@ -1897,12 +2055,30 @@ main(int argc, char *argv[])
     iodata_size %= MAX_IODATA_SIZE;
 
     /* This must be first so it gets picked up before any allocations. */
-    rv = setenv("GENSIO_MEMTRACK", "abort", 1);
+    rv = gensio_os_env_set("GENSIO_MEMTRACK", "abort");
     if (rv) {
-	fprintf(stderr, "Unable to set GENSIO_MEMTRACK");
+	fprintf(stderr, "Unable to set GENSIO_MEMTRACK: %s",
+		gensio_err_to_str(rv));
 	exit(1);
     }
 
+    rv = gensio_os_env_get("GENSIO_TEST_OS_HANDLER", oshstr, &len);
+    if (rv == 0) {
+	if (strcmp(oshstr, "glib") == 0) {
+	    use_glib = true;
+	} else if (strcmp(oshstr, "tcl") == 0) {
+	    use_tcl = true;
+	} else if (strcmp(oshstr, "default") == 0) {
+	    /* Nothing to do. */
+	} else {
+	    fprintf(stderr, "Unknown OS handler fron environment: %s\n",
+		    oshstr);
+	    exit(1);
+	}
+    } else if (rv != GE_NOTFOUND) {
+	fprintf(stderr, "Error getting GENSIO_TEST_OS_HANDLER: %s\n",
+		gensio_err_to_str(rv));
+    }
 #ifndef ENABLE_INTERNAL_TRACE
     fprintf(stderr, "Internal tracing disabled, cannot run oomtest\n");
     fprintf(stderr, "Configure with --enable-internal-trace to enable internal"
@@ -1910,49 +2086,9 @@ main(int argc, char *argv[])
     exit(77);
 #endif
 
-    rv = gensio_default_os_hnd(SIGUSR1, &o);
-    if (rv) {
-	fprintf(stderr, "Could not allocate OS handler: %s\n",
-		gensio_err_to_str(rv));
-	goto out_err;
-    }
-    o->vlog = do_vlog;
-
     for (j = 0; oom_tests[j].connecter; j++)
 	numtests++;
 
-    sigemptyset(&sigs);
-    sigaddset(&sigs, SIGCHLD);
-    sigaddset(&sigs, SIGPIPE); /* Ignore broken pipes. */
-    rv = sigprocmask(SIG_BLOCK, &sigs, NULL);
-    if (rv) {
-	perror("Could not set up signal mask");
-	exit(1);
-    }
-    rv = sigprocmask(SIG_BLOCK, NULL, &waitsigs);
-    if (rv) {
-	perror("Could not get signal mask");
-	exit(1);
-    }
-    sigdelset(&waitsigs, SIGCHLD);
-
-    memset(&sigdo, 0, sizeof(sigdo));
-    sigdo.sa_handler = handle_sigchld;
-    sigdo.sa_flags = SA_NOCLDSTOP;
-    rv = sigaction(SIGCHLD, &sigdo, NULL);
-    if (rv) {
-	perror("Could not set up sigchld handler");
-	exit(1);
-    }
-
-    sigdo.sa_handler = handle_sigusr1;
-    sigdo.sa_flags = 0;
-    rv = sigaction(SIGUSR1, &sigdo, NULL);
-    if (rv) {
-	perror("Could not set up siguser1 handler");
-	exit(1);
-    }
-
     for (i = 1; i < argc; i++) {
 	if (argv[i][0] != '-')
 	    break;
@@ -1965,15 +2101,7 @@ main(int argc, char *argv[])
 	} else if (strcmp(argv[i], "-b") == 0) {
 	    sleep_on_timeout_err = true;
 	} else if (strcmp(argv[i], "-l") == 0) {
-	    for (j = 0; oom_tests[j].connecter; j++) {
-		if (!check_oom_test_present(o, oom_tests + j))
-		    continue;
-		printf("%d : %s %s\n", j, oom_tests[j].connecter,
-		       oom_tests[j].accepter ? oom_tests[j].accepter : "");
-		if (oom_tests[j].end_test_suite)
-		    oom_tests[j].end_test_suite(o, oom_tests + j);
-	    }
-	    exit(0);
+	    list_tests = true;
 	} else if (strcmp(argv[i], "-t") == 0) {
 	    i++;
 	    if (i >= argc) {
@@ -2000,7 +2128,6 @@ main(int argc, char *argv[])
 		exit(1);
 	    }
 	    testnrstart = strtol(argv[i], NULL, 0);
-#ifdef USE_PTHREADS
 	} else if (strcmp(argv[i], "-n") == 0) {
 	    i++;
 	    if (i >= argc) {
@@ -2008,7 +2135,6 @@ main(int argc, char *argv[])
 		exit(1);
 	    }
 	    num_extra_threads = strtol(argv[i], NULL, 0);
-#endif
 	} else if (strcmp(argv[i], "-e") == 0) {
 	    i++;
 	    if (i >= argc) {
@@ -2039,12 +2165,65 @@ main(int argc, char *argv[])
 	    user_test.connecter = argv[i];
 	} else if (strcmp(argv[i], "-w") == 0) {
 	    user_test.allow_no_err_on_trig = true;
+	} else if (strcmp(argv[i], "--glib") == 0) {
+	    use_glib = true;
+	} else if (strcmp(argv[i], "--tcl") == 0) {
+	    use_tcl = true;
 	} else {
 	    fprintf(stderr, "Unknown argument: '%s'\n", argv[i]);
 	    exit(1);
 	}
     }
 
+    if (use_glib) {
+#ifndef HAVE_GLIB
+	fprintf(stderr, "glib specified, but glib OS handler not avaiable.\n");
+	exit(1);
+#else
+	os_func_str = " --glib";
+	rv = gensio_glib_funcs_alloc(&o);
+#endif
+    } else if (use_tcl) {
+#ifndef HAVE_TCL
+	fprintf(stderr, "tcl specified, but tcl OS handler not avaiable.\n");
+	exit(1);
+#else
+	if (num_extra_threads > 0)
+	    fprintf(stderr, "Number of extra threads is %u, incompatible with"
+		    " TCL, forcing to 0\n", num_extra_threads);
+	num_extra_threads = 0;
+	os_func_str = " --tcl";
+	rv = gensio_tcl_funcs_alloc(&o);
+#endif
+    } else {
+	rv = gensio_default_os_hnd(GENSIO_DEF_WAKE_SIG, &o);
+    }
+    if (rv) {
+	fprintf(stderr, "Could not allocate OS handler: %s\n",
+		gensio_err_to_str(rv));
+	goto out_err;
+    }
+    o->vlog = do_vlog;
+
+    rv = gensio_os_proc_setup(o, &proc_data);
+    if (rv) {
+	fprintf(stderr, "Error setting up process data: %s\n",
+		gensio_err_to_str(rv));
+	exit(1);
+    }
+
+    if (list_tests) {
+	for (j = 0; oom_tests[j].connecter; j++) {
+	    if (!check_oom_test_present(o, oom_tests + j))
+		continue;
+	    printf("%d : %s %s\n", j, oom_tests[j].connecter,
+		   oom_tests[j].accepter ? oom_tests[j].accepter : "");
+	    if (oom_tests[j].end_test_suite)
+		oom_tests[j].end_test_suite(o, oom_tests + j);
+	}
+	exit(0);
+    }
+
     printf("iodata_size is %lu\n", (unsigned long) iodata_size);
 
     iodata = malloc(iodata_size);
@@ -2056,16 +2235,22 @@ main(int argc, char *argv[])
 	return 1;
 
     if (i >= argc) {
-	gensiot = getenv("GENSIOT");
-	if (!gensiot) {
-	    fprintf(stderr, "No gensiot given\n");
+	rv = gensio_os_env_getalloc(o, "GENSIOT", &s);
+	if (rv != 0) {
+	    fprintf(stderr, "Can't get GENSIOT: %s\n", gensio_err_to_str(rv));
 	    exit(1);
 	}
     } else {
-	gensiot = argv[i];
+	s = argv[i];
     }
+    gensiot = gensio_quote_string(o, s);
+    if (!gensiot) {
+	fprintf(stderr, "Out of memory copying gensiot string\n");
+	exit(1);
+    }
+    if (i >= argc)
+	o->free(o, s);
 
-#ifdef USE_PTHREADS
     for (i = 0; i < num_extra_threads; i++) {
 	loopwaiter[i] = o->alloc_waiter(o);
 	if (!loopwaiter[i]) {
@@ -2073,38 +2258,47 @@ main(int argc, char *argv[])
 	    goto out_err;
 	}
 
-	rv = pthread_create(&loopth[i], NULL, gensio_loop, loopwaiter[i]);
+	rv = gensio_os_new_thread(o, gensio_loop, loopwaiter[i], &loopth[i]);
 	if (rv) {
-	    perror("Could not allocate loop thread");
+	    fprintf(stderr, "Could not allocate loop thread: %s",
+		    gensio_err_to_str(rv));
 	    goto out_err;
 	}
     }
-#endif
 
-    for (j = 0; j < repeat_count; j++) {
-	if (user_test.connecter) {
-	    run_tests(&user_test, testnrstart, testnrend,
-		      &skipcount, &errcount);
-	} else if (testnr < 0) {
-	    for (i = 0; oom_tests[i].connecter; i++) {
-		if (oom_tests[i].no_default_run)
-		    continue;
-		run_tests(oom_tests + i, testnrstart, testnrend,
-			  &skipcount, &errcount);
+    rv = gensio_os_argvenv_alloc(o, &env.argv, &env.args, &env.argc);
+    if (rv) {
+	fprintf(stderr, "Could not allocate environment array: %s",
+		gensio_err_to_str(rv));
+	goto out_err;
+    }
+
+    if (user_test.connecter) {
+	run_tests(&user_test, testnrstart, testnrend,
+		  &skipcount, &errcount, &testcount, &env);
+    } else {
+	for (j = 0; j < repeat_count; j++) {
+	    if (testnr < 0) {
+		for (i = 0; oom_tests[i].connecter; i++) {
+		    if (oom_tests[i].no_default_run)
+			continue;
+		    run_tests(oom_tests + i, testnrstart, testnrend,
+			      &skipcount, &errcount, &testcount, &env);
+		}
+	    } else {
+		run_tests(oom_tests + testnr, testnrstart, testnrend,
+			  &skipcount, &errcount, &testcount, &env);
 	    }
-	} else {
-	    run_tests(oom_tests + testnr, testnrstart, testnrend,
-		      &skipcount, &errcount);
 	}
     }
 
-#ifdef USE_PTHREADS
+    gensio_argv_free(o, env.argv);
+
     for (i = 0; i < num_extra_threads; i++) {
 	o->wake(loopwaiter[i]);
-	pthread_join(loopth[i], NULL);
+	gensio_os_wait_thread(loopth[i]);
 	o->free_waiter(loopwaiter[i]);
     }
-#endif
 
     for (i = 0; oom_tests[i].connecter; i++) {
 	if (oom_tests[i].free_connecter)
@@ -2116,11 +2310,16 @@ main(int argc, char *argv[])
 	;
     cleanup_ods();
     gensio_cleanup_mem(o);
-    gensio_sel_exit(!!errcount);
+    if (testcount == 0)
+	return 77;
+    gensio_osfunc_exit(!!errcount);
 
  out_err:
+    gensio_os_proc_cleanup(proc_data);
+    o->free(o, gensiot);
     cleanup_ods();
     gensio_cleanup_mem(o);
-    gensio_sel_exit(1);
+    o->free_funcs(o);
+    gensio_osfunc_exit(1);
     return 1;
 }
