@@ -11,6 +11,8 @@ import gensio
 check_sctp()
 
 print("Test mux sctp oob")
-ta = TestAccept(o, "mux,sctp,localhost,", "mux,sctp,0", do_oob_test,
-                chunksize = 64)
+TestAccept(o, "mux,sctp,localhost,", "mux,sctp,0", do_oob_test,
+           chunksize = 64)
 
+del o
+test_shutdown()
