@@ -11,5 +11,7 @@ import gensio
 check_sctp()
 
 print("Test mux sctp small")
-ta = TestAccept(o, "mux,sctp,localhost,", "mux,sctp,0", do_small_test,
-                chunksize = 64)
+TestAccept(o, "mux,sctp,localhost,", "mux,sctp,0", do_small_test,
+           chunksize = 64)
+del o
+test_shutdown()
