@@ -13,5 +13,8 @@ io = alloc_io(o, "stdio,cat", chunksize = 64)
 check_raddr(io, "stdio basic", 'stdio,"cat"')
 test_dataxfer(io, io, "This is a test string!")
 io_close(io)
+del io
+del o
+test_shutdown()
 print("  Success!")
 
