@@ -11,3 +11,5 @@ import gensio
 print("Test tcp accepter connect")
 TestAcceptConnect(o, "tcp,0", "tcp,0", "tcp,localhost,",
                   do_small_test)
+del o
+test_shutdown()
