@@ -9,5 +9,7 @@ from utils import *
 import gensio
 
 print("Test telnet small")
-ta = TestAccept(o, "telnet,tcp,localhost,", "telnet(rfc2217=true),tcp,0",
-                do_small_test, chunksize = 64, is_sergensio = True)
+TestAccept(o, "telnet,tcp,localhost,", "telnet(rfc2217=true),tcp,0",
+           do_small_test, chunksize = 64, is_sergensio = True)
+del o
+test_shutdown()
