@@ -66,7 +66,8 @@ class HandleData:
     """
 
     def __init__(self, o, iostr, name = None, chunksize=10240,
-                 io = None, password = None, expect_remclose = True):
+                 io = None, password = None, expect_remclose = True,
+                 val_2fa = None):
         """Start a gensio object with this handler"""
         if (name):
             self.name = name
@@ -88,8 +89,10 @@ class HandleData:
         self.expected_sig_server_cb = False
         self.expected_sig_server_val = None
         self.ignore_input = False
+        self.waiting_rem_close = False
         self.stream = None
         self.password = password
+        self.val_2fa = val_2fa
         if (io):
             self.io = io
             io.set_cbs(self)
@@ -175,7 +178,7 @@ class HandleData:
         if (debug or self.debug) and iolen != None and buf is not None:
             print("%s: Got %d bytes at pos %d of %d" % (self.name, len(buf),
                                                         self.compared, iolen))
-        if (debug >= 2 or self.debug >= 2):
+        if ((debug >= 2 or self.debug >= 2) and not err):
             s = ""
             for i in range(0, len(buf)):
                 if curses.ascii.isprint(buf[i]):
@@ -188,6 +191,8 @@ class HandleData:
             print("%s: Got data: (err %s %d bytes) %s" % (self.name, str(err),
                                                           len(buf), s))
         if self.expecting_remclose and err == "Remote end closed connection":
+            if (self.waiting_rem_close):
+                self.waiter.wake()
             io.read_cb_enable(False)
             return 0
         if (err):
@@ -307,6 +312,11 @@ class HandleData:
             return gensio.ENOTSUP
         return self.password
 
+    def request_2fa(self, io):
+        if self.val_2fa is None:
+            return gensio.ENOTSUP
+        return self.val_2fa
+
     def set_expecting_oob(self, data):
         self.to_compare_oob = data
         return
@@ -507,6 +517,7 @@ def alloc_io(o, iostr, do_open = True, chunksize = 10240):
 
     If do_open is True (default), open it, too.
     """
+    gensios_enabled.check_iostr_gensios(iostr)
     h = HandleData(o, iostr, chunksize = chunksize)
     if (do_open):
         h.io.open_s()
@@ -636,17 +647,52 @@ class Logger:
         print("***%s log: %s" % (level, log))
 
 gensio.gensio_set_log_mask(gensio.GENSIO_LOG_MASK_ALL)
-o = gensio.alloc_gensio_selector(Logger());
+
+oshndname = os.getenv("GENSIO_TEST_OS_HANDLER")
+if oshndname is None or oshndname == "default":
+    o = gensio.alloc_gensio_selector(Logger());
+elif oshndname == "glib":
+    import gensioglib
+    o = gensioglib.alloc_glib_os_funcs(Logger());
+elif oshndname == "tcl":
+    import gensiotcl
+    o = gensiotcl.alloc_tcl_os_funcs(Logger());
+else:
+    print("Unknown OS handler name: " + oshndname)
+    sys.exit(1)
+
+def test_shutdown():
+    global o
+    w = gensio.waiter(o)
+    count = 0
+    while gensio.gensio_num_alloced() > 0:
+        count += 1
+        if (count > 100):
+            raise Exception("All gensios were not freed in time" % c)
+        w.service(1)
+    while w.service_now() == 0:
+        # Give some time for everyting to clear out.
+        pass
+    del w
+    c = sys.getrefcount(o)
+    if c != 2:
+        raise Exception("OS object refcount was %d, not 2" % c)
+    c = gensio.get_os_funcs_refcount(o)
+    if c != 1:
+        raise Exception("OS funcs refcount was %d, not 1" % c)
+    gensio.gensio_cleanup_mem(o)
+    del o
 
 def check_raddr(io, testname, expected):
-    r = io.control(gensio.GENSIO_CONTROL_DEPTH_FIRST, True,
+    r = io.control(gensio.GENSIO_CONTROL_DEPTH_FIRST, gensio.GENSIO_CONTROL_GET,
                    gensio.GENSIO_CONTROL_RADDR, "0")
     if r != expected:
         raise Exception("%s raddr was not '%s', it was '%s'" %
                         (testname, expected, r));
 
 def check_laddr(acc, testname, expected):
-    r = acc.control(0, True, gensio.GENSIO_ACC_CONTROL_LADDR, "0")
+    r = acc.control(0, gensio.GENSIO_CONTROL_GET,
+                    gensio.GENSIO_ACC_CONTROL_LADDR, "0")
     if r != expected:
         expected = expected.replace("127.0.0.1", "::1")
         expected = expected.replace("ipv4", "ipv6")
@@ -655,7 +701,8 @@ def check_laddr(acc, testname, expected):
                         (testname, expected, r));
 
 def check_port(acc, testname, expected):
-    r = acc.control(0, True, gensio.GENSIO_ACC_CONTROL_LPORT, "0")
+    r = acc.control(0, gensio.GENSIO_CONTROL_GET,
+                    gensio.GENSIO_ACC_CONTROL_LPORT, "0")
     if r != expected:
         raise Exception("%s port was not '%s', it was '%s'" %
                         (testname, expected, r));
@@ -667,66 +714,92 @@ class TestAccept:
                  chunksize = 10240, get_port = True, except_on_log = False,
                  is_sergensio = False):
         self.o = o
-        self.except_on_log = except_on_log
-        if (name):
-            self.name = name
-        else:
-            self.name = accstr
+        self.io1 = None
         self.io2 = None
-        self.waiter = gensio.waiter(o)
-        gensios_enabled.check_iostr_gensios(accstr)
-        self.acc = gensio.gensio_accepter(o, accstr, self);
-        if is_sergensio:
-            sga = self.acc.cast_to_sergensio_acc()
-            if not sga:
-                raise Exception("Cast to sergensio_accepter failed");
-            ga = sga.cast_to_gensio_acc()
-            del sga
-            del ga
-        else:
-            sga = None
-            try:
+        self.acc = None
+
+        try:
+            self.except_on_log = except_on_log
+            if (name):
+                self.name = name
+            else:
+                self.name = accstr
+            if debug:
+                print("TestAccept " + self.name);
+            self.waiter = gensio.waiter(o)
+            gensios_enabled.check_iostr_gensios(accstr)
+            self.acc = gensio.gensio_accepter(o, accstr, self);
+            if is_sergensio:
                 sga = self.acc.cast_to_sergensio_acc()
+                if not sga:
+                    raise Exception("Cast to sergensio_accepter failed");
+                ga = sga.cast_to_gensio_acc()
+                del sga
+                del ga
+            else:
+                sga = None
+                try:
+                    sga = self.acc.cast_to_sergensio_acc()
+                except:
+                    pass
+                if sga:
+                    raise Exception("Cast to sergensio_accepter succeeded");
+            if debug:
+                print("acc startup");
+            self.acc.startup()
+            self.waiter.service(1) # Wait a bit for the accepter to start up.
+
+            if get_port:
+                port = self.acc.control(gensio.GENSIO_CONTROL_DEPTH_FIRST,
+                                        gensio.GENSIO_CONTROL_GET,
+                                        gensio.GENSIO_ACC_CONTROL_LPORT, "0")
+            else:
+                port = ""
+            io1str = io1str + port
+            io1 = alloc_io(o, io1str, do_open = False,
+                           chunksize = chunksize)
+            self.io1 = io1
+            if expected_acc_laddr:
+                expected_acc_laddr = expected_acc_laddr + port
+            if expected_raddr:
+                expected_raddr = expected_raddr + port
+
+            if expected_acc_laddr:
+                check_laddr(self.acc, self.name, expected_acc_laddr)
+            if debug:
+                print("io1 open " + self.name);
+            try:
+                io1.open_s()
             except:
-                pass
-            if sga:
-                raise Exception("Cast to sergensio_accepter succeeded");
-        self.acc.startup()
-        self.waiter.service(1) # Wait a bit for the accepter to start up.
-
-        if get_port:
-            port = self.acc.control(gensio.GENSIO_CONTROL_DEPTH_FIRST, True,
-                                    gensio.GENSIO_ACC_CONTROL_LPORT, "0")
-        else:
-            port = ""
-        io1str = io1str + port
-        io1 = alloc_io(o, io1str, do_open = False,
-                       chunksize = chunksize)
-        self.io1 = io1
-        if expected_acc_laddr:
-            expected_acc_laddr = expected_acc_laddr + port
-        if expected_raddr:
-            expected_raddr = expected_raddr + port
-
-        if expected_acc_laddr:
-            check_laddr(self.acc, self.name, expected_acc_laddr)
-        io1.open_s()
-        if expected_raddr:
-            check_raddr(io1, self.name, expected_raddr)
-        if (io1_dummy_write):
-            # For UDP, kick start things.
-            io1.write(io1_dummy_write, None)
-        self.wait()
-        if (io1_dummy_write):
-            self.io2.handler.set_compare(io1_dummy_write)
-            if (self.io2.handler.wait_timeout(1000) == 0):
-                raise Exception(("%s: %s: " % ("test_accept",
-                                               self.io2.handler.name)) +
-                        ("Timed out waiting for dummy read at byte %d" %
-                         self.io2.handler.compared))
-        tester(self.io1, self.io2)
-        if do_close:
-            self.close()
+                del io1.handler.io
+                del io1.handler
+                raise
+            if expected_raddr:
+                check_raddr(io1, self.name, expected_raddr)
+            if (io1_dummy_write):
+                # For UDP, kick start things.
+                io1.write(io1_dummy_write, None)
+            if debug:
+                print("wait 1 " + self.name);
+            # Wait for the accept to happen
+            if (self.wait_timeout(1000) == 0):
+                raise Exception(("%s: %s: " % ("test_accept", self.name)) +
+                                ("Timed out waiting for initial connection"))
+            if (io1_dummy_write):
+                self.io2.handler.set_compare(io1_dummy_write)
+                if (self.io2.handler.wait_timeout(1000) == 0):
+                    raise Exception(("%s: %s: " % ("test_accept",
+                                                   self.io2.handler.name)) +
+                                ("Timed out waiting for dummy read at byte %d" %
+                                     self.io2.handler.compared))
+            tester(self.io1, self.io2)
+            if do_close:
+                self.close()
+        except:
+            self.io1 = None
+            self.io2 = None
+            self.acc = None
+            raise
 
     def close(self):
         self.io1.read_cb_enable(False)
@@ -740,12 +813,13 @@ class TestAccept:
             io_close(self.io2)
 
         # Break all the possible circular references.
-        del self.io1
-        del self.io2
-        del self.acc
+        self.io1 = None
+        self.io2 = None
+        self.acc = None
 
     def new_connection(self, acc, io):
         HandleData(self.o, None, io = io, name = self.name)
+        print("New connection " + self.name);
         self.io2 = io
         self.waiter.wake()
 
@@ -759,6 +833,9 @@ class TestAccept:
     def wait(self):
         self.waiter.wait(1)
 
+    def wait_timeout(self, timeout):
+        return self.waiter.wait_timeout(1, timeout)
+
 def do_test(io1, io2):
     test_dataxfer(io1, io2, "This is a test string!")
     print("  Success!")
@@ -806,56 +883,72 @@ class TestAcceptConnect:
                  io1_dummy_write = None, CA=None, do_close = True,
                  auth_begin_rv = gensio.GE_NOTSUP, expect_pw = None,
                  expect_pw_rv = gensio.GE_NOTSUP, password = None,
-                 expect_remclose = True, use_port = True):
+                 expect_remclose = True, use_port = True,
+                 expect_2fa = None, expect_2fa_rv = gensio.GE_NOTSUP,
+                 val_2fa = None):
         self.o = o
-        if (name):
-            self.name = name
-        else:
-            self.name = iostr
-        self.waiter = gensio.waiter(o)
-        gensios_enabled.check_iostr_gensios(iostr)
-        gensios_enabled.check_iostr_gensios(io2str)
-        self.acc = gensio.gensio_accepter(o, iostr, self);
-        self.acc.startup()
-        self.acc2 = gensio.gensio_accepter(o, io2str, self);
-        self.acc2.startup()
-        if (use_port):
-            port = self.acc.control(gensio.GENSIO_CONTROL_DEPTH_FIRST, True,
-                                    gensio.GENSIO_ACC_CONTROL_LPORT, "0")
-            io3str = io3str + port
-        self.io1 = self.acc2.str_to_gensio(io3str, None);
+        self.io1 = None
         self.io2 = None
-        self.CA = CA
-        h = HandleData(o, io3str, io = self.io1, password = password,
-                             expect_remclose = expect_remclose)
-        self.auth_begin_rv = auth_begin_rv
-        self.expect_pw = expect_pw
-        self.expect_pw_rv = expect_pw_rv
+        self.acc = None
+        self.acc2 = None
         try:
-            self.io1.open_s()
-        except:
-            self.io1 = None
-            self.close()
-            raise
-        self.io1.read_cb_enable(True)
-        if (io1_dummy_write):
-            # For UDP, kick start things.
-            self.io1.write(io1_dummy_write, None)
-        try:
-            self.wait()
+            if (name):
+                self.name = name
+            else:
+                self.name = iostr
+            self.waiter = gensio.waiter(o)
+            gensios_enabled.check_iostr_gensios(iostr)
+            gensios_enabled.check_iostr_gensios(io2str)
+            self.acc = gensio.gensio_accepter(o, iostr, self);
+            self.acc.startup()
+            self.acc2 = gensio.gensio_accepter(o, io2str, self);
+            self.acc2.startup()
+            if (use_port):
+                port = self.acc.control(gensio.GENSIO_CONTROL_DEPTH_FIRST,
+                                        gensio.GENSIO_CONTROL_GET,
+                                        gensio.GENSIO_ACC_CONTROL_LPORT, "0")
+                io3str = io3str + port
+            self.io1 = self.acc2.str_to_gensio(io3str, None);
+            self.io2 = None
+            self.CA = CA
+            h = HandleData(o, io3str, io = self.io1, password = password,
+                           expect_remclose = expect_remclose,
+                           val_2fa = val_2fa)
+            self.auth_begin_rv = auth_begin_rv
+            self.expect_pw = expect_pw
+            self.expect_pw_rv = expect_pw_rv
+            self.expect_2fa = expect_2fa
+            self.expect_2fa_rv = expect_2fa_rv
+            try:
+                self.io1.open_s()
+            except:
+                del self.io1.handler.io
+                del self.io1.handler
+                self.io1 = None
+                self.close()
+                raise
+            self.io1.read_cb_enable(True)
+            if (io1_dummy_write):
+                # For UDP, kick start things.
+                self.io1.write(io1_dummy_write, None)
+            try:
+                self.wait()
+            except:
+                self.close()
+                raise
+            if (io1_dummy_write):
+                self.io2.handler.set_compare(io1_dummy_write)
+                if (self.io2.handler.wait_timeout(1000) == 0):
+                    raise Exception(("%s: %s: " % ("test_accept",
+                                                   self.io2.handler.name)) +
+                               ("Timed out waiting for dummy read at byte %d" %
+                                     self.io2.handler.compared))
+            tester(self.io1, self.io2)
+            if do_close:
+                self.close()
         except:
             self.close()
             raise
-        if (io1_dummy_write):
-            self.io2.handler.set_compare(io1_dummy_write)
-            if (self.io2.handler.wait_timeout(1000) == 0):
-                raise Exception(("%s: %s: " % ("test_accept",
-                                               self.io2.handler.name)) +
-                        ("Timed out waiting for dummy read at byte %d" %
-                         self.io2.handler.compared))
-        tester(self.io1, self.io2)
-        if do_close:
-            self.close()
 
     def close(self):
         if (self.io1):
@@ -868,12 +961,14 @@ class TestAcceptConnect:
             io_close(self.io2)
 
         # Break all the possible circular references.
-        del self.io1
-        del self.io2
-        self.acc.shutdown_s()
-        del self.acc
-        self.acc2.shutdown_s()
-        del self.acc2
+        self.io1 = None
+        self.io2 = None
+        if self.acc:
+            self.acc.shutdown_s()
+        self.acc = None
+        if self.acc2:
+            self.acc2.shutdown_s()
+        self.acc2 = None
 
     def new_connection(self, acc, io):
         HandleData(self.o, None, io = io, name = self.name)
@@ -885,7 +980,8 @@ class TestAcceptConnect:
 
     def precert_verify(self, acc, io):
         if self.CA:
-            io.control(0, False, gensio.GENSIO_CONTROL_CERT_AUTH, self.CA)
+            io.control(0, gensio.GENSIO_CONTROL_SET,
+                       gensio.GENSIO_CONTROL_CERT_AUTH, self.CA)
             return gensio.GE_NOTSUP
         return gensio.GE_NOTSUP
 
@@ -897,6 +993,15 @@ class TestAcceptConnect:
                             % (self.expect_pw, password))
         return self.expect_pw_rv
 
+    def verify_2fa(self, acc, io, val_2fa):
+        if self.expect_2fa is None:
+            raise Exception("got 2-factor auth verify when none expected")
+        if self.expect_2fa != val_2fa:
+            raise Exception(
+                "Invalid 2-factor auth in verify, expected %s, got %s"
+                % (self.expect_2fa, val_2fa))
+        return self.expect_2fa_rv
+
     def accepter_log(self, acc, level, logstr):
         print("***%s LOG: %s: %s" % (level, self.name, logstr))
 
@@ -977,7 +1082,7 @@ if e is not None:
     else:
         ttypipe = None
 else:
-    ttypipe = [ "/dev/ttyPipeA0", "/dev/ttyPipeB0" ]
+    ttypipe = [ "/dev/ttyPipeA1", "/dev/ttyPipeB1" ]
 
 ttypipe_def = ttypipe
 if ttypipe and (not os.path.exists(ttypipe[0])
@@ -1000,7 +1105,8 @@ def check_pipe_dev(is_serialsim = False):
         sys.exit(77)
 
 def remote_id_int(io):
-    return int(io.control(0, True, gensio.GENSIO_CONTROL_REMOTE_ID, None))
+    return int(io.control(0, gensio.GENSIO_CONTROL_GET,
+                          gensio.GENSIO_CONTROL_REMOTE_ID, None))
 
 def check_sctp():
     try:
