@@ -34,10 +34,10 @@
 #include <pwd.h>
 #include <assert.h>
 #include <sys/ioctl.h>
-#include <termios.h>
+#include <fcntl.h>
+#include <arpa/inet.h> /* For htonl and friends. */
 
 #include <security/pam_appl.h>
-#include <security/pam_misc.h>
 
 #include <gensio/gensio.h>
 #include <gensio/gensio_builtins.h>
@@ -97,6 +97,159 @@ static const char *default_configfile = SYSCONFDIR "/gtlssh/gtlsshd.conf";
 
 static const char *pid_file = NULL;
 
+static int
+write_s_nl_addc(struct gensio *io, char *obuf, char c,
+		gensiods *pos, gensiods len, gensio_time *timeout)
+{
+    int err = 0;
+
+    obuf[(*pos)++] = c;
+    if (*pos >= len) {
+	err = gensio_write_s(io, NULL, obuf, len, timeout);
+	*pos = 0;
+    }
+    return err;
+}
+
+static int
+write_s_nl(struct gensio *io, const char *buf, gensiods len,
+	   gensio_time *timeout)
+{
+    char buf2[100];
+    gensiods i, j;
+    int err;
+
+    for (i = 0, j = 0; i < len; i++) {
+	if (buf[i] == '\n') {
+	    err = write_s_nl_addc(io, buf2, '\r', &j, sizeof(buf2), timeout);
+	    if (err)
+		break;
+	}
+	err = write_s_nl_addc(io, buf2, buf[i], &j, sizeof(buf2), timeout);
+	if (err)
+	    break;
+    }
+    if (!err && j)
+	err = gensio_write_s(io, NULL, buf2, j, timeout);
+
+    return err;
+}
+
+static int
+write_file_to_gensio(const char *filename, struct gensio *io,
+		     struct gensio_os_funcs *o, gensio_time *timeout,
+		     bool xlatnl)
+{
+    int err;
+    int fd;
+    char buf[100];
+    int count;
+
+    err = gensio_set_sync(io);
+    if (err)
+	return err;
+
+    fd = open(filename, O_RDONLY);
+    if (fd == -1) {
+	err = gensio_os_err_to_err(o, errno);
+	goto out_unsync;
+    }
+
+    while (true) {
+	count = read(fd, buf, sizeof(buf));
+	if (count == -1) {
+	    err = gensio_os_err_to_err(o, errno);
+	    break;
+	}
+	if (count == 0)
+	    break;
+	if (xlatnl)
+	    err = write_s_nl(io, buf, count, timeout);
+	else
+	    err = gensio_write_s(io, NULL, buf, count, timeout);
+	if (err)
+	    break;
+    }
+
+    close(fd);
+
+ out_unsync:
+    gensio_clear_sync(io);
+
+    return err;
+}
+
+static int
+write_buf_to_gensio(const char *buf, gensiods len, struct gensio *io,
+		    gensio_time *timeout, bool xlatnl)
+{
+    int err;
+
+    err = gensio_set_sync(io);
+    if (err)
+	return err;
+
+    if (xlatnl)
+	err = write_s_nl(io, buf, len, timeout);
+    else
+	err = gensio_write_s(io, NULL, buf, len, timeout);
+
+    gensio_clear_sync(io);
+
+    return err;
+}
+
+static int
+write_str_to_gensio(const char *str, struct gensio *io,
+		    gensio_time *timeout, bool xlatnl)
+{
+    return write_buf_to_gensio(str, strlen(str), io, timeout, xlatnl);
+}
+
+static int
+read_rsp_from_gensio(char *buf, gensiods *len, struct gensio *io,
+		     gensio_time *timeout, bool echo)
+{
+    int err;
+    gensiods pos = 0, count;
+    gensiods size = *len;
+    char c;
+
+    err = gensio_set_sync(io);
+    if (err)
+	return err;
+
+    while (true) {
+	err = gensio_read_s(io, &count, &c, 1, timeout);
+	if (err)
+	    break;
+	if (count == 0) {
+	    err = GE_TIMEDOUT;
+	    break;
+	}
+	if (c == '\r' || c == '\n')
+	    break;
+	if (c == '\b' || c == 0x7f) {
+	    if (pos > 0)
+		pos--;
+	    if (echo)
+		gensio_write_s(io, NULL, "\b \b", 3, timeout);
+	    continue;
+	}
+	if (pos < size - 1) {
+	    buf[pos++] = c;
+	    if (echo)
+		gensio_write_s(io, NULL, &c, 1, timeout);
+	}
+    }
+
+    gensio_clear_sync(io);
+    buf[pos] = '\0';
+    *len = pos;
+
+    return err;
+}
+
 static void
 make_pidfile(void)
 {
@@ -245,8 +398,8 @@ handle_winch(struct per_con_info *pcinfo,
     if (msglen < 8)
 	return;
 
-    err = gensio_control(pcinfo->io2, 0, true, GENSIO_CONTROL_RADDR_BIN,
-			 (char *) &ptym, &len);
+    err = gensio_control(pcinfo->io2, 0, GENSIO_CONTROL_GET,
+			 GENSIO_CONTROL_RADDR_BIN, (char *) &ptym, &len);
     if (err)
 	return;
 
@@ -335,13 +488,26 @@ acc_shutdown(struct gensio_accepter *acc, void *done_data)
 
 static pam_handle_t *pamh;
 static char *passwd;
+static char *val_2fa;
+static gensiods len_2fa;
 static bool pam_started = false;
 static bool pam_cred_set = false;
+static bool authed_by_cert = false;
 static char username[100];
 static char *homedir;
 static int pam_err;
 static uid_t uid = -1;
 static gid_t gid = -1;
+static bool interactive_login = true;
+
+static struct gtlssh_aux_data aux_data;
+static gensiods aux_data_len;
+
+/*
+ * If this is set and a certificate auth happens, we use this to start
+ * PAM.  This way we can do 2-factor auth with certificates.
+ */
+static const char *pam_cert_auth_progname;
 
 /*
  * Ambiguity in spec: is it an array of pointers or a pointer to an array?
@@ -353,64 +519,9 @@ static gid_t gid = -1;
 # define PAM_MSG_MEMBER(msg, n, member) ((msg)[(n)]->member)
 #endif
 
-static int
-pam_cb(int num_msg, const struct pam_message **msg,
-       struct pam_response **resp, void *appdata_ptr)
-{
-    int i, j;
-    struct pam_response *reply = NULL;
-
-    if (num_msg <= 0 || num_msg > PAM_MAX_NUM_MSG)
-	return PAM_CONV_ERR;
-
-    reply = malloc(sizeof(*reply) * num_msg);
-    if (!reply)
-	return PAM_CONV_ERR;
-
-    for (i = 0; i < num_msg; i++) {
-	reply[i].resp = NULL;
-	reply[i].resp_retcode = 0;
-
-	switch (PAM_MSG_MEMBER(msg, i, msg_style)) {
-	case PAM_PROMPT_ECHO_OFF:
-	    if (passwd) {
-		reply[i].resp = strdup(passwd);
-		if (!reply[i].resp)
-		    goto out_err;
-	    }
-	    break;
-
-	case PAM_PROMPT_ECHO_ON:
-	    break;
-
-	case PAM_ERROR_MSG:
-	    syslog(LOG_ERR, "Error from pam: %s",
-		   PAM_MSG_MEMBER(msg, i, msg));
-	    break;
-
-	case PAM_TEXT_INFO:
-	    break;
-
-	default:
-	    goto out_err;
-	}
-    }
-    *resp = reply;
-    return PAM_SUCCESS;
-
- out_err:
-    for (j = 0; j < i; j++) {
-	if (reply[j].resp)
-	    free(reply[j].resp);
-    }
-    free(reply);
-    return PAM_CONV_ERR;
-}
-
-static struct pam_conv auth_conv = { pam_cb, NULL };
-
 static bool permit_root = false;
 static bool pw_login = false;
+static bool do_2fa = false;
 
 static int
 get_vals_from_service(char ***rvals, unsigned int *rvlen,
@@ -454,6 +565,33 @@ get_vals_from_service(char ***rvals, unsigned int *rvlen,
     return 0;
 }
 
+static int
+get_2fa(struct gensio *io)
+{
+    int err;
+    char dummy;
+
+    len_2fa = 0;
+    err = gensio_control(io, 0, GENSIO_CONTROL_GET, GENSIO_CONTROL_2FA,
+			 &dummy, &len_2fa);
+    if (err) {
+	if (err == GE_DATAMISSING)
+	    return 0;
+	return err;
+    }
+    val_2fa = malloc(len_2fa + 1);
+    if (!val_2fa)
+	return GE_NOMEM;
+    val_2fa[len_2fa] = '\0'; /* nil terminate, 2fa may be binary. */
+    err = gensio_control(io, 0, GENSIO_CONTROL_GET, GENSIO_CONTROL_2FA,
+			 val_2fa, &len_2fa);
+    if (err) {
+	free(val_2fa);
+	val_2fa = NULL;
+    }
+    return err;
+}
+
 static int
 certauth_event(struct gensio *io, void *user_data, int event, int ierr,
 	       unsigned char *buf, gensiods *buflen,
@@ -468,8 +606,8 @@ certauth_event(struct gensio *io, void *user_data, int event, int ierr,
 	struct passwd *pw;
 
 	len = sizeof(username);
-	err = gensio_control(io, 0, true, GENSIO_CONTROL_USERNAME, username,
-			     &len);
+	err = gensio_control(io, 0, GENSIO_CONTROL_GET, GENSIO_CONTROL_USERNAME,
+			     username, &len);
 	if (err) {
 	    syslog(LOG_ERR, "No username provided by remote: %s",
 		   gensio_err_to_str(err));
@@ -488,20 +626,12 @@ certauth_event(struct gensio *io, void *user_data, int event, int ierr,
 	}
 	uid = pw->pw_uid;
 	gid = pw->pw_gid;
-
-	pam_err = pam_start(progname, username, &auth_conv, &pamh);
-	if (pam_err != PAM_SUCCESS) {
-	    syslog(LOG_ERR, "pam_start failed for %s: %s", username,
-		   pam_strerror(pamh, pam_err));
-	    return GE_AUTHREJECT;
-	}
-	pam_started = true;
 	homedir = pw->pw_dir;
 
 	len = snprintf(authdir, sizeof(authdir), "%s/.gtlssh/allowed_certs/",
 		       pw->pw_dir);
-	err = gensio_control(io, 0, false, GENSIO_CONTROL_CERT_AUTH,
-			     authdir, &len);
+	err = gensio_control(io, 0, GENSIO_CONTROL_SET,
+			     GENSIO_CONTROL_CERT_AUTH, authdir, &len);
 	if (err) {
 	    syslog(LOG_ERR, "Could not set authdir %s: %s", authdir,
 		   gensio_err_to_str(err));
@@ -521,23 +651,29 @@ certauth_event(struct gensio *io, void *user_data, int event, int ierr,
 		   auxdata[0] ? auxdata[0] : "");
 	    return GE_AUTHREJECT;
 	}
-	if (!ierr)
+	if (!ierr) {
 	    syslog(LOG_INFO, "Accepted certificate for %s\n", username);
+	    authed_by_cert = true;
+	}
 	return GE_NOTSUP;
 
     case GENSIO_EVENT_PASSWORD_VERIFY:
-	passwd = (char *) buf;
-	pam_err = pam_authenticate(pamh, PAM_SILENT);
-	passwd = NULL;
-	if (pam_err == PAM_AUTH_ERR) {
-	    return GE_NOTSUP;
-	} else if (pam_err != PAM_SUCCESS) {
-	    syslog(LOG_ERR, "pam_authenticate failed for %s: %s", username,
-		   pam_strerror(pamh, pam_err));
-	    return GE_INVAL;
-	}
-	syslog(LOG_INFO, "Accepted password for %s\n", username);
-	return 0;
+	passwd = strdup((char *) buf);
+	if (!passwd)
+	    return GE_NOMEM;
+	err = get_2fa(io);
+	if (err)
+	    return err;
+	return GE_NOTSUP;
+
+    case GENSIO_EVENT_2FA_VERIFY:
+	len_2fa = *buflen;
+	val_2fa = malloc(len_2fa + 1);
+	if (!val_2fa)
+	    return GE_NOMEM;
+	memcpy(val_2fa, buf, len_2fa);
+	val_2fa[len_2fa] = '\0';
+	return GE_NOTSUP;
 
     default:
 	return GE_NOTSUP;
@@ -573,6 +709,23 @@ gensio_pam_cb(int num_msg, const struct pam_message **msg,
 	switch (style) {
 	case PAM_PROMPT_ECHO_OFF:
 	case PAM_PROMPT_ECHO_ON:
+	    /* If we have passwd or 2fa data, just supply it. */
+	    if (passwd) {
+		reply[i].resp = passwd;
+		if (!reply[i].resp)
+		    goto out_err;
+		passwd = NULL;
+		break;
+	    } else if (val_2fa) {
+		reply[i].resp = val_2fa;
+		if (!reply[i].resp)
+		    goto out_err;
+		val_2fa = NULL;
+		break;
+	    }
+	    if (!interactive_login)
+		goto out_err;
+
 	    if (msgdata) {
 		err = write_str_to_gensio(msgdata, io, &timeout, true);
 		if (err)
@@ -642,7 +795,7 @@ new_rem_io(struct gensio *io, struct gdata *ginfo)
     bool set_uid = true;
 
     len = 0;
-    err = gensio_control(io, 0, true, GENSIO_CONTROL_SERVICE,
+    err = gensio_control(io, 0, GENSIO_CONTROL_GET, GENSIO_CONTROL_SERVICE,
 			 NULL, &len);
     if (err) {
 	gensio_time timeout = {10, 0};
@@ -657,7 +810,7 @@ new_rem_io(struct gensio *io, struct gdata *ginfo)
 	syslog(LOG_ERR, "Could not allocate service memory");
 	goto out_free;
     }
-    err = gensio_control(io, 0, true, GENSIO_CONTROL_SERVICE,
+    err = gensio_control(io, 0, GENSIO_CONTROL_GET, GENSIO_CONTROL_SERVICE,
 			 service, &len);
     if (err) {
 	syslog(LOG_ERR, "Could not get service(2): %s",
@@ -678,7 +831,7 @@ new_rem_io(struct gensio *io, struct gdata *ginfo)
 	    goto out_free;
 	}
 	/* Dummy out the program, we will set it later with a control. */
-	s = alloc_sprintf("stdio(stderr-to-stdout),dummy");
+	s = alloc_sprintf("stdio(stderr-to-stdout,readbuf=16384),dummy");
     } else if (strstartswith(service, "login:")) {
 	char *str = strchr(service, ':') + 1;
 
@@ -814,7 +967,7 @@ new_rem_io(struct gensio *io, struct gdata *ginfo)
     pcinfo->io1 = io;
 
     if (login) {
-	err = gensio_control(io, GENSIO_CONTROL_DEPTH_ALL, false,
+	err = gensio_control(io, GENSIO_CONTROL_DEPTH_ALL, GENSIO_CONTROL_SET,
 			     GENSIO_CONTROL_NODELAY, "1", NULL);
 	if (err) {
 	    fprintf(stderr, "Could not set nodelay: %s\n",
@@ -834,7 +987,7 @@ new_rem_io(struct gensio *io, struct gdata *ginfo)
     pcinfo->io2 = pty_io;
 
     if (progv) {
-	err = gensio_control(pty_io, 0, false, GENSIO_CONTROL_ARGS,
+	err = gensio_control(pty_io, 0, GENSIO_CONTROL_SET, GENSIO_CONTROL_ARGS,
 			     (char *) progv, NULL);
 	if (err) {
 		syslog(LOG_ERR, "Setting program arguments failed: %s",
@@ -844,8 +997,8 @@ new_rem_io(struct gensio *io, struct gdata *ginfo)
     }
 
     if (progv || login) {
-	err = gensio_control(pty_io, 0, false, GENSIO_CONTROL_ENVIRONMENT,
-			     (char *) env, NULL);
+	err = gensio_control(pty_io, 0, GENSIO_CONTROL_SET,
+			     GENSIO_CONTROL_ENVIRONMENT, (char *) env, NULL);
 	if (err) {
 	    syslog(LOG_ERR, "set env failed for %s: %s", username,
 		   gensio_err_to_str(err));
@@ -939,7 +1092,19 @@ open_mux(struct gensio *io, struct gdata *ginfo, const char *service)
     struct gensio_os_funcs *o = ginfo->o;
     struct gensio *mux_io;
     int err;
-    static const char *isclient[2] = { "mode=server", NULL };
+    /*
+     * The buffer sizes are carefully chosen here to mesh with ssl and
+     * mux.  ssl can encrypt up to 16384 bytes at a time, and the
+     * overhead of a mux data packet is 10 bytes.  So we set this up
+     * so writes can be up to 16384 * 4 bytes total.  On the read size of
+     * mux, each read packet has a 3-byte overhead in the buffer, so
+     * we will be getting 16374 bytes of data, + 3 for overhead, 64
+     * packets is 1048128 bytes.
+     */
+    static const char *isclient[4] = { "mode=server",
+				       "writebuf=65496",
+				       "readbuf=1048128",
+				       NULL };
 
     err = mux_gensio_alloc(io, isclient, o, mux_event, ginfo, &mux_io);
     if (err) {
@@ -966,11 +1131,13 @@ handle_new(struct gensio_runner *r, void *cb_data)
     int err;
     const char *ssl_args[] = { ginfo->key, ginfo->cert, "mode=server", NULL };
     const char *certauth_args[] = { "mode=server", "allow-authfail", NULL,
-				    NULL };
+				    NULL, NULL };
     struct gensio *ssl_io, *certauth_io, *top_io;
     gensiods len;
     char tmpservice[20];
+    const char *pn;
     bool interactive = false;
+    unsigned int i;
 
     o->free_runner(r);
 
@@ -987,8 +1154,13 @@ handle_new(struct gensio_runner *r, void *cb_data)
 	exit(1);
     }
 
+    i = 2;
     if (pw_login)
-	certauth_args[2] = "enable-password";
+	certauth_args[i++] = "enable-password";
+
+    if (do_2fa)
+	certauth_args[i++] = "enable-2fa";
+
     err = certauth_gensio_alloc(ssl_io, certauth_args, o,
 				certauth_event, NULL, &certauth_io);
     if (err) {
@@ -1003,11 +1175,25 @@ handle_new(struct gensio_runner *r, void *cb_data)
 	exit(1);
     }
 
+    aux_data_len = sizeof(aux_data);
+    err = gensio_control(certauth_io, 0, GENSIO_CONTROL_GET,
+			 GENSIO_CONTROL_REM_AUX_DATA,
+			 (char *) &aux_data, &aux_data_len);
+    if (err)
+	aux_data_len = 0;
+
+    if (aux_data_len >= sizeof(aux_data)) {
+	aux_data.flags = ntohl(aux_data.flags);
+
+	if (aux_data.flags & GTLSSH_AUX_FLAG_NO_INTERACTIVE)
+	    interactive_login = false;
+    }
+
     /* FIXME - figure out a way to unstack certauth_io after authentication */
 
     len = sizeof(tmpservice);
-    err = gensio_control(certauth_io, 0, true, GENSIO_CONTROL_SERVICE,
-			 tmpservice, &len);
+    err = gensio_control(certauth_io, 0, GENSIO_CONTROL_GET,
+			 GENSIO_CONTROL_SERVICE, tmpservice, &len);
     if (err) {
 	gensio_time timeout = {10, 0};
 	write_str_to_gensio("Could not get service\n", certauth_io,
@@ -1020,42 +1206,55 @@ handle_new(struct gensio_runner *r, void *cb_data)
     else
 	top_io = certauth_io;
 
+    pn = progname;
+    if (pam_cert_auth_progname && authed_by_cert)
+	pn = pam_cert_auth_progname;
     ginfo->rem_io = top_io;
     gensio_conv.appdata_ptr = ginfo;
-    pam_err = pam_set_item(pamh, PAM_CONV, &gensio_conv);
-    if (pam_err) {
-	syslog(LOG_ERR, "Unable to set PAM_CONV");
+    pam_err = pam_start(pn, username, &gensio_conv, &pamh);
+    if (pam_err != PAM_SUCCESS) {
+	syslog(LOG_ERR, "pam_start failed for %s: %s", username,
+	       pam_strerror(pamh, pam_err));
 	exit(1);
     }
+    pam_started = true;
 
-    if (!gensio_is_authenticated(top_io)) {
+    if (!gensio_is_authenticated(top_io) || pam_cert_auth_progname) {
 	int tries = 3;
 
-	do {
+	if (!interactive_login)
+	    tries = 1;
+	pam_err = pam_authenticate(pamh, 0);
+	while (pam_err != PAM_SUCCESS && tries > 0) {
 	    gensio_time timeout = {10, 0};
 
 	    err = write_str_to_gensio("Permission denied, please try again\n",
 				      top_io, &timeout, true);
 	    if (err) {
-		syslog(LOG_INFO, "Error getting password: %s\n",
+		syslog(LOG_INFO, "Error writing password prompt: %s\n",
 		       gensio_err_to_str(err));
 		exit(1);
 	    }
 	    pam_err = pam_authenticate(pamh, 0);
 	    tries--;
-	} while (pam_err != PAM_SUCCESS && tries > 0);
+	}
 
 	if (pam_err != PAM_SUCCESS) {
 	    gensio_time timeout = {10, 0};
 
-	    err = write_str_to_gensio("Too many tries, giving up\n",
-				      top_io, &timeout, true);
+	    if (interactive_login) {
+		err = write_str_to_gensio("Too many tries, giving up\n",
+					  top_io, &timeout, true);
+		syslog(LOG_ERR, "Too many login tries for %s\n", username);
+	    } else {
+		err = write_str_to_gensio("Non-interactive login only\n",
+					  top_io, &timeout, true);
+		syslog(LOG_ERR, "Non-interactive login only %s\n", username);
+	    }
 	    if (err) {
-		syslog(LOG_INFO, "Error getting password: %s\n",
+		syslog(LOG_INFO, "Error writing login error: %s\n",
 		       gensio_err_to_str(err));
-		exit(1);
 	    }
-	    syslog(LOG_ERR, "Too many login tries for %s\n", username);
 	    exit(1);
 	}
 	syslog(LOG_INFO, "Accepted password for %s\n", username);
@@ -1247,12 +1446,17 @@ help(int err)
     printf("  --allow-password - Allow password-based logins.\n");
     printf("  --oneshot - Do not fork new connections, do one and exit.\n");
     printf("  --nodaemon - Do not daemonize.\n");
+    printf("  --nointeractive - Do not do interactive login queries.\n");
     printf("  --nosctp - Disable SCTP support.\n");
     printf("  --notcp - Disable TCP support.\n");
     printf("  --other_acc <accepter> - Allows the user to specify the\n");
     printf("     accepter used by gtlsshd, in addition to sctp and tcp.\n");
     printf("  -4 - Do IPv4 only.\n");
     printf("  -6 - Do IPv6 only.\n");
+    printf("  --do-2fa - Have the client get 2-factor auth data.\n");
+    printf("  --pam-cert-auth <name> - When doing a certificate auth,\n");
+    printf("     use the name as the PAM program name and run the PAM auth\n");
+    printf("     after the certificate auth succeeds.  For 2-factor auth.\n");
     printf("  -P, --pidfile <file> - Create the given pidfile.\n");
     printf("  -h, --help - This help\n");
     exit(err);
@@ -1296,12 +1500,13 @@ main(int argc, char *argv[])
     const char *keyfile = default_keyfile;
     const char *certfile = default_certfile;
     const char *configfile = default_configfile;
-    int port = 852;
+    unsigned int port = 852;
     char *s;
     bool notcp = false, nosctp = false;
     bool daemonize = true;
     const char *iptype = ""; /* Try both IPv4 and IPv6 by default. */
     const char *other_acc_str = NULL;
+    struct gensio_os_proc_data *proc_data;
 
     localport_err = pr_localport;
 
@@ -1317,8 +1522,8 @@ main(int argc, char *argv[])
 	    arg++;
 	    break;
 	}
-	if ((rv = cmparg_int(argc, argv, &arg, "-p", "--port",
-			     &port)))
+	if ((rv = cmparg_uint(argc, argv, &arg, "-p", "--port",
+			      &port)))
 	    ;
 	else if ((rv = cmparg(argc, argv, &arg, "-f", "--configfile",
 			      &configfile)))
@@ -1340,12 +1545,19 @@ main(int argc, char *argv[])
 	    permit_root = true;
 	else if ((rv = cmparg(argc, argv, &arg, NULL, "--allow-password", NULL)))
 	    pw_login = true;
+	else if ((rv = cmparg(argc, argv, &arg, NULL, "--do-2fa", NULL)))
+	    do_2fa = true;
+	else if ((rv = cmparg(argc, argv, &arg, NULL, "--pam-cert-auth",
+			      &pam_cert_auth_progname)))
+	    ;
 	else if ((rv = cmparg(argc, argv, &arg, NULL, "--oneshot", NULL)))
 	    oneshot = true;
 	else if ((rv = cmparg(argc, argv, &arg, NULL, "--nodaemon", NULL)))
 	    daemonize = false;
 	else if ((rv = cmparg(argc, argv, &arg, "-P", "--pidfile", &pid_file)))
 	    ;
+	else if ((rv = cmparg(argc, argv, &arg, NULL, "--nointeractive", NULL)))
+	    interactive_login = false;
 	else if ((rv = cmparg(argc, argv, &arg, "-4", NULL, NULL)))
 	    iptype = "ipv4,0.0.0.0,";
 	else if ((rv = cmparg(argc, argv, &arg, "-6", NULL, NULL)))
@@ -1385,6 +1597,14 @@ main(int argc, char *argv[])
     }
     o->vlog = do_vlog;
 
+    rv = gensio_os_proc_setup(o, &proc_data);
+    if (rv) {
+	fprintf(stderr, "Could not setup process data: %s\n",
+		gensio_err_to_str(rv));
+	o->free_funcs(o);
+	return 1;
+    }
+
     ginfo.o = o;
 
     ginfo.key = alloc_sprintf("key=%s", keyfile);
@@ -1405,7 +1625,7 @@ main(int argc, char *argv[])
     }
 
     if (!notcp) {
-	s = alloc_sprintf("tcp,%s%d", iptype, port);
+	s = alloc_sprintf("tcp(readbuf=20000),%s%d", iptype, port);
 	if (!s) {
 	    fprintf(stderr, "Could not allocate tcp descriptor\n");
 	    return 1;
@@ -1429,7 +1649,7 @@ main(int argc, char *argv[])
     }
 
     if (!nosctp) {
-	s = alloc_sprintf("sctp,%s%d", iptype, port);
+	s = alloc_sprintf("sctp(readbuf=20000),%s%d", iptype, port);
 	if (!s) {
 	    fprintf(stderr, "Could not allocate sctp descriptor\n");
 	    return 1;
@@ -1572,6 +1792,18 @@ main(int argc, char *argv[])
 
     o->free_waiter(ginfo.waiter);
 
+    gensio_os_proc_cleanup(proc_data);
+    o->free_funcs(o);
+
+    if (passwd) {
+	memset(passwd, 0, strlen(passwd));
+	free(passwd);
+    }
+    if (val_2fa) {
+	memset(val_2fa, 0, len_2fa);
+	free(val_2fa);
+    }
+
     if (pam_cred_set) {
 	pam_err = pam_setcred(pamh, PAM_DELETE_CRED | PAM_SILENT);
 	if (pam_err != PAM_SUCCESS)
